#ifndef MHCL_BASE_SRCBASE_CORE_H
#define MHCL_BASE_SRCBASE_CORE_H

/*!F::Core declarations and definitions for Machel's C projects.
 *
 *  ::Purpose:
 *    Contains things common for every Machel's C project:
 *    - headers
 *    - helper macros and functions
 *    - thread-local storage facilities
 *    - type definitions and aliases
 *    - GNU extensions: attributes
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    Created:            2013-12-16
 *    (c) Copyright:      "Machel" 2013
 *    License:            GNU Public License
 */




/* 32/64b */
/* -------------------------------------------------------------------------- */

/* Functions that take 32b parameters (like lseek) wil now take 64b parameters.
 * Thus lseek on 32b system will be transparently replaced by lseek64.
 * Allows 32b systems to use 64b large files and large data.
 */
#define _FILE_OFFSET_BITS 64




/* Common headers */
/* -------------------------------------------------------------------------- */
#include <assert.h>     // assert()
#include <stddef.h>     // NULL, size_t, ptrdiff_t, offsetof
#include <stdbool.h>    // bool, true, false


/* Pathname component separator */
#define PNCS '/'




/* Thread local storage. */
/* -------------------------------------------------------------------------- */

/* Thread local storage. Different defines for different compilers.
 * In-code usage:
 *    / extern/static __tls int var; /
 *
 * !!! C++:
 *    GCC do not allow construct/destruct for '__thread' variables!
 *    use C++'s 'thread_local'
 */
#if defined(__GNUC__) || defined(__GNUG__) \
   || defined(__SUNPRO_C) || defined(__SUNPRO_CC) \
   || defined(__ICC) || defined(__INTEL_COMPILER)
#  define __tls __thread
#elif defined(_MSC_VER) && defined(__cplusplus)
#  define __tls __declspec(thread)
#else
#  define __tls
#endif




/* Type definitions. */
/* -------------------------------------------------------------------------- */

/* ssize_t - signed size type. For Windows and for other systems.
 */
#if defined(_MSC_VER)
#  include <BaseTsd.h>
   typedef SSIZE_T ssize_t;
#else
#  include <unistd.h>
#endif


/* SIZE_MAX - max size_t value
 */
#include <limits.h>        // <C++/Solaris>SIZE_MAX
#if HAVE_STDINT_H
#include <stdint.h>        // <C99>SIZE_MAX
#endif
#ifndef SIZE_MAX
#define SIZE_MAX ((size_t)(-1))
#endif


/*!:: Special type that can take three values:
 *       negative (less then 0), neutral (0), positive (greater then 0).
 */
enum trool { tr_neg=-1, tr_less=-1, tr_neut=0, tr_equal=0, tr_pos=1, tr_great=1 };
typedef enum trool trool;


/* aliases for built-in types */
typedef unsigned short ushort;
typedef signed short sshort;
typedef unsigned int uint;
typedef signed int sint;
typedef unsigned long ulong;
typedef signed long slong;

/* file descriptor */
typedef int FD;




/* Helper functions. */
/* -------------------------------------------------------------------------- */

/*!:: Stringize macro constant
 */
#define __STR__(x) #x
#define _STR_(x) __STR__(x)


/*!:: Ask for short answer and get input.
 */
#include <stdio.h>
#include <stdarg.h>        // va_*
static inline
bool
ask_yn (
   const char* yes_chars,  //! any of this chars is 'yes' answer
   const char* fmt,        //! format string conforming to printf specification
   ...                     //! data to print
){
   int ch = '\0';
   if (fmt != NULL) {
      va_list ap;
      va_start(ap, fmt);
      vfprintf(stdout, fmt, ap);
      fflush(stdout);
      va_end(ap);
   }
   ch = getchar();  //==getc(stdin)
   if (ch == '\n') { ch = getchar(); }    //scanf() leaves \n in input buffer
   while (*yes_chars != '\0') {
      if (ch == *yes_chars) { return true; }
      ++yes_chars;
   }
   return false;
}




/* GCC extensions. */
/* -------------------------------------------------------------------------- */

/* Function attributes.
 * Define attribute macros only if compiler is GCC or compatible (particulary
 * Clang), otherwise define empty macros that do nothing.
 */
#if defined(__GNUC__) || defined(__GNUG__)

   // function has no effects, ONLY: ret val, may depend on outside var,
   // not on volatile mem etc.
#  define __attr_pure__ __attribute__((__pure__))

   // function has no effects, ONLY: ret val, may NOT depend on outside var,
   // not on volatile mem etc.
   // examining data pointed by ptr args is depending on outside data!
#  define __attr_const__ __attribute__((__const__))

   // list contains parameter numbers starting from 1 that can not be NULL.
#  define __attr_nonnull__(...) __attribute__((__nonnull__(__VA_ARGS__)))

   // return value can not be NULL
#  define __attr_ret_nonnull__ __attribute__((__returns_nonnull__))

   // return value can not be ignored
#  define __attr_wures__ __attribute__((__warn_unused_result__))

   // num of parameter that specifies size of memory that function will alloc
#  define __attr_alloc_size__(PARAM_NUM) __attribute__((__alloc_size__(PARAM_NUM)))

   // mallo-like function - see documentation
#  define __attr_malloc__ __attribute__((__malloc__))

   // printf-like function - see documentation
#  define __attr_format__(FUN,FMT_NUM,VA_NUM) __attribute__((__format__(FUN,FMT_NUM,VA_NUM)))

   // num of parameter from back starting from 0 that is sentinel in va function
#  define __attr_sentinel__(PARAM_NUM_BACK) __attribute__((__sentinel__(PARAM_NUM_BACK)))

   // attribute can be unused, suppress warnings about it
#  define __attr_unused__ __attribute((__unused__))

#else /*- __GNUC__ -*/

#  define __attribute__(x) /*-nothing-*/
#  define __attr_pure__ /*-nothing-*/
#  define __attr_const__ /*-nothing-*/
#  define __attr_nonnull__(...) /*-nothing-*/
#  define __attr_ret_nonnull__ /*-nothing-*/
#  define __attr_wures__ /*-nothing-*/
#  define __attr_alloc_size__(PARAM_NUM) /*-nothing-*/
#  define __attr_malloc__ /*-nothing-*/
#  define __attr_format__(FUN,FMT_NUM,VA_NUM) /*-nothing-*/
#  define __attr_sentinel__(PARAM_NUM_BACK) /*-nothing-*/
#  define __attr_unused__ /*-nothing-*/
#endif /*- __GNUC__ -*/

// for descriptions of attributes see documentation at:
//    http://gcc.gnu.org/onlinedocs/gcc-4.3.0/gcc/Function-Attributes.html








/* REMAINDERS */

/* macros defined by compilers */
#if defined(__GNUC__) || defined(__GNUG__)
   /* <C&C++> <C++> GNU GCC/G++, Clang/LLVM, IBM XL C/C++. ------------------- */
#elif (defined(__GNUC__) || defined(__GNUG__)) && !(defined(__clang__) || defined(__INTEL_COMPILER))
   /* GNU GCC/G++. -----------------------------------only------ */
#elif defined(__clang__)
   /* <C&C++> Clang/LLVM. ---------------------------------------------- */
#elif defined(__IBMC__) || defined(__IBMCPP__)
   /* <C> <C++> IBM XL C/C++. -------------------------------------------- */
#elif defined(__ICC) || defined(__INTEL_COMPILER)
   /* <C&C++> <C&C++> Intel ICC/ICPC. ------------------------------------------ */
#elif defined(__HP_cc) || defined(__HP_aCC)
   /* <C> <C++> Hewlett-Packard C/aC++. ---------------------------------- */
#elif defined(_MSC_VER) && defined(__cplusplus)
   /* <C&C++> <C++> Microsoft Visual Studio. --------------------------------- */
#elif defined(__PGI) && defined(__cplusplus)
   /* <C&C++> <C++> Portland Group PGCC/PGCPP. ------------------------------- */
#elif defined(__SUNPRO_C) || defined(__SUNPRO_CC)
   /* <C> <C++> Oracle Solaris Studio. ----------------------------------- */
#endif

/* macros defined for different operating systems */
// __unix__    - any unix-like, NOT Apple-like
//             - HP-UX
// __linux__   - linuxes
// _POSIX*     - linuxes
// BSD         - any BSD or specific: __DragonFly__, __Free/Net/OpenBSD__
//             - OSX, iOS
// __APPLE__ && __MACH__   - OSX, iOS     + TARGET_OS_MAC (OSX), TARGET_OS_IPHONE etc.
// __sun && __SVR4
//
// __hpux      - HP-UX
// _AIX        - Ibm AIX
//
// _WIN64, _WIN32             - WinXP,Vista,7,8
// __CYGWIN__ && !_WIN32      - Cygwin (only 32b version exist)
// __MINGW32/64_              - MinGW

#if defined(__unix__)
#include <sys/param.h>
#if defined(BSD)
   /* BSD (DragonFly BSD, FreeBSD, OpenBSD, NetBSD) ----------- */
#endif
#endif

#if defined(__unix__) || (defined(__APPLE__) && defined(__MACH__))
#include <sys/param.h>
#if defined(BSD)
   /* BSD (DragonFly BSD, FreeBSD, OpenBSD, NetBSD) + OSX, iOS, Darwin ----------- */
#endif
#endif

#if defined(__APPLE__) && defined(__MACH__)
   /* Apple OSX and iOS (Darwin). ------------------------------ */
#include <TargetConditionals.h>
#if TARGET_IPHONE_SIMULATOR == 1
   /* iOS in Xcode simulator */
#elif TARGET_OS_IPHONE == 1
   /* iOS on iPhone, iPad, etc. */
#elif TARGET_OS_MAC == 1
   /* OSX */
#endif
#endif

/* UNIX line and POSIX compliant - currently BSD,linux,OSX,Solaris */
#if !defined(_WIN32) && (defined(__unix__) || defined(__unix) || (defined(__APPLE__) && defined(__MACH__)))
   /* UNIX-style OS. ------------------------------------------- */
#include <unistd.h>
#if defined(_POSIX_VERSION)
   /* POSIX compliant */
#endif
#endif




#endif /*- MHCL_BASE_SRCBASE_CORE_H -*/
