#ifndef MHCL_BASE_SRCBASE_DEBUG_H
#define MHCL_BASE_SRCBASE_DEBUG_H

/*!F::Debug procedures - external interface file.
 *
 *  ::Provides function declarations for other modules.
 */




/* current project */
#include "common_mhlbase.h"
#include "logs.h"

/* mh library */
#include "core.h"
#include "logs.h"

/* libc */
#include <stdlib.h>     // abort()
#include <stdio.h>      // printf(), fflush()




#define LogAssert(...) do { log_string(LOG_DEF_STREAM, LOG_PREF_ASSERT, __FILE__, __LINE__, __func__, __VA_ARGS__); } while(0)
#define ___Sentinel(...) do { LogAssert(__VA_ARGS__); } while(0)




/* DEBUG */
/* ========================================================================== */
#ifdef DEBUG


/*!1-6
 * :: Log debug messages.
 * :: Print messages of certain type to 'LOG_DEF_STREAM'. Messages are
 *    consistently formatted with use of adequate prefix, current file name,
 *    line number, function name and additional printf-like string customized
 *    by user.
 * :: Intended for printing various debug messages.
 * :: These macro functions are in effect only if 'DEBUG' macro is defined.
 */
#define DLog(...) do { log_string(LOG_DEF_STREAM, LOG_PREF_DEBUG, __FILE__, __LINE__, __func__, __VA_ARGS__); } while(0)
#define DLogInfo(...) do { log_string(LOG_DEF_STREAM, LOG_PREF_INFO, __FILE__, __LINE__, __func__, __VA_ARGS__); } while(0)
#define DLogWarn(...) do { log_string(LOG_DEF_STREAM, LOG_PREF_WARN, __FILE__, __LINE__, __func__, __VA_ARGS__); } while(0)
#define DLogError(...) do { log_string(LOG_DEF_STREAM, LOG_PREF_ERROR, __FILE__, __LINE__, __func__, __VA_ARGS__); } while(0)
#define DLogTest(...) do { log_string(LOG_DEF_STREAM, LOG_PREF_TEST, __FILE__, __LINE__, __func__, __VA_ARGS__); } while(0)
#define DLogAssert(...) do { log_string(LOG_DEF_STREAM, LOG_PREF_ASSERT, __FILE__, __LINE__, __func__, __VA_ARGS__); } while(0)


/*!1-6
 * :: Checks.
 * :: Check if condition is true and take actions if it is not. Log message or
 *    eventually abort.
 * :: This macro functions are in effect only if 'DEBUG' macro is defined.
 */
#define ___Dcheck(Cond, ...) do { if(!(Cond)) { DLogError(__VA_ARGS__); } } while(0)
#define ___Dassert(Cond, ...) do { if(!(Cond)) { DLogAssert(__VA_ARGS__); abort(); } } while(0)
#define ___Dsentinel(...) do { DLogAssert(__VA_ARGS__); } while(0)

#define ___DParam(Param, Cond) do { if(!((Param) Cond)) { DLogAssert("'"#Param"' parameter invalid"); abort(); } } while(0)
#define ___DInter(Param, Cond) do { if(!((Param) Cond)) { DLogAssert("'"#Param"' internal check failed"); abort(); } } while(0)
#define ___DReturn(Param, Cond) do { if(!((Param) Cond)) { DLogAssert("'"#Param"' return expression invalid"); abort(); } } while(0)

#else
#define DLog(...)
#define DLogInfo(...)
#define DLogWarn(...)
#define DLogError(...)
#define DLogTest(...)
#define DLogAssert(...)
#define ___Dcheck(Cond, ...)
#define ___Dassert(Cond, ...)
#define ___Dsentinel(...)
#define ___DParam(Param, Cond)
#define ___DInter(Param, Cond)
#define ___DReturn(Param, Cond)
#endif /*- DEBUG -*/




/* DEBUG2 */
/* ========================================================================== */
#ifdef DEBUG2


/*!1-4
 * :: Log program state.
 * :: Simply print printf-like formatted message to 'stdout' preceded by
 *    function name and line number, then flush stream.
 * :: Intended for printing state of program during debugging time,
 *    ie. value of variables.
 */
#define ___DPr(...) do { \
   fprintf(stdout, ">> %s:%d: ", __func__, __LINE__); \
   fprintf(stdout, __VA_ARGS__); \
   fflush(stdout); \
} while(0)
#define ___DVs(VAR) do { \
   fprintf(stdout, #VAR"=|%s|", (VAR)); \
   fflush(stdout); \
} while(0)
#define ___DVd(VAR) do { \
   fprintf(stdout, #VAR"=%d", (VAR)); \
   fflush(stdout); \
} while(0)
#define ___DVz(VAR) do { \
   fprintf(stdout, #VAR"=%zu", (VAR)); \
   fflush(stdout); \
} while(0)
#define ___DVp(VAR) do { \
   fprintf(stdout, #VAR"=%p", (VAR)); \
   fflush(stdout); \
} while(0)

#else
#define ___DPr(...)
#define ___DVs(VAR)
#define ___DVd(VAR)
#define ___DVz(VAR)
#define ___DVp(VAR)
#endif /*- DEBUG2 -*/




#ifdef DEBUG
bool debug_all_tests_run (void)
      __attr_wures__
      ;
#endif




#endif /*- MHCL_BASE_SRCBASE_DEBUG_H -*/
