/*!F::Errors management - main implementation file.
 *
 *  ::Purpose:
 *    Detecting, propagating and handling of errors in consistent and convenient
 *    way. Thread-safe.
 *    Test for specified conditions and take actions on failing, like jumping to
 *    label.
 *
 *    It is part of error handling facilities which involve error.s, debug.c and
 *    faults.c.
 *    It covers OS errors: access denied, bad permissions, operation not supported,
 *    no memory, quota exceeded etc.
 *    It doesn't cover: programmer's errors (covered by debug.c), user's errors
 *    (covered by faults.c).
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *
 *  ::Functionality:
 *    - propagate error information from within a function
 *    - set/cancel errors
 *    - handle errors
 *
 *  ::Limitations:
 *    - information is limited to fixed size arrays
 *    - error handling uses labels that must be placed within scope
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2013-12-16
 *    (c) Copyright:    Maciej Helminiak 2013
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* snprintf(), vsnprintf():
 *    _BSD_SOURCE || _XOPEN_SOURCE >= 500 || _ISOC99_SOURCE
 *    || _POSIX_C_SOURCE >= 200112L;
 *    or -std=c99
 */
#define _POSIX_C_SOURCE 200112L




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"
#include "debug.h"
#include "logs.h"

/* libc */
#include <errno.h>         // errno
#include <stdarg.h>        // va_*
#include <stdio.h>         // printf(), fflush()
#include <string.h>        // strerror()

/* current module */
#include "errors.h"




/* Error structure */
/* -------------------------------------------------------------------------- */

/*!:: Error object instance.
 * :: Used by error setting and handling routines.
 *    Created in thread local storage.
 */
__tls st_mh_error __err;




/* Error codes */
/* -------------------------------------------------------------------------- */

/*!:: Labels for enumeration constants in 'errt_en'.
 * :: Printable, unmodifiable string literals describing each enum constant.
 *    Constant itself minus MH_ET_STARTING_NUMBER can be used as index in this
 *    array.
 */
static const char* const errnum_Labels[errnum_count-MH_ET_STARTING_NUMBER] = {
   "Parameter out of range",
   "Data length exceeded",
   "Integer overflow",
   "Buffer overflow",
   "Output error",
   "Functionality not implemented",
   "State changed during operation",
   "Subject exist",
   "Format is wrong",
   "all types of errors",
   "unknown error"
};




/* Error manipulating routines. */
/* -------------------------------------------------------------------------- */

/*!1-3
 * :: Return cstring with current error's signature, trace, message.
 * ::Preconditions: NONE.
 * ::Errors: NONE.
 */
char*
err_ret_sig (
   void
){
   return &__err.sig[0];
}

char*
err_ret_trace (
   void
){
   return &__err.trace[0];
}

char*
err_ret_msg (
   void
){
   return &__err.msg[0];
}

/*!1-3
 * :: Place formatted in printf-like manner cstring in error's signature, trace,
 *    message.
 * ::Preconditions:
 * ::Errors:
 *    no Errors.
 * ..Note: vsnprintf() returns -1 on output error, but it is ignored as we are
 *    here because of error condition anyway.
 * ::Parameters:
 *    `fmt`: format cstring
 *    `...`: arguments
 */
char*
err_carve_sig (
   const char* fmt,
   ...
){
   va_list ap;
   va_start(ap, fmt);

   vsnprintf(&__err.sig[0], ERROR_SIGNATURE_BYTES, fmt, ap);

   va_end(ap);

   return &__err.sig[0];
}

char*
err_carve_trace (
   const char* fmt,
   ...
){
   va_list ap;
   va_start(ap, fmt);

   vsnprintf(&__err.trace[0], ERROR_TRACE_BYTES, fmt, ap);

   va_end(ap);

   return &__err.trace[0];
}

char*
err_carve_msg (
   const char* fmt,
   ...
){
   va_list ap;
   va_start(ap, fmt);

   vsnprintf(&__err.msg[0], ERROR_MESSAGE_BYTES, fmt, ap);

   va_end(ap);

   return &__err.msg[0];
}

/*!1-2
 * :: Append formatted in printf-like manner cstring to error's trace, message.
 * ::Preconditions:
 * ::Errors:
 *    no Errors.
 * ..Note: vsnprintf() returns -1 on output error, but it is ignored as we are
 *    here because of error condition anyway.
 * ::Parameters:
 *    `fmt`: format cstring
 *    `...`: arguments
 */
char*
err_append_trace (
   const char* fmt,
   ...
){
   va_list ap;
   va_start(ap, fmt);

   const size_t nc = strlen(__err.trace);
   vsnprintf(&__err.trace[nc], ERROR_TRACE_BYTES-nc, fmt, ap);

   va_end(ap);

   return &__err.trace[0];
}

char*
err_append_msg (
   const char* fmt,
   ...
){
   va_list ap;
   va_start(ap, fmt);

   const size_t nc = strlen(__err.msg);
   vsnprintf(&__err.msg[nc], ERROR_MESSAGE_BYTES-nc, fmt, ap);

   va_end(ap);

   return &__err.msg[0];
}

/*!:: Return pointer to string describing given error code.
 * ::Errors:
 */
const char*
err_descr (
   int err     //! error code
){
   /* error in range of errno values defined in libc */
   if (err < MH_ET_STARTING_NUMBER) {
      return strerror(err);
   }
   /* error type defined outside libc */
   else {
      return errnum_Labels[err - MH_ET_STARTING_NUMBER];
   }
}




/* Error handling routines. */
/* -------------------------------------------------------------------------- */


/*!:: Error code must be different from 'NONE' and must be equal to current
 *    code in error or to constant indicating ALL codes.
 * ::Errors:
 */
bool
err_catch_cond_fulfiled (
   errnum_en errnum        //! error code
){
   if (errno == 0) { return false; }
   if ( ((errnum) != E_NONE) && ((uint)errno == (errnum) || (errnum) == E_ALL_TYPES) )
   {
      return true;
   }

   return false;
}

/*!:: Error information is printed accordingly to specified print action.
 * ::Errors:
 */
void
err_catch_print_action (
   errprint_en errpr
){
   switch (errpr)
   {
      case EPR_NONE:
         /*- nothing -*/
      break;
      case EPR_NOTICE:
         LogError("Error: %s\n", ErrDescr());
      break;
      case EPR_MESSAGE:
         LogError("%s\n", ErrMessage());
      break;
      case EPR_DEBUG_INFO:
         LogError("%s\n\ttype: %s\n\toriginated: %s\n\ttrace: %s\n\tmsg: %s",
            "Error encountered:", ErrDescr(), ErrSignat(), ErrTrace(), ErrMessage());
      break;
      case EPR_UNKNOWN:
      case errprint_count:
      default:
         ___Sentinel("Wrong print error action"); abort();
   };
}
