#ifndef MHCL_BASE_SRCBASE_ERRORS_H
#define MHCL_BASE_SRCBASE_ERRORS_H

/*!F::Errors management - external interface file.
 *
 *  ::Provides function declarations for other modules.
 */




/* mh library */
#include "core.h"

/* current project */
#include "common_mhlbase.h"

/* libc */
#include <errno.h>         // errno
#include <stdlib.h>     // abort()
#include <stdio.h>      // printf(), fflush()




/* Error structure */
/* -------------------------------------------------------------------------- */


/* Number of bytes of cstrings representing error's signature, trace, message
   in error structure. */
#define ERROR_SIGNATURE_BYTES 128
#define ERROR_TRACE_BYTES 1024
#define ERROR_MESSAGE_BYTES 1024


typedef struct st_mh_error st_mh_error;
/*!:: Error structure.
 *    Data describing current error. In conjunction with 'errno' provided by
 *    libc makes overall error description.
 */
struct st_mh_error
{
   char sig[ERROR_SIGNATURE_BYTES]; //! signature - where error originated from
   char trace[ERROR_TRACE_BYTES];   //! subsequent places of error propagation
   char msg[ERROR_MESSAGE_BYTES];   //! message used if more elaborate description
                                    //. then provided by 'errno' needed
   bool aux;                        //! for use by error routines
};


extern __tls st_mh_error __err;




/* Error codes and cathegory */
/* -------------------------------------------------------------------------- */


/* Number of first error in 'enum errnum'. It must be greater then highest
   'errno' code.*/
#define MH_ET_STARTING_NUMBER 4096

/*!:: Error codes.
 *    Error codes complementary to those used by 'errno'.
 */
enum errnum_en {
   E_NONE=0,
   E_PARAMOOR=MH_ET_STARTING_NUMBER,   //! parameter has value out of range
   E_DATALENEXC,        //! data is too long
   E_OVERFLOWINT,       //! integer overflow
   E_OVERFLOWBUF,       //! buffer overflow
   E_OUTPUTERR,         //! output to file, screen, network error
   E_FUNC_NOTIMPL,
   E_STATE_CHANGE,      //! state of subject changed during operation
   E_EXIST,             //! file, object, entity exist
   E_FORMAT_WRONG,      //! file, string, user input is in wrong format
   E_ALL_TYPES,         //! all types
   E_UNKNOWN,
   errnum_count };
typedef enum errnum_en errnum_en;


/*!:: Error print action.
 */
enum errprint_en {
   EPR_NONE=0,       //! do not print
   EPR_NOTICE,       //! short notice
   EPR_MESSAGE,      //! informative message
   EPR_DEBUG_INFO,   //! full debug info
   EPR_UNKNOWN,      //! unknown cathegory
   errprint_count };
typedef enum errprint_en errprint_en;




/* Error manipulating routines. */
/* -------------------------------------------------------------------------- */


/*!1-3
 * :: Return error's signature, trace, message.
 */
extern char* err_ret_sig (void)
               __attr_wures__
               ;
extern char* err_ret_trace (void)
               __attr_wures__
               ;
extern char* err_ret_msg (void)
               __attr_wures__
               ;

/*!1-3
 * :: Carve error's signature, trace, message.
 */
extern char* err_carve_sig (const char* format, ...)
               __attr_nonnull__(1)
               __attr_format__(__printf__,1,2)
               ;
extern char* err_carve_trace (const char* format, ...)
               __attr_nonnull__(1)
               __attr_format__(__printf__,1,2)
               ;
extern char* err_carve_msg (const char* format, ...)
               __attr_nonnull__(1)
               __attr_format__(__printf__,1,2)
               ;

/*!1-3
 * :: Append to error's signature, trace, message.
 */
extern char* err_append_trace (const char* format, ...)
               __attr_nonnull__(1)
               __attr_format__(__printf__,1,2)
               ;
extern char* err_append_msg (const char* format, ...)
               __attr_nonnull__(1)
               __attr_format__(__printf__,1,2)
               ;

/*!1:: Return description of error type.
 */
extern const char* err_descr (int errno_val)
                     __attr_wures__
                     ;




/*!1-5
 * :: Error code.
 * :: Set, get, test, clear current error code.
 */
#define SetErr(errnum) do { \
   errno = (errnum); \
} while(0)

#define GetErr() errno

#define IsErr() errno

#define ClrErrCode() do { \
   errno = 0; \
} while(0)

#define ClrErr() do { \
   __err.sig[0] = '\0'; \
   __err.trace[0] = '\0'; \
   __err.msg[0] = '\0'; \
   __err.aux = false; \
   errno = 0; \
} while(0)




/*!1-4
 * :: Get error parameters.
 * :: Return current error's code description, signature, trace or message.
 */
#define ErrDescr() err_descr(errno)
#define ErrSignat() err_ret_sig()
#define ErrTrace() err_ret_trace()
#define ErrMessage() err_ret_msg()




/* Error triggering routines. */
/* -------------------------------------------------------------------------- */


/*!1-5
 * :: Set error.
 * :: Set error parameters, ev. set custom message, ev. do it conditionally,
 *    ev. do not assign code. Then goto label.
 */
//#define ErrorTrig(errnum, label)
#define ErrorSet(errnum, label) do { \
   if (errno == 0) SetErr(errnum); \
   err_carve_sig("%s:%d", __func__, __LINE__); \
   err_carve_trace("%s:%d", __func__, __LINE__); \
   goto label; \
} while(0)

//#define ErrorTrig_msg(errnum, label, ...)
#define ErrorSetm(errnum, label, ...) do { \
   if (errno == 0) SetErr(errnum); \
   err_carve_sig("%s:%d", __func__, __LINE__); \
   err_carve_trace("%s:%d", __func__, __LINE__); \
   err_carve_msg(__VA_ARGS__); \
   goto label; \
} while(0)

//#define ErrorTrigIf(Cond, errnum, label)
#define ErrorSetIf(Cond, errnum, label) do { \
   if ((Cond)) { \
      if (errno == 0) SetErr(errnum); \
      err_carve_sig("%s:%d", __func__, __LINE__); \
      err_carve_trace("%s:%d", __func__, __LINE__); \
      goto label; \
   } \
} while(0)

//#define ErrorIf(Cond, label)
#define ErrorTrigIf(Cond, label) do { \
   if ((Cond)) { \
      err_carve_sig("%s:%d", __func__, __LINE__); \
      err_carve_trace("%s:%d", __func__, __LINE__); \
      goto label; \
   } \
} while(0)

#define ErrorTrigIfm(Cond, label, ...) do { \
   if ((Cond)) { \
      err_carve_sig("%s:%d", __func__, __LINE__); \
      err_carve_trace("%s:%d", __func__, __LINE__); \
      err_carve_msg(__VA_ARGS__); \
      goto label; \
   } \
} while(0)




/*!1-2
 * :: Overflows.
 * :: Set error if overflow occured.
 */
#define OverflowIntIf(Cond, label) do { \
   if ((Cond)) { \
      if (errno == 0) SetErr(E_OVERFLOWINT); \
      err_carve_sig("%s:%d", __func__, __LINE__); \
      err_carve_trace("%s:%d", __func__, __LINE__); \
      goto label; \
   } \
} while(0)

/*!:: If condition is true buffer overflow occured.
 *    Set error approprietely and goto `label`.
 */
#define OverflowBufIf(Cond, label) do { \
   if ((Cond)) { \
      if (errno == 0) SetErr(E_OVERFLOWBUF); \
      err_carve_sig("%s:%d", __func__, __LINE__); \
      err_carve_trace("%s:%d", __func__, __LINE__); \
      goto label; \
   } \
} while(0)




/* Error handling routines. */
/* -------------------------------------------------------------------------- */


/*!1-2
 * :: Handle (propagate) errors.
 * :: If error is set append trace to parameters and goto label. Eventually
 *    carve custom message.
 */
#define HandleErr(label) do { \
   if (errno != 0) { \
      err_append_trace(" >> %s:%d", __func__, __LINE__); \
      goto label; \
   } \
} while(0)

#define HandleErrm(label, ...) do { \
   if (errno != 0) { \
      err_append_trace(" >> %s:%d", __func__, __LINE__); \
      err_carve_msg(__VA_ARGS__); \
      goto label; \
   } \
} while(0)




/*!:: True if error code in current error is as specified.
 */
extern bool err_catch_cond_fulfiled (errnum_en errnum)
               __attr_wures__
               ;

/*!:: Error information print.
 */
extern void err_catch_print_action (errprint_en errpr)
               ;


/*!1-2
 * :: Catch errors.
 * :: If error is set do any of following actions: print notice, message or
 *    debug info, ignore, continue, retry, exit, abort.
 */
#define CatchError(err_num, err_pr) \
   __err.aux = err_catch_cond_fulfiled((err_num)); \
   if (__err.aux == true) { \
      err_append_trace(" )) %s:%d", __func__, __LINE__); \
      err_catch_print_action((err_pr)); } \
   if (__err.aux == true)




#ifdef DEBUG
/* Used only by tests in debug mode. */
bool errors_all_tests_run (void)
      __attr_wures__
      ;
#endif




#endif /*- MHCL_BASE_SRCBASE_ERRORS_H -*/
