/*!F::User's faults - main implementation file.
 *
 *  ::Purpose:
 *    Record, manage and present to user detected faults.
 *
 *    It is part of error handling facilities which involve error.s, debug.c and
 *    faults.c.
 *    It covers user's mistakes like bad keyboard or file input.
 *    It doesn't cover programmer's mistakes (debug.c) and OS errors (errors.c).
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *
 *  ::Functionality:
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    Created:            2014-05-06
 *    (c) Copyright:      "Machel" 2014
 *    License:            GNU Public License
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"
#include "cstrings.h"
#include "strings.h"

/* libc */

/* current module */
#include "faults.h"
#include "faults_internal.h.inc"
#include "faults_aux.h.inc"




st_fault*
fault_create (
   faultcath_en cath,      //! fault cathegory
   const char* fmt_descr,  //! printf-like format description of fault
   ...                     //! data for format
){
   ___Dassert(fmt_descr, "'fmt_descr' parameter invalid");

   /*: variadic transmission: call helper variadic function */
   va_list ap;
   va_start(ap, fmt_descr);
   st_fault* stFault_new = fault_vcreate(cath, fmt_descr, ap);
   va_end(ap);

   return stFault_new;
}

st_fault*
fault_vcreate (
   faultcath_en cath,      //! fault cathegory
   const char* fmt_descr,  //! printf-like format description of fault
   va_list vlist           //! data for format
){
   ___Dassert(fmt_descr, "'fmt_descr' parameter invalid");

   st_fault* stFault_new = NULL;
   char* cs_descr = NULL;

   stFault_new = mem_alloc(sizeof *stFault_new);
      HandleErr(handle_err);

   cs_descr = cs_vcref(fmt_descr, vlist);
      HandleErr(handle_err);

   stFault_new->cath = cath;
   str_fill(&stFault_new->descr, cs_descr);

   return stFault_new;

  handle_err:
   cs_des(&cs_descr);
   if (stFault_new != NULL) { fault_destroy(&stFault_new); }
   return NULL;
}

void
fault_destroy (
   st_fault** pstFault  //! pointer to fault
){
   ___Dassert(pstFault, "'pstFault' parameter invalid");
   ___Dassert(*pstFault, "'*pstFault' parameter invalid");

   st_fault* stFault = *pstFault;   //helper var
   str_fin(&stFault->descr);
   mem_free(*pstFault);
   return;
}

void
fault_print (
   FILE* fs,               //! filestream
   st_fault* stFault,      //! fault
   const char* pref,       //! string preceding each fault in same line
   const char* post        //! string succeding each fault in same line
){
   ___Dassert(fs, "'fs' parameter invalid");
   ___Dassert(stFault, "'stFault' parameter invalid");

   if (pref == NULL) { pref = ""; }
   if (post == NULL) { post = ""; }
   fprintf(fs, "%s%s%s", pref, stFault->descr.buf, post);
   return;
}




void
faultsarr_add_fault (
   st_faults_arr* stFaultsArr,   //! [out] array of faults
   faultcath_en cath,      //! fault cathegory
   const char* fmt_descr,  //! printf-like format description of fault
   ...                     //! data for format
){
   ___Dassert(stFaultsArr, "'stFaultsArr' parameter invalid");
   ___Dassert(fmt_descr, "'fmt_descr' parameter invalid");

   /*: variadic transmission: call helper variadic function */
   va_list ap;
   va_start(ap, fmt_descr);
   faultsarr_vadd_fault(stFaultsArr, cath, fmt_descr, ap);
   va_end(ap);
   return;
}

void
faultsarr_vadd_fault (
   st_faults_arr* stFaultsArr,   //! [out] array of faults
   faultcath_en cath,      //! fault cathegory
   const char* fmt_descr,  //! printf-like format description of fault
   va_list vlist           //! data for format
){
   ___Dassert(stFaultsArr, "'stFaultsArr' parameter invalid");
   ___Dassert(fmt_descr, "'fmt_descr' parameter invalid");

   st_fault* stFault_new = fault_vcreate(cath, fmt_descr, vlist);
      HandleErr(handle_err);

   extend_if_needed(stFaultsArr, 1);
      HandleErr(handle_err);

   stFaultsArr->buf[stFaultsArr->ns] = stFault_new;
   ++stFaultsArr->ns;
   ++stFaultsArr->caths_num[cath];

   return;

  handle_err:
   if (stFault_new != NULL) { fault_destroy(&stFault_new); }
   return;
}

void
faultsarr_finalize (
   st_faults_arr* stFaultsArr    //! [out] array of faults
){
   ___Dassert(stFaultsArr, "'stFaultsArr' parameter invalid");
   for (size_t i = 0; i < stFaultsArr->ns; ++i) {
      fault_destroy(&stFaultsArr->buf[i]);
   }
   mem_free(stFaultsArr->buf);
   stFaultsArr->na = 0;
   stFaultsArr->ns = 0;
   for (size_t i = Faultcath_none; i < faultcath_count; ++i) {
      stFaultsArr->caths_num[i] = 0;
   }
   return;
}

/*!:: Return fault in buffer inside array at specified index and increment that
 *    index by one.
 *    NULL returned if index if out of array bounds.
 *    After function call `idx` holds next index to retrieve, that is it consist
 *    number of elements retrieved so far.
 * ::Errors:
 *    No errors.
 */
st_fault*                     //! fault in buffer
faultsarr_it (
   const st_faults_arr* stFaultsArr,//! array of strings
   uint *idx                  //! [in/out] index
){
   ___Dassert(stFaultsArr, "'stFaultsArr' parameter invalid");
   ___Dassert(idx, "'idx' parameter invalid");

   size_t i = *idx;
   if (i >= stFaultsArr->ns) { return NULL; }    //: out-of-bounds check
   *idx = *idx + 1;              //: increase index by 1
   return stFaultsArr->buf[i];   //: return fault at old index in array
}

void faultsarr_print (
   faultcath_en cath,      //! print only this cathegory, 0 to print all
   FILE* fs,               //! filestream
   const st_faults_arr* stFaultsArr,//! array of faults
   const char* prec_line,  //! line preceding each fault
   const char* pref,       //! string preceding each fault in same line
   const char* post,       //! string succeding each fault in same line
   const char* suc_line    //! line succeding each fault
){
   ___Dassert(fs, "'fs' parameter invalid");
   ___Dassert(stFaultsArr, "'stFaultsArr' parameter invalid");

   st_fault* stFault = NULL;
   uint i = 0;
   while ((stFault = faultsarr_it(stFaultsArr, &i)) != NULL) {
      if (cath != Faultcath_none && stFault->cath != cath) { continue; }
      if (prec_line != NULL) { fprintf(fs, "%s\n", prec_line); }
      fault_print(fs, stFault, pref, post);
      if (suc_line != NULL) { fprintf(fs, "\n%s", suc_line); }
   }
}
