/*!F::Messages logging - main implementation file.
 *
 *  ::Purpose:
 *    Log messages in consistent manner.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - log messages
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    Created:            2013-12-16
 *    (c) Copyright:      "Machel" 2013
 *    License:            GNU Public License
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"

/* libc */
#include <stdarg.h>        // va_*
#include <stdio.h>         // printf(), fflush(), FILE*

/* current module */
#include "logs.h"




/*!:: First, fixed part of message is composed from established elements: prefix,
 *    program name, file name, line number and function name. Second part is
 *    customized by user in printf-like manner.
 *    Message is printed to `fs` stream, stream is flushed.
 * :: Intended for logging, debugging and errors printing, wenever message needs
 *    to carry info about file, line and function where event occured.
 *
 * ::Preconditions:
 *  - all parameters except variadic list are required and can't be NULL
 *
 * ::Special preconditions:
 *  - 'MHLC_BASE_PR_NAME' macro should be defined or "UNKNOWN" string will be
 *    used as program name
 *
 * ::Errors: NONE.
 *    printf-like functions return -1 on output error, but it is ignored.
 */
void
log_string (
   FILE* fs,         //! message goes to this stream
   const char* pref, //! prefix, begining of message
   const char* file, //! file where message originated from
   uint line,        //! line in file
   const char* func, //! function where message originated from
   const char* fmt,  //! format string conforming to printf specification
   ...               //! data to print
){
   assert(pref && "'pref' parameter invalid");
   assert(file && "'file' parameter invalid");
   assert(line > 0 && "'line' parameter invalid");
   assert(func && "'func' parameter invalid");
   assert(fmt && "'fmt' parameter invalid");

#ifndef MHLC_BASE_PR_NAME
#  define MHLC_BASE_PR_NAME "UNKNOWN"
#endif

   va_list ap;
   va_start(ap, fmt);

   /*: print fixed part */
   fprintf(fs, "%s <%s> %s(): %s:%d:: ",
               pref, MHLC_BASE_PR_NAME, func, file, line);
   /*: print custom part */
   vfprintf(fs, fmt, ap);

   fprintf(fs,"\n");
   fflush(fs);

   va_end(ap);
}
