/*!F::Array of pointers - main implementation file.
 *
 *  ::Purpose:
 *    Manage arrays of pointers to arbitrary objects.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *
 *  ::Functionality:
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-05-04
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"

/* libc */
#include <stdlib.h>     // abort()
#include <stdio.h>      // printf(), fflush()
#include <string.h>     // memset()

/* current module */
#include "arrptrs.h"
#include "arrptrs_internal.h.inc"




#define ARRPTRS_INFLATE_LEAP 128




void
arrptrs_ini (
   arr_ptrs_t* arrptrs,
   arr_ptrs_el_destr_ft pf_el_destr
){
   ___Dassert(arrptrs, "'arrptrs' parameter invalid");

   arrptrs_ini2(arrptrs, 0, pf_el_destr);
      HandleErr(handle_err);
   return;

  handle_err:
   return;
}

void
arrptrs_ini2 (
   arr_ptrs_t* arrptrs,
   uint nels_init,
   arr_ptrs_el_destr_ft pf_el_destr
){
   ___Dassert(arrptrs, "'arrptrs' parameter invalid");

   if (nels_init == 0) {
      ass_(arrptrs, NULL, 0, 0, pf_el_destr);
      return;
   }

   void* buf = prep_(arrptrs, nels_init);
      ErrorTrigIf(buf == NULL, handle_err);
   arrptrs->pf_el_destr = pf_el_destr;

   rubout_(buf, 0, nels_init);   //: zero allocated pointers

   return;

  handle_err:
   free(buf);
   return;
}

void
arrptrs_ext (
   arr_ptrs_t* arrptrs,
   uint nels_add
){
   ___Dassert(arrptrs, "'arrptrs' parameter invalid");

   void** res = infl_(arrptrs, nels_add);
      ErrorTrigIf(res == NULL, HANDLE_ERR);

   return;

  HANDLE_ERR:
     return;
}

void*
arrptrs_fin (
   arr_ptrs_t* arrptrs,
   void* destr_data     //! data passed to function destroying elements
){
   ___Dassert(arrptrs, "'arrptrs' parameter invalid");

   void* des_res = NULL;

   if (arrptrs->pf_el_destr != NULL) {
      void* ptr = NULL;
      for (uint ii = 0; (ptr = arrptrs_it(arrptrs, &ii)) != NULL; ) {
         des_res = arrptrs->pf_el_destr(ptr, destr_data);
      }
   }

   clean_(arrptrs);
   return des_res;
}

void*
arrptrs_cpy (
   arr_ptrs_t* arrptrs,
   uint idx,
   void* ptr
){
   ___Dassert(arrptrs, "'arrptrs' parameter invalid");

   arrptrs->buf[idx] = ptr;
   if (arrptrs->ns <= idx) { arrptrs->ns = idx + 1; }
   return ptr;
}

void*
arrptrs_cpyr (
   arr_ptrs_t* arrptrs,
   uint idx,
   uint num,
   void* ptr
){
   ___Dassert(arrptrs, "'arrptrs' parameter invalid");
   for (size_t i = idx; i < num; ++i) {
      arrptrs_cpy(arrptrs, i, ptr);
   }
   //TODO:if (arrptrs->ns <= idx+num) { arrptrs->ns = idx + 1; }
   return ptr;
}

void*
arrptrs_app (
   arr_ptrs_t* arrptrs,
   void* ptr
){
   ___Dassert(arrptrs, "'arrptrs' parameter invalid");

   /*: Extend if needed: */
   if (arrptrs->ns == arrptrs->na) {
      void** res = infl_(arrptrs, ARRPTRS_INFLATE_LEAP);
         ErrorTrigIf(res == NULL, handle_err);
   }

   /*: Write given pointer after last set element. */
   arrptrs->buf[arrptrs->ns] = ptr;
   ++arrptrs->ns;

   /*: Finish: */
   return ptr;

  handle_err:
   return NULL;
}

void*
arrptrs_el_ret (
   arr_ptrs_t* arrptrs,
   uint idx
){
   ___Dassert(arrptrs, "'arrptrs' parameter invalid");
   if (idx >= arrptrs->ns) { return NULL; }  //: out-of-bounds check
   return ret_(arrptrs, idx);
}

void*                         //! pointer in buffer
arrptrs_it (
   const arr_ptrs_t* arrptrs, //! array of pinters
   uint *idx                   //! [in/out] index
){
   ___Dassert(arrptrs, "'arrptrs' parameter invalid");
   ___Dassert(idx, "'idx' parameter invalid");

   size_t i = *idx;
   if (i >= arrptrs->ns) { return NULL; }    //: out-of-bounds check
   *idx = *idx + 1;              //: increase index by 1
   return arrptrs->buf[i];       //: return pointer in buffer at old index
}
