/*!F::Memory management - main implementation file.
 *
 *  ::Purpose:
 *    Safe management of memory allocation. Reserving and freeing memory space
 *    in reliable, error aware and consistent way.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - allocate and reallocate memory
 *    - free memory, including setting pointer to NULL
 *
 *  ::Limitations:
 *    - not possible to allocate more then one area at once (ex. for several
 *       pointers)
 *    - not possible to free more then one area at once
 *    - not possible to preallocate one big area for later use
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2013-12-16
 *    (c) Copyright:    Maciej Helminiak 2013
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"

/* libc */
#include <stdlib.h>     // malloc() free()
#include <string.h>     // memset()

/* current module */
#include "memory.h"




/*!:: Return pointer to allocated heap area. Memory is not initialized.
 *
 * ::Preconditions:
 *  - `size` can't be 0
 *
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    ENOMEM      -- not enough memory available
 */
void*             //! return pointer to allocated memory or NULL
mem_alloc (
   size_t size    //! size of memory block in bytes
){
   ___Dassert(size > 0, "'size' parameter invalid");

   void* ptr_new = NULL;
   ptr_new = malloc(size);
      ErrorTrigIf(ptr_new == NULL, handle_err);

      ___Dassert(ptr_new, "'ptr_new' return assert failed");
   return ptr_new;

  handle_err:
   return NULL;
}

/*!:: Return pointer to allocated heap area. Memory is initialized to 0.
 *
 * ::Preconditions:
 *  - `size` can't be 0
 *
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    ENOMEM      -- not enough memory available
 */
void*             //! return pointer to allocated memory or NULL
mem_allocz (
   size_t size    //! size of memory block in bytes
){
   ___Dassert(size > 0, "'size' parameter invalid");

   void* ptr_new = NULL;
   ptr_new = calloc(1, size);
      ErrorTrigIf(ptr_new == NULL, handle_err);

      ___Dassert(ptr_new, "'ptr_new' return assert failed");
   return ptr_new;

  handle_err:
   return NULL;
}

/*!:: Return pointer to reallocated heap area. Added memory is not initialized.
 *    The contents of memory is not modified up to `new_size` bytes. Returned
 *    pointer may but doesn't have to be the same as pointer passed to function.
 *
 * ::Preconditions:
 *  - `ptr` must point to previously allocated memory area
 *  - `size` can't be 0
 *
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    ENOMEM      -- not enough memory available
 */
void*                //! return pointer to rallocated memory or NULL
mem_realloc (
   void* ptr,        //! pointer to memory area to be reallocated
   size_t new_size   //! new size of memory block in bytes
){
   ___Dassert(ptr, "'ptr' parameter invalid");
   ___Dassert(new_size > 0, "'new_size' parameter invalid");

   ptr = realloc(ptr, new_size);
      ErrorTrigIf(ptr == NULL, handle_err);

      ___Dassert(ptr, "'ptr' return assert failed");
   return ptr;

  handle_err:
   return NULL;
}

/*!:: Return pointer to reallocated heap area.
 *    The contents of memory is not modified up to `new_size` bytes. Returned
 *    pointer may but doesn't have to be the same as pointer passed to function.
 *
 * ::Preconditions:
 *  - `ptr` must point to previously allocated memory area
 *  - `size` can't be 0
 *
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
void*                //! return pointer to rallocated memory or NULL
mem_reallocz (
   void* ptr,        //! pointer to memory area to be reallocated
   size_t old_size,  //! old size of memory block in bytes
   size_t new_size   //! new size of memory block in bytes
){
   ___Dassert(ptr, "'ptr' parameter invalid");
   ___Dassert(new_size > 0, "'new_size' parameter invalid");

   /*: reallocate */
   ptr = realloc(ptr, new_size);
      ErrorTrigIf(ptr == NULL, handle_err);

   /*: initialize added memory to 0 */
   memset((char*)ptr + old_size, 0, new_size - old_size);

      ___Dassert(ptr, "'ptr' return assert failed");
   return ptr;

  handle_err:
   return NULL;
}
