/*!F::C-like strings management - main implementation file.
 *
 *  ::Purpose:
 *    Functions to work with \0 terminated c-like strings.
 *  ..Main improvement over standard library functions is implicit allocation of
 *    memory space for strings and care about always terminating with \0 char.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - basic activities: create c-string in various manners, destroy
 *    - manipulate existing c-strings: prepend, append
 *
 *  ::Limitations:
 *    - much more functions maniuplating cstrings in many ways can be implemented
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2013-12-16
 *    (c) Copyright:    Maciej Helminiak 2013
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"

/* libc */
#include <string.h>     // str*() strn*() mem*()
#include <stdarg.h>     // va_*, vsnprintf()

/* current module */
#include "cstrings.h"
#include "cstrings_internal.h.inc"




/* Prepare and cleanup. */
/* -------------------------------------------------------------------------- */


/*!:: Allocate sufficient memory and copy characters.
 * ::Preconditions:
 *    - `cs` must be '\0' terminated c-string
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! created c-string
cs_cre (
   const char* cs    //! c-string used for creation
){
   ___Dassert(cs, "'cs' parameter invalid");

   const size_t nc = strlen(cs);
   char* cs_new = cs_cre2(cs, nc);
      HandleErr(handle_err);

      ___Dassert(cs_new, "'cs_new' return assert failed");
   return cs_new;

  handle_err:
   free(cs_new);
   return NULL;
}

/*!:: Allocate sufficient memory and copy characters there. Use supplied lengths
 *    instead computing them self.
 * ::Preconditions:
 *    - `cs` must be '\0' terminated c-string
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! created c-string or NULL
cs_cre2 (
   const char* cs,   //! c-string used for creation
   size_t nc         //! length
){
   ___Dassert(cs, "'cs' parameter invalid");

      OverflowIntIf(nc + 1 < nc, handle_err);
   const size_t nb = nc + 1;
   char* cs_new = constr_(nb);
      ErrorTrigIf(cs_new == NULL, handle_err);
   overw_(cs_new, 0, nc, cs);
   cs_new[nc] = '\0';

      ___Dassert(cs_new, "'cs_new' return assert failed");
   return cs_new;

  handle_err:
   free(cs_new);
   return NULL;
}

/*!:: Allocate sufficient memory block and copy there c-string which is
 *    formatted in printf-like manner.
 * ::Preconditions:
 *    - if `fmt` is NULL no actions are taken
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 *    . vsnprintf() -- caused by this libc function
 */
char*                //! created cstring or NULL
cs_cref (
   const char* fmt,  //! format string (like printf)
   ...               //! data for format string (like printf)
){
   /*: variadic transmission: call helper variadic function */
   va_list ap;
   va_start(ap, fmt);
   char* ret = cs_vcref(fmt, ap);
   va_end(ap);
   return ret;
}

/*! ~~cs_cref()
 */
char*                //! pointer to new cstring or NULL
cs_vcref (
   const char* fmt,  //! format string (like printf)
   va_list vlist     //! data for format string (like printf)
){
   if (fmt == NULL) { goto handle_noaction; }

   va_list ap;
   char* cstr_new = NULL;

   int nc_to_print = 0;
   size_t nb = 0;

   /*: Compute length of string */
   va_copy(ap, vlist);
   nc_to_print = vsnprintf(NULL, 0, fmt, ap);
   va_end(ap);
      ErrorTrigIf(nc_to_print == -1, handle_err);
   nb = (size_t)nc_to_print + 1;

   /*: Allocate memory */
   cstr_new = constr_(nb);
      ErrorTrigIf(cstr_new == NULL, handle_err);

   /*: Compose c-string */
   va_copy(ap, vlist);
   nc_to_print = vsnprintf(cstr_new, nb, fmt, ap);
   va_end(ap);
      ErrorTrigIf(nc_to_print == -1, handle_err);

   /*: Finish */
      ___Dassert(cstr_new, "'cstr_new' return assert failed");
   return cstr_new;

  handle_noaction:
   return NULL;

  handle_err:
   destr_(cstr_new);
   return NULL;
}

/*!:: Release memory pointed by `pcs`, set pointer to NULL.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
 */
void
cs_des (
   char** pcs     //! [in/out] pointer to c-string to destroy
){
   ___Dassert(pcs, "'pcs' parameter invalid");
   destr_(*pcs);
   *pcs = NULL;
   return;
}




/* Manipulate. */
/* -------------------------------------------------------------------------- */


/*!:: Allocate suffcient memory and copy there both c-strings.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                   //! new c-string
cs_prep (
   char** pcs_d,        //! [out] pointer to destination c-string
   const char* cs_s     //! c-string being prepended
){
   ___Dassert(pcs_d, "'pcs_d' parameter invalid");
   ___Dassert(cs_s, "'cs_s' parameter invalid");

   /*: Compute lengths. */
   char* cs_d = *pcs_d;
   size_t nc_d = strlen(cs_d);
   size_t nc_s = strlen(cs_s);

   /*: Use other function to prepend. */
   cs_prep2(pcs_d, nc_d, cs_s, nc_s);
      HandleErr(handle_err);

   /*: Finish. */
      ___Dassert(*pcs_d, "'*pcs_d' return assert failed");
   return *pcs_d;

  handle_err:
   return NULL;
}

/*!:: Allocate suffcient memory and copy there both c-strings. Use supplied
 *    lengths instead computing them internally.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                   //! new c-string
cs_prep2 (
   char** pcs_d,        //! [out] pointer to destination c-string
   size_t nc_d,         //! length of destination
   const char* cs_s,    //! c-string being prepended
   size_t nc_s          //! length of source
){
   ___Dassert(pcs_d, "'pcs_d' parameter invalid");
   ___Dassert(cs_s, "'cs_s' parameter invalid");

   char* cs_new = NULL;
   char* cs_d = *pcs_d; // helper variable

   /*: Allocate c-string according to given lengths. */
      OverflowIntIf(nc_d + 1 < nc_d, handle_err);
      OverflowIntIf(nc_d + nc_s + 1 < nc_d + 1, handle_err);
   cs_new = constr_(nc_d + nc_s + 1);
      ErrorTrigIf(cs_new == NULL, handle_err);

   /*: Copy c-strings there in proper order. */
      OverflowIntIf(nc_s + 1 < nc_s, handle_err);
      OverflowIntIf(nc_d + 1 < nc_d, handle_err);
   overw_(cs_new, 0, nc_s+1, cs_s);
   overw_(cs_new, nc_s, nc_d+1, cs_d);

   /*: Release old destination c-string and direct pointer to new one. */
   destr_(*pcs_d);
   *pcs_d = cs_new;

   /*: Finish. */
      ___Dassert(*pcs_d, "'*pcs_d' return assert failed");
   return *pcs_d;

  handle_err:
   free(cs_new);
   return NULL;
}

/*!:: Allocate suffcient memory and copy there both c-strings.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                   //! new c-string
cs_app (
   char** pcs_d,        //! [out] pointer to destination c-string
   const char* cs_s     //! c-string being appended
){
   ___Dassert(pcs_d, "'pcs_d' parameter invalid");
   ___Dassert(cs_s, "'cs_s' parameter invalid");

   char* cs_d = *pcs_d; // helper variable

   /*: Compute lengths. */
   size_t nc_d = strlen(cs_d);
   size_t nc_s = strlen(cs_s);

   /*: Use other function to prepend. */
   cs_app2(pcs_d, nc_d, cs_s, nc_s);
      HandleErr(handle_err);

   /*: Finish. */
      ___Dassert(*pcs_d, "'*pcs_d' return assert failed");
   return *pcs_d;

  handle_err:
   return NULL;
}

/*!:: Allocate suffcient memory and copy there both c-strings. Use supplied
 *    lengths instead computing them internally.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                   //! new c-string
cs_app2 (
   char** pcs_d,        //! [out] pointer to destination c-string
   size_t nc_d,         //! length of destination
   const char* cs_s,    //! c-string being appended
   size_t nc_s          //! length of source
){
   ___Dassert(pcs_d, "'pcs_d' parameter invalid");
   ___Dassert(cs_s, "'cs_s' parameter invalid");

   char* cs_new = NULL;
   char* cs_d = *pcs_d; // helper variable

   /*: Allocate c-string according to given lengths. */
      OverflowIntIf(nc_d + 1 < nc_d, handle_err);
      OverflowIntIf(nc_d + nc_s + 1 < nc_d + 1, handle_err);
   cs_new = constr_(nc_d + nc_s + 1);
      ErrorTrigIf(cs_new == NULL, handle_err);

   /*: Copy c-strings there in proper order. */
      OverflowIntIf(nc_s + 1 < nc_s, handle_err);
      OverflowIntIf(nc_d + 1 < nc_d, handle_err);
   overw_(cs_new, 0, nc_d+1, cs_d);
   overw_(cs_new, nc_d, nc_s+1, cs_s);

   /*: Release old destination c-string and direct pointer to new one. */
   destr_(*pcs_d);
   *pcs_d = cs_new;

   /*: Finish. */
      ___Dassert(*pcs_d, "'*pcs_d' return assert failed");
   return *pcs_d;

  handle_err:
   free(cs_new);
   return NULL;
}

/*!:: Allocate sufficient memory block and copy there both c-string. Appended
 *    one is formatted in printf-like manner.
 * ::Preconditions:
 *    - if `fmt` is NULL no actions are taken
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 *    . vsnprintf() -- caused by this libc function
 */
void
cs_appf (
   char** cs_dest,   //! [out] destination c-string
   const char* fmt,  //! format string (like printf)
   ...               //! data for format string (like printf)
){
   /*: variadic transmission: call helper variadic function */
   va_list ap;
   va_start(ap, fmt);
   cs_vappf(cs_dest, fmt, ap);
   va_end(ap);
   return;
}

/*! ~~cs_appf()
 */
void
cs_vappf (
   char** cs_dest,   //! [out] destination c-string
   const char* fmt,  //! format string (like printf)
   va_list vlist     //! data for format string (like printf)
){
   ___Dassert(cs_dest, "'cs_dest' parameter invalid");

   /*: Create temporary c-string using another function. */
   char* cs_toapp = cs_vcref(fmt, vlist);
      HandleErr(handle_err);

   /*: Append temporary c-string. */
   cs_app(cs_dest, cs_toapp);
      HandleErr(handle_err);

   /*: Finish. */
   cs_des(&cs_toapp);   //: Destroy temporary c-string
   return;

  handle_err:
   cs_des(&cs_toapp);   //: Destroy temporary c-string
   return;
}
