/*!F::String structure management - main implementation file.
 *
 *  ::Purpose:
 *    Wrap cstrings into managable string structure.
 *    Introduces str_t structure, wrapper for cstrings and functions to
 *    manipulate it.
 *  ..Main improvement over standard cstrings is implicit memory
 *    allocation and realase, attaching size info to string, also
 *    safe manipulation and implicit care about terminating \0.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - basic activities: creation in various manners, destroy
 *    - manipulate existing strings: prepend, append, replace
 *    - examination: is empty? is zero?
 *    - print string
 *
 *  ::Limitations:
 *    - more functions maniuplating cstrings in many ways can be implemented
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    Created:            2013-12-16
 *    (c) Copyright:      "Machel" 2013
 *    License:            GNU Public License
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"
#include "cstrings.h"

/* libc */
#include <errno.h>      // errno
#include <string.h>     // str*() strn*() mem*()
#include <stdarg.h>     // va_*, vsnprintf()
//#include <stdio.h>         // printf()

/* current module */
#include "strings.h"
#include "strings_internal.h.inc"




/* Prepare and cleanup. */
/* -------------------------------------------------------------------------- */


/*!:: Allocate memory for string structure, allocate buffer inside and copy
 *    characters. Set structure data.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
str_t*               //! pointer to created string
str_cre (
   const char* cs    //! c-string used for creation
){
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Allocate str_t struct: */
   str_t* str_new = str_cre2(cs, strlen(cs));
      HandleErr(handle_err);

   /* Finish: */
      ___Dassert(str_new, "'str_new' return assert failed");
   return str_new;

  handle_err:
   return NULL;
}

/*!:: Allocate memory for string structure, allocate buffer inside and copy
 *    characters. Set structure data.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
str_t*               //! pointer to created string
str_cres (
   const str_t* str    //! string used for creation
){
   ___Dassert(str, "'str' parameter invalid");

   /*: Allocate str_t struct: */
   str_t* str_new = str_cre2(str->buf, str->nc);
      HandleErr(handle_err);

   /* Finish: */
      ___Dassert(str_new, "'str_new' return assert failed");
   return str_new;

  handle_err:
   return NULL;
}

/*!:: Allocate memory for string structure, allocate buffer inside and copy
 *    characters. Set structure data.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
str_t*               //! pointer to created string
str_cre2 (
   const char* cs,   //! c-string used for creation
   size_t nc         //! number of characters
){
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Allocate str_t struct: */
   str_t* str_new = constr_();
      ErrorTrigIf(str_new == NULL, handle_err);

   /*: Allocate c-string buffer inside: */
      OverflowIntIf(nc + 1 < 1, handle_err);
   size_t nb = nc + 1;
   char* res = prep_(str_new, nb);
      ErrorTrigIf(res == NULL, handle_err);

   /*: Copy `cs` into buffer, set structure data: */
   overw_(str_new, 0, nb, cs);
   ass_ncs_(str_new, nb-1, 0);

   /* Finish: */
      ___Dassert(str_new, "'str_new' return assert failed");
   return str_new;

  handle_err:
   if (str_new != NULL) { clean_(str_new); }
   destr_(str_new);
   return NULL;
}


/*!:: Release c-string buffer inside struct and struct itself, set pointer to
 *    NULL.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
 */
void
str_des (
   str_t** pstr      //! [in/out] pointer to string struct
){
   ___Dassert(pstr, "'pstr' parameter invalid");
   clean_(*pstr);
   destr_(*pstr);
   *pstr = NULL;
   return;
}

/*!:: Allocate buffer inside and copy characters. Set structure data.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! pointer to buffer inside string struct
str_ini (
   str_t* str,       //! [out] string struct to initialize
   const char* cs    //! c-string used for creation
){
   ___Dassert(str, "'str' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Use other function to initialize. */
   str_ini2(str, cs, strlen(cs));
      HandleErr(handle_err);

   /*: Finish: */
      ___Dassert(str->buf, "'str->buf' return assert failed");
   return str->buf;

  handle_err:
   return NULL;
}

/*!:: Allocate buffer inside string structure and copy characters. Set structure
 *    data approrietely. Use supplied length instead computing it self.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! pointer to buffer inside string struct
str_ini2 (
   str_t* str,       //! [out] string struct to initialize
   const char* cs,   //! c-string used for creation
   size_t nc         //! length
){
   ___Dassert(str, "'str' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Allocate c-string buffer inside. */
      OverflowIntIf(nc + 1 < 1, handle_err);
   uint nb = nc + 1;
   char* res = prep_(str, nb);
      ErrorTrigIf(res == NULL, handle_err);

   /*: Copy `cs` into buffer, set structure data: */
   overw_(str, 0, nc, cs);
   str->buf[nb-1] = '\0';
   ass_ncs_(str, nb-1, 0);

   /*: Finish: */
      ___Dassert(res, "'res' return assert failed");
   return res;

  handle_err:
   return NULL;
}

/*!:: Allocate buffer inside string structure and copy characters. Set structure
 *    data approrietely. Use supplied length instead computing it self.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*
str_inif (
   str_t* str,
   const char* fmt,
   ...
){
   /*: variadic transmission: call helper variadic function */
   va_list ap;
   va_start(ap, fmt);
   char* cs = cs_vcref(fmt, ap);
      HandleErr(handle_err);
   va_end(ap);
   str_fill(str, cs);
   return cs;

  handle_err:
   return NULL;
}

/*!:: Allocate buffer inside and copy characters from string structure buffer.
 *    Set structure data.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! pointer to buffer inside string struct
str_inis (
   str_t* str_dest,  //! [out] string struct to initialize
   const str_t* str  //! string used for creation
){
   ___Dassert(str_dest, "'str_dest' parameter invalid");
   ___Dassert(str, "'str' parameter invalid");

   /*: Use other function to initialize. */
   str_ini2(str_dest, str->buf, str->nc);
      HandleErr(handle_err);

   /*: Finish: */
      ___Dassert(str_dest->buf, "'str_dest->buf' return assert failed");
   return str_dest->buf;

  handle_err:
   return NULL;
}

/*!:: Release buffer inside string struct, zero struct data.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
 */
void
str_fin (
   str_t* str     //! string struct
){
   ___Dassert(str, "'str' parameter invalid");
   clean_(str);
}

/*!:: Point buffer inside string struct to given c-string, compute properties.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
 */
void
str_fill (
   str_t* str_dest,  //! [out] string struct to fill
   char* cs          //! c-string used to fill in string
){
   ___Dassert(str_dest, "'str_dest' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");
   size_t nc = strlen(cs);
   ass_(str_dest, cs, nc+1, nc, 0);
   return;
}

/*!:: Point buffer inside string struct to given c-string. Use supplied
 *    properties instead computing them self.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
 */
void
str_fill2 (
   str_t* str_dest,  //! [out] string struct to fill
   char* cs,         //! c-string used to fill in string
   size_t nc,        //! length
   size_t mnc        //! multibyte length
){
   ___Dassert(str_dest, "'str_dest' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");
   ass_(str_dest, cs, nc+1, nc, mnc);
   return;
}

/*!:: Point buffer inside string struct to given string struct's buffer. Use
 *    properties ready inside string.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
 */
void
str_fills (
   str_t* str_dest,  //! [out] string struct to fill
   str_t* str_src    //! string used to fill in string
){
   ___Dassert(str_dest, "'str_dest' parameter invalid");
   ___Dassert(str_src, "'str_src' parameter invalid");
   ass_(str_dest, str_src->buf, str_src->nb, str_src->nc, str_src->mnc);
   return;
}

/*!:: Zero all data inside string struct.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
 */
void
str_empty (
   str_t* str     //! [out] string struct to fill
){
   ___Dassert(str, "'str' parameter invalid");
   zero_(str);
   return;
}




/* Manipulate. */
/* -------------------------------------------------------------------------- */


/*!:: Replace c-string inside string struct with given c-string. Buffer is
 *    resized and properties are set appropriately.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! pointer to buffer inside string struct
str_repl (
   str_t* str,       //! [out] string struct
   const char* cs    //! c-string used to replace string
){
   ___Dassert(str, "'str' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Resize buffer. */
   size_t nc = strlen(cs);
      OverflowIntIf(nc + 1 < 1, handle_err);
   char* res = levl_(str, nc+1);
      ErrorTrigIf(res == NULL, handle_err);

   /*: Copy c-string and set data in struct. */
   overw_(str, 0, nc+1, cs);
   ass_ncs_(str, nc, 0);

   /*: Finish. */
      ___Dassert(res, "'res' return assert failed");
   return res;

  handle_err:
   return NULL;
}

/*!:: Replace c-string inside string struct with given c-string. Buffer is
 *    resized and properties are set appropriately. Use supplied properties
 *    instead computing them self.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! pointer to buffer inside string struct
str_repl2 (
   str_t* str,       //! [out] string struct
   const char* cs,   //! c-string used to replace string
   size_t nc,        //! length
   size_t mnc        //! multibyte length
){
   ___Dassert(str, "'str' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Resize buffer. */
      OverflowIntIf(nc + 1 < 1, handle_err);
   char* res = levl_(str, nc+1);
      ErrorTrigIf(res == NULL, handle_err);

   /*: Copy c-string and set data in struct. */
   overw_(str, 0, nc+1, cs);
   ass_ncs_(str, nc, mnc);
   str->buf[nc] = '\0';   //nc my be less then strlen(cs) thus \0 may be not copied

   /*: Finish. */
      ___Dassert(res, "'res' return assert failed");
   return res;

  handle_err:
   return NULL;
}

/*!:: Replace c-string inside string struct with given string. Buffer is
 *    resized and properties are set appropriately. Use properties already set
 *    in string.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! pointer to buffer inside string struct
str_repls (
   str_t* str_dest,  //! [out] string struct
   const str_t* str  //! string used for replacing
){
   ___Dassert(str_dest, "'str_dest' parameter invalid");
   ___Dassert(str, "'str' parameter invalid");

   /*: Resize buffer. */
      OverflowIntIf(str->nc + 1 < 1, handle_err);
   char* res = levl_(str_dest, str->nc+1);
      ErrorTrigIf(res == NULL, handle_err);

   /*: Copy c-string and set data in struct. */
   overw_(str_dest, 0, str->nc+1, str->buf);
   ass_ncs_(str_dest, str->nc, str->mnc);

   /*: Finish. */
      ___Dassert(res, "'res' return assert failed");
   return res;

  handle_err:
   return NULL;
}

/*!:: Allocate suffcient buffer inside string struct and copy there c-string
 *    and old buffer. Set struct properties approprietely.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! pointer to buffer inside string struct
str_prep (
   str_t* str,       //! [out] string struct
   const char* cs    //! c-string to prepend
){
   ___Dassert(str, "'str' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   size_t nc_cs = strlen(cs);
   cs_prep2 (&str->buf, str->nc, cs, nc_cs);
      HandleErr(handle_err);
   //TODO: if mnc!= count new mnc
      OverflowIntIf(str->nb + nc_cs < str->nb, handle_err);
      OverflowIntIf(str->nc + nc_cs < str->nc, handle_err);
   ass_(str, str->buf, str->nb + nc_cs, str->nc + nc_cs, 0);

   /*: Finish. */
      ___Dassert(str->buf, "'str->buf' return assert failed");
   return str->buf;

  handle_err:
   return NULL;
}

/*!:: Allocate suffcient buffer inside string struct and copy there old buffer
 *    and c-string. Set struct properties approprietely.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                //! pointer to buffer inside string struct
str_app (
   str_t* str,       //! [out] string struct
   const char* cs    //! c-string to append
){
   ___Dassert(str, "'str' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   size_t nc_cs = strlen(cs);
   cs_app2 (&str->buf, str->nc, cs, nc_cs);
      HandleErr(handle_err);
   //TODO: if mnc!= count new mnc
      OverflowIntIf(str->nb + nc_cs < str->nb, handle_err);
      OverflowIntIf(str->nc + nc_cs < str->nc, handle_err);
   ass_(str, str->buf, str->nb + nc_cs, str->nc + nc_cs, 0);

   /*: Finish. */
      ___Dassert(str->buf, "'str->buf' return assert failed");
   return str->buf;

  handle_err:
   return NULL;
}

/*!:: Allocate suffcient buffer inside string struct and copy there old buffer
 *    and given string. Set struct properties approprietely.
 * ::Preconditions:
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                   //! pointer to buffer inside string struct
str_apps (
   str_t* str_dest,     //! [out] string struct
   const str_t* str_src //! string to append
){
   ___Dassert(str_dest, "'str_dest' parameter invalid");
   ___Dassert(str_src, "'str_src' parameter invalid");

   cs_app2 (&str_dest->buf, str_dest->nc, str_src->buf, str_src->nc);
      HandleErr(handle_err);
   //TODO: if mnc!= count new mnc
      OverflowIntIf(str_dest->nb + str_src->nc < str_dest->nb, handle_err);
      OverflowIntIf(str_dest->nc + str_src->nc < str_dest->nc, handle_err);
   ass_(str_dest, str_dest->buf, str_dest->nb + str_src->nc, str_dest->nc + str_src->nc, 0);

   /*: Finish. */
      ___Dassert(str_dest->buf, "'str_dest->buf' return assert failed");
   return str_dest->buf;

  handle_err:
   return NULL;
}




/* Examine. */
/* -------------------------------------------------------------------------- */


/*!:: Return buffer.
 */
char*
str_ret_buf (
   str_t* str
){
   ___Dassert(str, "'str' parameter invalid");
   return str->buf;
}

/*!:: True if c-string buffer inside string struct is not NULL.
 */
bool
str_is_empty (
   str_t* str
){
   ___Dassert(str, "'str' parameter invalid");
   if (str->buf == NULL) { return true; }
   else { return false; }
}

/*!:: True if c-string buffer inside string struct is not NULL and other
 *    struct data is zero.
 */
bool
str_is_zero (
   str_t* str
){
   ___Dassert(str, "'str' parameter invalid");
   if (str->buf == NULL && str->nb == 0 && str->nc == 0) { return true; }
   else { return false; }
}




/* Compare. */
/* -------------------------------------------------------------------------- */
int
str_cmp (
   const str_t* str1,
   const str_t* str2
){
   ___Dassert(str1, "'str1' parameter invalid");
   ___Dassert(str2, "'str2' parameter invalid");

   if (str1->nc > str2->nc) {
      int res = memcmp(str1->buf, str2->buf, str2->nc);
      if (res == 0) { return 1; } else { return res; }
   } else if  (str1->nc < str2->nc) {
      int res = memcmp(str1->buf, str2->buf, str1->nc);
      if (res == 0) { return -1; } else { return res; }
   } else {
      return memcmp(str1->buf, str2->buf, str1->nc);
   }

   //return strcmp("aaa", "ab");
   //aaa aa    1
   //aab aa    1
   //aaa ab    -1
   //aaa bb    -1
   //bbb aa    1
   //
   //aa aaa    -1
   //aa aab    -1
   //aa bbb    -1
   //bb aaa    1
   //
   //bbb aaa   1
   //aaa bbb   -1
   //
   //
   // - s1<s2
   // + s1>s2
}




/* Print. */
/* -------------------------------------------------------------------------- */


/*!:: Print buffer.
 * ::Errors:
 *    On error nothing is printed.
 *    . fprintf()    -- errors caused by this function
 */
void
str_print (
   FILE* fs,
   str_t* str
){
   ___Dassert(str, "'str' parameter invalid");
   ___Dassert(fs, "'fs' parameter invalid");

   if (str->buf == NULL) { return; }
   fprintf(fs, "%s", str->buf);
   return;
}
