/*!F::Arrays of strings management - main implementation file.
 *
 *  ::Purpose:
 *    Provide easy way to manage arrays of strings.
 *    Introduces array of strings structure together with functions to
 *    manipulate it.
 *    Allows easy management of multiple strings in consistent and
 *    clear way with implicit memory allocation, release and care
 *    about terminating \0 character.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - basic activities: creation in various manners, destroy
 *    - manipulate existing arrays: copy, prepend, append, attach
 *    - other: split, compare
 *    - examination: return buffer, iterate
 *    - print strings in array
 *
 *  ::Limitations:
 *    - more functions maniuplating cstrings in many ways can be implemented
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2013-12-16
 *    (c) Copyright:    Maciej Helminiak 2013
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"
#include "cstrings.h"
#include "strings.h"

/* current project */
#include "common_mhlbase.h"

/* libc */
#include <errno.h>      // errno
#include <string.h>     // str*() strn*() mem*()
#include <stdarg.h>     // va_*, vsnprintf()
//#include <stdio.h>         // printf()

/* current module */
#include "strings_arr.h"
#include "strings_arr_internal.h.inc"




#define STRARR_INFLATE_LEAP 128




/* Prepare and cleanup. */
/* -------------------------------------------------------------------------- */


/*!:: Allocate memory for array structure, allocate inside buffer able to hold
 *    one element and copy given c-string to first element. Set structure data.
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
strarr_t*            //! pointer to created strings array
sarr_cre (
   const char* cs    //! cstring used
){
   ___Dassert(cs, "'cs' parameter invalid");

   strarr_t* sarr = NULL;

   /*: Allocate array struct. */
   sarr = constr_();
      ErrorTrigIf(sarr == NULL, handle_err);

   /*: Place c-string in frist array string. */
   sarr_ini(sarr, cs);
      HandleErr(handle_err);

   /*: Finish: */
      ___Dassert(sarr, "'sarr' return assert failed");
   return sarr;

  handle_err:
   free(sarr);
   return NULL;
}

/*!:: Allocate memory for array structure, allocate inside buffer able to hold
 *    specified number of elements. Set structure data.
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
strarr_t*         //! pointer to created strings array
sarr_cre2 (
   uint ne        //! number of elements in array
){
   strarr_t* sarr = NULL;

   /*: Allocate array struct. */
   sarr = constr_();
      ErrorTrigIf(sarr == NULL, handle_err);

   /*: Allocate buffer for `ne` strings */
   sarr_ini2(sarr, ne);
      HandleErr(handle_err);

   /*: Finish: */
      ___Dassert(sarr, "'sarr' return assert failed");
   return sarr;

  handle_err:
   free(sarr);
   return NULL;
}

/*!:: Allocate memory for array structure, allocate suffcient buffer and create
 *    element from each given c-string. Set structure data.
 * ::Preconditions:
 *  - last element of variadic arguments must be NULL
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
strarr_t*               //! pointer to created strings array or NULL
sarr_crev (
   const char* cs_fst,  //! first c-string
   ...                  //! variadic number of c-strings
){
   ___Dassert(cs_fst, "'cs_fst' parameter invalid");
   /*: variadic transmission: call helper variadic function */
   va_list ap;
   va_start(ap, cs_fst);
   strarr_t* ret = sarr_vcrev(cs_fst, ap);
   va_end(ap);
   return ret;
}

/*! ~~sarr_crev()
 */
strarr_t*            //! pointer to created strings array
sarr_vcrev (
   const char* cs_fst,     //! first c-string
   va_list vlist           //! variadic list of c-strings
){
   va_list ap;
   size_t n_css = 0;    //: number of cstrings to assign

   /*: Compute num of strings to assign */
   va_copy(ap, vlist);
   while (true)
   {
      const char* cs_ass = va_arg(ap, char*);
      if (cs_ass == NULL) { break; } //: NULL cstr ends va list
      ++n_css;
   };
   va_end(ap);

   /*: Create strings array struct */
   strarr_t* sarr_new = sarr_cre(cs_fst);
      HandleErr(handle_err);

   /*: No-NULL cstrings not present: skip copying */
   if (n_css == 0) { goto handle_succ; }

   /*: Append each cstring */
   va_copy(ap, vlist);
   for (size_t i = 0; i < n_css; ++i)
   {
      const char* cs_ass = va_arg(ap, char*);
      sarr_app(sarr_new, cs_ass);
         HandleErr(handle_err);
   };
   va_end(ap);

   /*: Finish: */
  handle_succ:
      ___Dassert(sarr_new, "'sarr_new' return assert failed");
   return sarr_new;         //: return pointer to new array

  handle_err:
   if (sarr_new != NULL) { sarr_des(&sarr_new); } //: release function resources
   return NULL;
}

/*!:: Allocate memory for array structure, allocate inside buffer able to hold
 *    one element and copy given c-string to first element. Set structure data.
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
strarr_t*            //! pointer to created strings array
sarr_cref (
   const char* fmt,  //! format string (printf-like)
   ...               //! data for `fmt` (printf-like)
){
   char* cs_new = NULL;
   strarr_t* sarr_new = NULL;

   /*: Cover case when format is NULL. */
   if (fmt == NULL) {
      sarr_new = sarr_cre2(0);      //: create array with 0 elements (buffer NULL)
         HandleErr(handle_err);
      goto handle_succ;
   }

   /*: Create c-string. */
   va_list ap;
   va_start(ap, fmt);
   cs_new =  cs_vcref (fmt, ap);
   va_end(ap);
      HandleErr(handle_err);

   /*: Create array initialized by this c-string. */
   sarr_new = sarr_cre(cs_new);
      HandleErr(handle_err);

   /*: Finish: */
  handle_succ:
   if (cs_new != NULL) { cs_des(&cs_new); }
      ___Dassert(sarr_new, "'sarr_new' return assert failed");
   return sarr_new;

  handle_err:
   if (cs_new != NULL) { cs_des(&cs_new); }
   return NULL;
}

/*!:: Release resources taken by every element in array, release buffer in
 *    array and array itself.
 * ::Errors:
 *    no Errors.
 */
void
sarr_des (
   strarr_t** psarr     //! pointer to array of strings
){
   ___Dassert(psarr, "'psarr' parameter invalid");
   ___Dassert(*psarr, "'*psarr' parameter invalid");
   strarr_t* sarr = *psarr;   // helper variable
   /*: Finalize internals of each string in array. */
   for (size_t i = 0; i < sarr->na; ++i) {
      str_fin(&sarr->buf[i]);
   }
   /*: Release buffer inside and array struct itself. */
   clean_(sarr);
   destr_(sarr);
   *psarr = NULL;
   return;
}

/*!:: Inside array allocate buffer able to hold one string and initialize it
 *    with given c-string.
 * ::Errors:
 *    nothing allocated, error set
 *    . ENOMEM      -- not enough memory available
 */
void
sarr_ini (
   strarr_t* sarr,   //! [out] array of strings
   const char* cs    //! c-string
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   str_t* buf = NULL;

   /*: Allocate buffer for one string. */
   buf = prep_(sarr, 1);
      ErrorTrigIf(buf == NULL, handle_err);
   //rubout_(buf, 0, 1);

   /*: Initialize string with c-string. */
   str_ini(&buf[0], cs);
      HandleErr(handle_err);
   sarr->ns = 1;

   /*: Finish: */
   return;

  handle_err:
   free(buf);
   return;
}

/*!:: Inside array allocate buffer able to hold specified number of elements.
 * ::Errors:
 *    nothing allocated, error set
 *    . ENOMEM      -- not enough memory available
 */
void
sarr_ini2 (
   strarr_t* sarr,      //! [out] array of strings
   uint ne              //! number of elements
){
   ___Dassert(sarr, "'sarr' parameter invalid");

   /*: Cover case when 0 elements requested. */
   if (ne == 0) {
      ass_(sarr, NULL, 0, 0);
      return;
   }

   /*: Allocate buffer in array for `ne` strings. */
   str_t* buf = prep_(sarr, ne);
      ErrorTrigIf(buf == NULL, handle_err);
   rubout_(buf, 0, ne); //: zero allocated strings
   sarr->ns = ne;

   /*: Finish */
   return;

  handle_err:
   free(buf);
   return;
}

/*!:: Inside array allocate buffer able to hold one string and initialize it
 *    with given string.
 * ::Errors:
 *    nothing allocated, error set
 *    . ENOMEM      -- not enough memory available
 */
void
sarr_inis (
   strarr_t* sarr,   //! [out] array of strings
   const str_t* str  //! string
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(str, "'str' parameter invalid");

   str_t* buf = NULL;

   /*: Allocate buffer for one string. */
   buf = prep_(sarr, 1);
      ErrorTrigIf(buf == NULL, handle_err);
   //rubout_(buf, 0, 1);

   /*: Initialize string in buffer with given string. */
   str_ini2(&buf[0], str->buf, str->nc);
      HandleErr(handle_err);
   sarr->ns = 1;

   /*: Finish: */
   return;

  handle_err:
   free(buf);
   return;
}

/*!:: Release resources taken by every element in array, release buffer in
 *    array and array itself.
 * ::Errors:
 *    no Errors.
 */
void
sarr_fin (
   strarr_t* sarr    //! array of strings
){
   ___Dassert(sarr, "'sarr' parameter invalid");

   /*: Finalize internals of each string in array. */
   for (size_t i = 0; i < sarr->na; ++i) {
      str_fin(&sarr->buf[i]);
   }
   /*: Release buffer inside array. */
   clean_(sarr);
   return;
}




/* Manipulate. */
/* -------------------------------------------------------------------------- */


/*!:: Resize element as needed and copy c-string over, old content is lost.
 *    Array will not be expanded if needed.
 * ::Errors:
 *    nothing copied, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 * TODO: index check
 */
char*                   //! pointer to buffer of copied string
sarr_cpy (
   strarr_t* sarr,      //! [out] array of strings
   uint idx,            //! index in array
   const char* cs       //! c-string
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   char* res = str_repl(&sarr->buf[idx], cs);
      HandleErr(handle_err);

   return res;

  handle_err:
   return NULL;
}

/*!:: Resize element as needed and copy string buffer over, old content is lost.
 * ::Errors:
 *    nothing copied, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 * TODO: index check
 */
char*                   //! pointer to buffer of copied string
sarr_cpys (
   strarr_t* sarr,      //! [out] array of strings
   uint idx,            //! index in array
   const str_t* str     //! string
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(str, "'str' parameter invalid");

   char* res = str_repls(&sarr->buf[idx], str);
      HandleErr(handle_err);

   return res;

  handle_err:
   return NULL;
}

/*!:: Add one element at array begining and copy given c-string there.
 * ::Errors:
 *    nothing prepended, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                   //! pointer to buffer of prepended string
sarr_prep (
   strarr_t* sarr,      //! [out] array of strings
   const char* cs       //! c-string
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Add empty element at begining. */
   str_t* ress = infl2_(sarr, 1);
      ErrorTrigIf(ress == NULL, handle_err);

   /*: Write given c-string to this element. */
   char* res = str_repl(&sarr->buf[0], cs);
      HandleErr(handle_err);
   ++sarr->ns;

   /*: Finish: */
   return res;

  handle_err:
   return NULL;
}

/*!:: Add one element at array end and copy given c-string there.
 * ::Errors:
 *    nothing appended, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                   //! pointer to buffer of appended string
sarr_app (
   strarr_t* sarr,      //! [out] array of strings
   const char* cs       //! c-string
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Extend if needed: */
   if (sarr->ns == sarr->na) {
      str_t* res = infl_(sarr, STRARR_INFLATE_LEAP);
         ErrorTrigIf(res == NULL, handle_err);
   }

   /*: Write given c-string to this element. */
   char* res = str_repl(&sarr->buf[sarr->ns], cs);
      HandleErr(handle_err);
   ++sarr->ns;

   /*: Finish: */
   return res;

  handle_err:
   return NULL;
}

/*!:: Add one element at array end and copy given c-string there. Use supplied
 *    c-string properties instead computing them self.
 * ::Errors:
 *    nothing appended, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                   //! pointer to buffer of appended string
sarr_app2 (
   strarr_t* sarr,      //! [out] array of strings
   const char* cs,      //! c-string
   size_t nc,           //! size
   size_t mnc           //! multibyte length
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Add empty element at end. */
   if (sarr->ns == sarr->na) {   //: only if no unset elements available
      str_t* res = infl_(sarr, STRARR_INFLATE_LEAP);
         ErrorTrigIf(res == NULL, handle_err);
   }

   /*: Write given c-string to this element using given properites. */
   char* res = str_repl2(&sarr->buf[sarr->ns], cs, nc, mnc);
      HandleErr(handle_err);
   ++sarr->ns;

   /*: Finish: */
   return res;

  handle_err:
   return NULL;
}

/*!:: Add one element at array end and copy given c-string there.
 * ::Errors:
 *    nothing appended, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                   //! pointer to buffer of appended string
sarr_apps (
   strarr_t* sarr,      //! [out] array of strings
   const str_t* str     //! string
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(str, "'str' parameter invalid");

   /*: Add empty element at end. */
   if (sarr->ns == sarr->na) {
      str_t* res = infl_(sarr, STRARR_INFLATE_LEAP);
         ErrorTrigIf(res == NULL, handle_err);
   }

   /*: Write given string to this element. */
   char* res = str_repl2(&sarr->buf[sarr->ns], str->buf, str->nc, str->mnc);
      HandleErr(handle_err);
   ++sarr->ns;

   /*: Finish: */
   return res;

  handle_err:
   return NULL;
}

/*!:: Add one element at array end and copy formatted c-string there.
 * ::Errors:
 *    nothing appended, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 *    . fprintf()   -- errors caused by this function
 */
char*
sarr_appf (
   strarr_t* stStrsArr, //! strings array to append to
   const char* fmt,     //! format string (printf-like) to be appended
   ...                  //! data for `fmt` (printf-like)
){
   ___Dassert(stStrsArr, "'stStrsArr' parameter invalid");

   char* cstr_new =  NULL;

   /*: Cover case when format is NULL. */
   if (fmt == NULL) { goto handle_noaction; }

   /*: Create c-string. */
   va_list ap;
   va_start(ap, fmt);
   cstr_new =  cs_vcref(fmt, ap);
   va_end(ap);
      HandleErr(handle_err);

   /*: Appned this c-string to array. */
   char* res = sarr_app(stStrsArr, cstr_new);
      HandleErr(handle_err);

   /*: Finish: */
   if (cstr_new != NULL) { cs_des(&cstr_new); }
   return res;

  handle_noaction:
   return NULL;

  handle_err:
   if (cstr_new != NULL) { cs_des(&cstr_new); }
   return NULL;
}

/*!:: Append each string in source array to destination array.
 * ::Errors:
 *    nothing appended, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
char*                      //! pointer to buffer of last appended string
sarr_appA (
   strarr_t* sarr_d,       //! [out] destination array
   const strarr_t* sarr_s  //! source array
){
   ___Dassert(sarr_d, "'sarr_d' parameter invalid");
   ___Dassert(sarr_s, "'sarr_s' parameter invalid");

   /*: Append each string in source array to destination array */
   for (size_t i = 0; i < sarr_s->ns; ++i) {
      sarr_apps(sarr_d, sarr_el_rets(sarr_s, i));
         HandleErr(handle_err);
   }

   /*: Finish: */
   return sarr_el_ret(sarr_d, sarr_d->ns-1);

  handle_err:
   return NULL;
}

/*!:: Append each string in source array to destination array adding prefix
 *    and postfix to each element.
 * ::Errors:
 *    nothing appended, NULL returned, error set
 *    . ENOMEM      -- not enough memory available
 */
void
sarr_appAp (
   strarr_t* sarr_d,       //! [out] destination array
   const char* pref,       //! prefix c-string
   const strarr_t* sarr_s, //! source array
   const char* post        //! postfix c-string
){
   ___Dassert(sarr_d, "'sarr_d' parameter invalid");
   ___Dassert(pref, "'pref' parameter invalid");
   ___Dassert(sarr_s, "'sarr_s' parameter invalid");
   ___Dassert(pref, "'post' parameter invalid");

   const char* cs = NULL;
   /*: Each string in source array: */
   for (uint i = 0; (cs = sarr_it(sarr_s, &i)) != NULL;) {
      sarr_appf(sarr_d, "%s%s%s", pref, cs, post); //: apend to destination array
         HandleErr(handle_err);                       //: with given prefix and postfix
   }

   /*: Finish: */
   return;

  handle_err:
   return;
}

/*!:: Add empty string element at array end and assign given c-string into its
 *    buffer.
 * ::Errors:
 *    nothing attached, error set
 *    . ENOMEM      -- not enough memory available
 */
void
sarr_att (
   strarr_t* sarr,   //! [out] array of strings
   char* cs          //! c-string
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Use other function to attach c-string. */
   size_t nc = strlen(cs);
   sarr_att2 (sarr, cs, nc+1, 0);
      HandleErr(handle_err);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/*!:: Add empty string element at array end and assign given c-string into its
 *    buffer. Use supplied properties instead computing them self.
 * ::Errors:
 *    nothing attached, error set
 *    . ENOMEM      -- not enough memory available
 */
void
sarr_att2 (
   strarr_t* sarr,   //! [out] array of strings
   char* cs,         //! c-string
   size_t nc,        //! length
   size_t mnc        //! multibyte length
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");

   /*: Add empty string at end. */
   if (sarr->ns == sarr->na) {   //: only if no unset elements available
      str_t* res = infl_(sarr, STRARR_INFLATE_LEAP);
         ErrorTrigIf(res == NULL, handle_err);
   }
   /*: Assign c-string and properties to string. */
   str_fill2(&sarr->buf[sarr->ns], cs, nc, mnc);
   ++sarr->ns;

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* Append each element of second array to each element of dest array and
   place result in dest array.
   ENOMEM
   errors not covered properly !!!
 */
void
sarr_multipleA (
   strarr_t* sarr_dest,    //! [in/out] dest array, place for result at the same time
   const strarr_t* sarr    //! second array
){
   ___Dassert(sarr_dest, "'sarr_dest' parameter invalid");
   ___Dassert(sarr, "'sarr' parameter invalid");

   strarr_t sarr_result = {0};   //: temporary result array
   const str_t* str_d = NULL;

   /*: For each el in dest array: */
   for (uint i = 0; (str_d = sarr_its(sarr_dest, &i)) != NULL;) {
      const str_t* str = NULL;
      /*: For each el in second array: */
      for (uint j = 0; (str = sarr_its(sarr, &j)) != NULL;) {
         /*: Create result string: */
         char* cstr_res = cs_cre2(str_d->buf, str_d->nc);
         size_t nc = strlen(cstr_res);
         cs_app2 (&cstr_res, nc, str->buf, str->nc);
         /*: Append result string to array: */
         sarr_att2(&sarr_result, cstr_res, nc + str->nc, 0);
            HandleErr(handle_err);  //common error check
      }
   }

   /*: Set members of dest array and finish temp result array. */
   str_t* tmp_buf = sarr_dest->buf;
   size_t tmp_ne = sarr_dest->na;
   size_t tmp_ns = sarr_dest->ns;
   sarr_dest->buf = sarr_result.buf;
   sarr_dest->na = sarr_result.na;
   sarr_dest->ns = sarr_result.ns;
   sarr_result.buf = tmp_buf;
   sarr_result.na = tmp_ne;
   sarr_result.ns = tmp_ns;
   sarr_fin(&sarr_result);

   /*: Finish: */
  handle_err:
   return;
}




/* Splitting. */
/* -------------------------------------------------------------------------- */


/*!:: Append each splitted word to array of strings.
 * ::Errors:
 *    words splitted before error occured may be already appended, error set
 *    . ENOMEM      -- not enough memory available
 */
void
sarr_split_line (
   strarr_t* sarr,      //! [out] array of strings
   const char* line,    //! line to split
   const char* delim    //! delimiter
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(line, "'line' parameter invalid");
   ___Dassert(delim, "'delim' parameter invalid");

   char* l = cs_cre(line);
      HandleErr(handle_err);

   size_t nc_delim = strlen(delim);
   const char* pl = l;
   char* pdelim = NULL;
   while ((pdelim = strstr(pl, delim)) != NULL)
   {
      *pdelim = '\0';
      sarr_app(sarr, pl);
         HandleErr(handle_err);
      pl = pdelim + nc_delim;
   }
   //append last part of line (after last delim or whole line if no delim found)
   sarr_app(sarr, pl);
      HandleErr(handle_err);

   cs_des(&l);
   return;

  handle_err:
   cs_des(&l);
   return;
}

/*! Same BUT ommit empty splits - when delim is streight after another. */
/*!:: Append each splitted word to array of strings. Ommit empty splits - if
 *    there is no character between delims no word is appended.
 * ::Errors:
 *    words splitted before error occured may be already appended, error set
 *    . ENOMEM      -- not enough memory available
 */
void
sarr_split_line2 (
   strarr_t* sarr,      //! [out] array of strings
   const char* line,    //! line to split
   const char* delim    //! delimiter
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(line, "'line' parameter invalid");
   ___Dassert(delim, "'delim' parameter invalid");

   char* l = cs_cre(line);
      HandleErr(handle_err);

   size_t nc_delim = strlen(delim);
   const char* pl = l;
   char* pdelim = NULL;
   while ((pdelim = strstr(pl, delim)) != NULL)
   {
      *pdelim = '\0';
      if (strlen(pl) != 0) {
         sarr_app(sarr, pl);
            HandleErr(handle_err);
      }
      pl = pdelim + nc_delim;
   }
   //append last part of line (after last delim or whole line if no delim found)
   sarr_app(sarr, pl);
      HandleErr(handle_err);

   cs_des(&l);
   return;

  handle_err:
   cs_des(&l);
   return;
}




/* Examine. */
/* -------------------------------------------------------------------------- */


/*!:: Return c-string in buffer of string element.
 * ::Errors:
 *    NULL returned if index if out of array bounds.
 */
char*                         //! c-string in buffer
sarr_el_ret (
   const strarr_t* sarr,      //! array of strings
   uint idx                   //! index
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   if (idx >= sarr->ns) { return NULL; }  //: out-of-bounds check
   return ret_(sarr, idx)->buf;
}

/*!:: Return pointer to 'str_t' element.
 * ::Errors:
 *    NULL returned if index if out of array bounds.
 */
str_t*                        //! pointer to string
sarr_el_rets (
   const strarr_t* sarr,      //! array of strings
   uint idx                   //! index
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   if (idx >= sarr->ns) { return NULL; }  //: out-of-bounds check
   return ret_(sarr, idx);
}

/*!:: Return c-string in buffer of string element at specified index and
 *    increment that index by one.
 *    NULL returned if index if out of array bounds.
 *    After function call `idx` holds next index to retrieve, that is it consist
 *    number of elements retrieved so far.
 * ::Errors:
 *    No errors.
 */
char*                         //! c-string in buffer
sarr_it (
   const strarr_t* sarr,      //! array of strings
   uint *idx                  //! [in/out] index
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(idx, "'idx' parameter invalid");

   size_t i = *idx;
   if (i >= sarr->ns) { return NULL; }    //: out-of-bounds check
   *idx = *idx + 1;              //: increase index by 1
   return sarr->buf[i].buf;      //: return string's buffer at old index in array
}

/*!:: Return pointer to 'str_t' element at specified index and increment that 
 *    index by one.
 *    NULL returned if index if out of array bounds.
 *    After function call `idx` holds next index to retrieve, that is it consist
 *    number of elements retrieved so far.
 * ::Errors:
 *    No errors.
 */
str_t*                        //! pointer to string
sarr_its (
   const strarr_t* sarr,      //! array of strings
   uint* idx                  //! [in/out] index
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(idx, "'idx' parameter invalid");

   size_t i = *idx;
   if (i >= sarr->ns) { return NULL; }    //: out-of-bounds check
   *idx = *idx + 1;              //: increase index by 1
   return &sarr->buf[i];         //: return pointer to string at old index in array
}




/* Compare. */
/* -------------------------------------------------------------------------- */


/*!:: Return 0 if c-sting inside array element at specified index is equal to
 *    given c-string. Return negative value if element is less then c-string
 *    or positive value if is greater.
 * ::Errors:
 *    No errors.
 */
int                        //! result of comparison
sarr_el_cmp (
   const strarr_t* sarr,   //! array of strings
   size_t idx,             //! index
   const char* cs          //! c-string
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");
   return strcmp(sarr->buf[idx].buf, cs);
}

ssize_t
sarr_cmp (
   const strarr_t* sarr,
   const char* cs
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(cs, "'cs' parameter invalid");
   const char* el_cs = NULL;
   for (uint i = 0; (el_cs = sarr_it(sarr, &i)) != NULL;) {
      if (strcmp(el_cs, cs) == 0) {
         return i;
      }
   }
   return -1;
}

ssize_t
sarr_cmps (
   const strarr_t* sarr,
   const str_t* str
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(str, "'str' parameter invalid");
   const str_t* el_str = NULL;
   for (uint ii = 0; (el_str = sarr_its(sarr, &ii)) != NULL;) {
      if (str_cmp(el_str, str) == 0) {
         return ii;
      }
   }
   return -1;
}

/*!:: Start from element 'el_num' and
 *    look for each element of array in c-string and stop at first hit. Place
 *    index of element found in `el_num` and return pointer to found substring
 *    in c-string. Return NULL if not found.
 * ::Errors:
 *    No errors.
 */
char*                      //! substring found in c-string or NULL if not found
sarr_str (
   size_t* el_num,         //! index of array element found
   const char* haystack,   //! c-string - haystack
   const strarr_t* needles //! array of strings - needles
){
   ___Dassert(el_num, "'el_num' parameter invalid");
   ___Dassert(haystack, "'haystack' parameter invalid");
   ___Dassert(needles, "'needles' parameter invalid");

   char* sub_ret = NULL;

   /*: Each string in array `needles`, starting from specified index,
    *    find it in haystack: */
   for (size_t i = *el_num; i < needles->ns; ++i)
   {
      if (sarr_el_ret(needles, i)[0] == '\0') { continue; } //: ignore empty string
      sub_ret = strstr(haystack, sarr_el_ret(needles, i));  //: find
      if (sub_ret != NULL)       //: if found:
      {
         *el_num = i;            //: indicate index of needle found
         return sub_ret;  //: pointer to substring in haystack
      }
   }
   return NULL;
}

/*!:: Look for c-string in each element of array and stop at first hit. Place
 *    index of element in which substring was found in `el_num` and return
 *    pointer to found substring in array element.
 * ::Errors:
 *    No errors.
 */
char*                         //! substring found in array element
sarr_rstr (
   size_t* el_num,            //! index of array element in which was found
   const strarr_t* haystacks, //! array of strings - haystack
   const char* needle         //! c-string - needle
){
   ___Dassert(el_num, "'el_num' parameter invalid");
   ___Dassert(needle, "'needle' parameter invalid");
   ___Dassert(haystacks, "'haystacks' parameter invalid");

   char* sub_ret = NULL;

   /*: Each string in array `haystack`, starting from 0,
    *    find needle in it: */
   for (size_t i = 0; i < haystacks->ns; ++i)
   {
      if (sarr_el_ret(haystacks, i)[0] == '\0') { continue; }  //: ignore empty string
      sub_ret = strstr(sarr_el_ret(haystacks, i), needle);     //: find
      if (sub_ret != NULL)       //: if found:
      {
         *el_num = i;            //: indicate in which haystack needle was found
         return sub_ret;  //: pointer to substring in haystack

      }
   }
   return NULL;
}




/* Print. */
/* -------------------------------------------------------------------------- */


/*!:: Print c-string inside element's buffer if it's not NULL one by one
 *    in same line.
 * ::Errors:
 *    On error printing is stopped.
 *    . fprintf()    -- errors caused by this function
 */
void
sarr_print (
   FILE* fs,               //! file stream
   const strarr_t* sarr    //! array of strings
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(fs, "'fs' parameter invalid");

   for (uint i = 0; i < sarr->ns; ++i) {
      if (sarr->buf[i].buf == NULL) { continue; }
      fprintf(fs, "%s", sarr->buf[i].buf);
   }
   return;
}

/*!:: Print c-string inside element's buffer if it's not NULL one by one
 *    in separate lines.
 * ::Errors:
 *    On error printing is stopped.
 *    . fprintf()    -- errors caused by this function
 */
void
sarr_printn (
   FILE* fs,               //! file stream
   const strarr_t* sarr    //! array of strings
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(fs, "'fs' parameter invalid");

   for (uint i = 0; i < sarr->ns; ++i) {
      if (sarr->buf[i].buf == NULL) { continue; }
      fprintf(fs, "%s\n", sarr->buf[i].buf);
   }
   return;
}

/*!:: Print c-string inside element's buffer (even if it's NULL) adding given
 *    prefix and postfix to each.
 * ::Errors:
 *    On error printing is stopped.
 *    . fprintf()    -- errors caused by this function
 */
void
sarr_printp (
   FILE* fs,               //! file stream
   const char* pref,       //! prefix c-string
   const strarr_t* sarr,   //! array of strings
   const char* post        //! postfix c-string
){
   ___Dassert(fs, "'fs' parameter invalid");
   ___Dassert(pref, "'pref' parameter invalid");
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(post, "'post' parameter invalid");

   for (uint i = 0; i < sarr->ns; ++i) {
      if (sarr->buf[i].buf == NULL) { continue; }
      fprintf(fs, "%s%s%s", pref, sarr->buf[i].buf, post);
   }
   return;
}
