#ifndef MH_LIB_BASE_STRINGS_STRINGS_ARR_H
#define MH_LIB_BASE_STRINGS_STRINGS_ARR_H


/*!F::Arrays of strings management - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* mh library */
#include "core.h"
#include "errors.h"
#include "strings.h"




/* Data. */
/* -------------------------------------------------------------------------- */


typedef struct strarr_t strarr_t;
/*!:: Array of strings.
 *    Contains array of string structs and array properties.
 */
struct strarr_t
{
   str_t* buf;    //! array of str_t
   uint na;       //! num of els - elements allocated
   uint ns;       //! num set - elements set
   int arb;       //! for arbitrary use by library user, library doesn't touch this at all
};
#define STRARR_T_Z (strarr_t){NULL,0,0,0}




/* Prepare and cleanup. */
/* -------------------------------------------------------------------------- */


/*!1:: Create array with first element created from given c-string. */
extern strarr_t* sarr_cre (const char* cstr)
                  __attr_nonnull__(1)
                  __attr_wures__
                  ;

/*!1:: Create array with buffer for specified number of empty strings. */
extern strarr_t* sarr_cre2 (uint nels)
                  __attr_wures__
                  ;

/*!1:: Create array with elements created form one or more c-strings given. */
extern strarr_t* sarr_crev (const char* cstr_first, ...)
                     __attr_nonnull__(1)
                     __attr_wures__
                     __attr_sentinel__(0)
                     ;

/*!1:: Same as `sarr_crev` but uses 'va_list' instead '...'. */
extern strarr_t* sarr_vcrev (const char* cstr_first, va_list vlist)
                     __attr_nonnull__(1)
                     __attr_wures__
                     ;

/*!1:: Create array of strings with one element from given printf-like format. */
extern strarr_t* sarr_cref (const char* format, ...)
                     __attr_wures__
                     __attr_format__(__printf__,1,2)
                     ;

/*!1:: Destroy array of strings. */
extern void sarr_des (strarr_t** ptr_strarr)
                  __attr_nonnull__(1)
                  ;

/*!1:: Initialize array with given c-string. */
extern void sarr_ini (strarr_t* strarr, const char* cstr)
                  __attr_nonnull__(1,2)
                  ;

/*!1:: Initialize array with buffer for specified number of empty strings. */
extern void sarr_ini2 (strarr_t* strarr, uint nels)
                  __attr_nonnull__(1)
                  ;

/*!1:: Initialize array with given string. */
extern void sarr_inis (strarr_t* strarr, const str_t* str)
                  __attr_nonnull__(1,2)
                  ;

/*!1:: Finalize array of strings. */
extern void sarr_fin (strarr_t* strarr)
                  __attr_nonnull__(1)
                  ;




/* Manipulate. */
/* -------------------------------------------------------------------------- */


/*!1:: Copy given c-string to array at specified index. */
extern char* sarr_cpy (strarr_t* strarr, uint idx, const char* cstr)
               __attr_nonnull__(1,3)
               ;

/*!1:: Copy given string to array at specified index. */
extern char* sarr_cpys (strarr_t* strarr, uint idx, const str_t* str)
               __attr_nonnull__(1,3)
               ;

/*!1:: Prepend given c-string to array. */
extern char* sarr_prep (strarr_t* strarr, const char* cstr)
               __attr_nonnull__(1,2)
               ;

/*!1:: Append given c-string to array. */
extern char* sarr_app (strarr_t* strarr, const char* cstr)
               __attr_nonnull__(1,2)
               ;

/*!1:: Append given c-string to array. Supply c-string properties. */
extern char* sarr_app2 (strarr_t* strarr, const char* cstr, size_t nc, size_t mnc)
               __attr_nonnull__(1,2)
               ;

/*!1:: Append given string to array. */
extern char* sarr_apps (strarr_t* strarr, const str_t* str)
               __attr_nonnull__(1,2)
               ;

/*!1:: Append printf-like formatted c-string to array. */
extern char* sarr_appf (strarr_t* stStrsArr, const char* fmt, ...)
               __attr_nonnull__(1,2)
               __attr_format__(__printf__,2,3)
               ;

/*!1:: Append all elements of given array of strings to destination array. */
extern char* sarr_appA (strarr_t* strarr_dest, const strarr_t* strarr_src)
               __attr_nonnull__(1,2)
               ;

/*!1:: Append all elements of given array of strings to destination array with
 *       prefix and postfix. */
extern void sarr_appAp (strarr_t* sarr_dest, const char* pref, const strarr_t* sarr_src, const char* post)
               __attr_nonnull__(1,2,3,4)
               ;

/*!1:: Attach given c-string to array. */
extern void sarr_att (strarr_t* sarr, char* cs)
               __attr_nonnull__(1,2)
               ;

/*!1:: Attach given c-string to array. Supply c-string properties. */
extern void sarr_att2 (strarr_t* sarr, char* cs, size_t nc, size_t mnc)
               __attr_nonnull__(1,2)
               ;

/* Mix elements of arrays, place result in dest array. */
extern void sarr_multipleA (strarr_t* sarr_dest, const strarr_t* sarr)
               __attr_nonnull__(1,2)
               ;




/* Splitting. */
/* -------------------------------------------------------------------------- */


/*!1:: Split `line` on `delim` and place each word in separate `sarr` element. */
extern void sarr_split_line (strarr_t* sarr, const char* line, const char* delim)
               __attr_nonnull__(1,2,3)
               ;

/*!1:: Split `line` on `delim` and place each word in separate `sarr` element,
 *       ommit empty splits. */
extern void sarr_split_line2 (strarr_t* sarr, const char* line, const char* delim)
               __attr_nonnull__(1,2,3)
               ;




/* Examine. */
/* -------------------------------------------------------------------------- */


/*!1:: Return c-string inside string element in array at specified index. */
extern char* sarr_el_ret (const strarr_t* sarr, uint idx)
            __attr_nonnull__(1)
            ;

/*!1:: Return pointer to string element in array at specified index. */
extern str_t* sarr_el_rets (const strarr_t* sarr, uint idx)
            __attr_nonnull__(1)
            ;

/*!1:: Return c-string inside subsequent elements in array of strings. */
extern char* sarr_it (const strarr_t* sarr, uint* idx)
               __attr_nonnull__(1,2)
               ;

/*!1:: Return pointers to subsequent elements in array of strings. */
extern str_t* sarr_its (const strarr_t* sarr, uint* idx)
               __attr_nonnull__(1,2)
               ;




/* Compare. */
/* -------------------------------------------------------------------------- */


/*!1:: Compare array element with c-string. */
extern int sarr_el_cmp (const strarr_t* sarr, size_t idx, const char* cs)
            __attr_nonnull__(1,3)
            ;

/*!1:: Compare each array element with c-string. */
extern ssize_t sarr_cmp (const strarr_t* sarr, const char* cs)
            __attr_nonnull__(1,2)
            ;

/*!1:: Compare each array element with string. */
extern ssize_t sarr_cmps (const strarr_t* sarr, const str_t* str)
            __attr_nonnull__(1,2)
            ;

/*!1:: Find elements of `needles` array in given c-string. */
//was sarr_cmp
extern char* sarr_str(size_t* el_num, const char* haystack, const strarr_t* needles)
                     __attr_nonnull__(1,2,3)
                     ;

/*!1:: Find given c-string in elements of `haystacks` array. */
//was sarr_rcmp
extern char* sarr_rstr(size_t* el_num, const strarr_t* haystacks, const char* needle)
                     __attr_nonnull__(1,2,3)
                     ;




/* Print. */
/* -------------------------------------------------------------------------- */


/*!1:: Print each element of array to given stream. */
extern void sarr_print (FILE* file_stream, const strarr_t* strs_arr)
               __attr_nonnull__(1,2)
               ;

/*!1:: Print each element of array to given stream on separate lines. */
extern void sarr_printn (FILE* file_stream, const strarr_t* strs_arr)
               __attr_nonnull__(1,2)
               ;

/*!1:: Print each element of array to given stream with prefix and postfix. */
extern void sarr_printp (FILE* file_stream,
      const char* prefix, const strarr_t* strs_arr, const char* postfix)
               __attr_nonnull__(1,2,3,4)
               ;




#ifdef DEBUG
bool strarr_all_tests_run (void)
         __attr_wures__;
#endif




#endif /*- MH_LIB_BASE_STRINGS_STRINGS_ARR_H -*/
