/*!F::Time manipulation - main implementation file.
 *
 *  ::Purpose:
 *    Reading, setting, transforming and manipulating time.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - not much yet, one function for stringize time
 *
 *  ::Limitations:
 *    - many other functions can be possibly implemented
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2013-12-30
 *    (c) Copyright:    Maciej Helminiak 2013
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */


/* localtime_r()  _POSIX_C_SOURCE >= 1 || _XOPEN_SOURCE
 *                || _BSD_SOURCE || _SVID_SOURCE || _POSIX_SOURCE
 */
#define _POSIX_C_SOURCE 1




/* current project */
#include "common_mhlbase.h"     // pr using this lib must contain that header

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"

/* libc */
#include <time.h>       // localtime(), strftime()

/* current module */
#include "time.h"       // time_t, struct tm, localtime(), strftime()




/*!:: Allocate memory for c-string representation of `stime`. Representation is
 *    formatted according to given format. Result is returned and placed in
 *    `pcstr` if it is not NULL.
 *    See strftime() libc function for format of `fmt`.
 *
 * ::Preconditions:
 *  - `stime` must contain valid simple time value
 *
 * ::Errors:
 *    nothing allocated, NULL returned, error set
 *    ENOMEM      -- not enough memory available
 *    EOVERFLOW   -- result can't be represented
 */
char*                   //! result
tm_stringize_time (
   char** pcstr,        //! [out] place for result (used if not NULL)
   time_t stime,        //! time in simple time format
   const char* fmt      //! strftime-like time format
){
   ___Dassert(stime, "'stime' parameter invalid");
   ___Dassert(fmt, "'fmt' parameter invalid");

   char* cstr_caltime = NULL;

   /*: Get high resolution time from simple time. */
   struct tm bdtime = {0};
   struct tm* res = localtime_r(&stime, &bdtime);
      ErrorTrigIf(res == NULL, handle_err);

   /*: Check size of stringized time. */
   char tmp_buf[4096];
   size_t nb = strftime(tmp_buf, 4096, fmt, &bdtime) + 1*sizeof(char);

   /*: Core action:
    *    . allocate memory for cstring
    *    . create formatted time
    */
   cstr_caltime = mem_alloc(nb);
      HandleErr(handle_err);
   strftime(cstr_caltime, nb, fmt, &bdtime);      //nl_langinfo(D_T_FMT)
   cstr_caltime[nb-1] = '\0';

   /*: Assign to destination pointer if provided. */
   if (pcstr != NULL) { *pcstr = cstr_caltime; }

   /*: Finish: */
      ___Dassert(cstr_caltime, "'cstr_caltime' return assert failed");
   return cstr_caltime;

  handle_err:
   mem_free(cstr_caltime);
   return NULL;
}




/* REMAINDER */
/*
 * time           - abstract concept
 * calendar/absoute time  - point in time continuum     04.01.2002 18:25:55 UTC
 *
 * interval    - contiguous part of time continuum, between two calendar times
 *             9:00 - 10:00 on 04.01.2001
 * elapsed time- length of interval
 *             1 hour
 * amount of time - sum of elapsed times (may be in different intervals)
 *             12 hours
 * period      - elapsed time of interval between two events
 *             2 days
 *
 *
 * CALENDAR TIME
 * time_t         - simple time, elapsed time, secs
 * struct timeval - high-resolution time, elasped time, fractions of sec
 * struct tm      - local/broken-down time, absolute time for specific timezone
 */
