/*!F::Basic operations on directory streams - main implementation file.
 *
 *  ::Purpose:
 *    Implement basic operations in safe, robust and clean manner.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - open, close directory streams
 *
 *  ::Limitations:
 *    - to implement: telldir, seekdir, rewinddir
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-02
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_mhlfs.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"

/* libc */
#include <unistd.h>     // fpathconf()
#include <dirent.h>     // opendir(), DIR, struct dirent, dirfd()
#include <sys/types.h>  // opendir(), dirfd()

/* current module */
#include "dstream_basic.h"




/*!:: Open given directory and associate it with directory stream.
 * ::Preconditions:
 * ::Errors:
 *    EACCES      -- permissions denied
 *    EMFILE      -- too many fds owned by process
 *    ENFILE      -- too many fds owned by OS
 *    ENOENT      -- directory doesn't exist or empty string given as name
 *    ENOTDIR     -- name is not directory
 *    ENOMEM      -- no memory
*/
DIR*                    //! directory stream
ds_open (
   const char* dname    //! directory name
){
   ___Dassert(dname, "'dname' parameter invalid");

   DIR* ds_new = opendir(dname);
      ErrorTrigIf(errno != 0, handle_err);

   return ds_new;

  handle_err:
   return NULL;
}

/*!:: Close given directory stream. Also closes underlying file descriptor.
 * ::Preconditions:
 * ::Errors:
 *    EBADF       -- file descriptor in directory stream is invalid
*/
void
ds_close (
   DIR** ds       //! directory stream
){
   ___Dassert(ds, "'ds' parameter invalid");
   ___Dassert(*ds, "'*ds' parameter invalid");

   closedir(*ds);
      ErrorTrigIf(errno != 0, handle_err);
   *ds = NULL;

   return;

  handle_err:
   return;
}

/*!:: Return maximum length of file system entry name for given directory
 *    stream.
 * ::Preconditions:
 *  - `ds` must point to valid directory stream, otherwise behavior is
 *    undefined
 * ::Errors:
 *    return -1:     length could not be determined
*/
long                 //! max length of entry name
file_name_len (
   const DIR* ds     //! directory stream
){
   ___Dassert(ds, "'ds' parameter invalid");

   long name_len = 0;
#  if defined(HAVE_FPATHCONF) && defined(HAVE_DIRFD) && defined(_PC_NAME_MAX)
      name_len = fpathconf(dirfd(ds), _PC_NAME_MAX);
      if (name_len == -1)
#     if defined(NAME_MAX)
         name_len = (NAME_MAX > 255) ? NAME_MAX : 255;
#     else
         name_len = -1;
#     endif
#  else
#     if defined(NAME_MAX)
         name_len = (NAME_MAX > 255) ? NAME_MAX : 255;
#     else
         name_len = -1;
#     endif
#  endif
   return name_len;
}
