/*!F::Directories information retrieve - main implementation file.
 *
 *  ::Purpose:
 *    Operations retrieving informations about directories.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - allocate memory for entry struct with proper 'struct dirent' size
 *    - read entry, read all entries in directory with stats
 *    - count entries in directory
 *    - globbing and word expanding (later not finished)
 *
 *  ::Limitations:
 *    - not implemented yet: readdir, scandir, getcwd, chdir an meybe others
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-02
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* readdir_r(): _POSIX_C_SOURCE >= 1 || _XOPEN_SOURCE || _BSD_SOURCE
 *              || _SVID_SOURCE || _POSIX_SOURCE */
//#define _POSIX_C_SOURCE 1

/* readlink(): _BSD_SOURCE || _XOPEN_SOURCE >= 500 || _XOPEN_SOURCE
 *             && _XOPEN_SOURCE_EXTENDED || _POSIX_C_SOURCE >= 200112L */
#define _POSIX_C_SOURCE 200112L




/* current project */
#include "common_mhlfs.h"
#include "fs_storage.h"
//#include "ds_basic.h"
#include "stat_info.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"
#include "cstrings.h"
#include "strings.h"
#include "strings_arr.h"

/* libc */
#include <errno.h>
#include <string.h>     // memcpy()
#include <dirent.h>     // opendir(), DIR, struct dirent
#include <sys/stat.h>   // stat(), struct stat
#include <unistd.h>     // readlink()
#include <wordexp.h>    // wprdexp()
#include <glob.h>       // glob()

/* current module */
#include "dstream_info.h"
#include "dstream_info_internal.h.inc"
#include "dstream_info_aux.h.inc"





/* Entry. */
/* -------------------------------------------------------------------------- */


/*!:: Allocate memory for structure and its members, particulary libc entry
 *    struct and libc stats struct.
 * ::Preconditions:
 *  - `ds` must point to valid dirstream, otherwise behavior is undefined
 * ::Errors:
 *    Nothing created.
 *    ENOMEM      -- no memory available
*/
st_ent*              //! entry structure
ds_ent_alloc (
   const DIR* ds     //! dirstream - needed only to compute 'struct dirent' size
){
   ___Dassert(ds, "'ds' parameter invalid");

   st_ent* stEntry_new = NULL;      // entry struct to create

   /*: Allocate entry structure. */
   stEntry_new = mem_alloc(sizeof(*stEntry_new));
      HandleErr(handle_err);
   ent_zero_(stEntry_new);

   /*: Init memebers, particulary allocate 'struct dirent'. */
   size_t dirent_size = dirent_st_size(ds);
   ent_init_(stEntry_new, dirent_size);
      HandleErr(handle_err);

   /*: Finish: */
   return stEntry_new;

  handle_err:
   mem_free(stEntry_new);
   return NULL;
}

/*!:: Release resources inside struct and struct itself.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
*/
void
ds_ent_free (
   st_ent** pstEnt      //! pointer to entry structure
){
   ___Dassert(pstEnt, "'pstEnt' parameter invalid");
   ___Dassert(*pstEnt, "'*pstEnt' parameter invalid");

   st_ent* stE = *pstEnt;     // helper variable

   /*: Release resources inside struct and struct itself. */
   ent_finish_(stE);
   mem_free(*pstEnt);

   /*: Finish: */
   return;
}

/*!:: Place information in libc dirent struct inside entry struct, compose full
 *    file name.
 * ::Preconditions:
 * ::Errors:
 *    Nothin read.
 *    EBADF    -- invalid dirstream
*/
bool                    //! true if entry read, false if no more entries or error
ds_read_entry (
   st_ent* stEnt,       //! [out] entry structure
   DIR* ds,             //! dirstream
   const char* dname    //! dir name - used only to compose file name
){
   ___Dassert(stEnt, "'stEnt parameter invalid");
   ___Dassert(ds, "'ds' parameter invalid");

   /*: Read next entry from dir stream. */
   struct dirent* result = NULL;
   int err = readdir_r(ds, stEnt->dent, &result);
      ErrorTrigIf(err != 0, handle_err);

   if (result == NULL) { goto entry_none; }  //: no more entries

   /*: Create name and filename from dirname and entry name. */
   str_fin(&stEnt->entryname);
   str_ini(&stEnt->entryname, stEnt->dent->d_name);
   str_fin(&stEnt->filename);
   str_inif(&stEnt->filename, "%s/%s", dname, stEnt->dent->d_name);
      HandleErr(handle_err);

   /*: Finish: */
   return true;      //: entry read

  entry_none:
   return false;     //: no more entries

  handle_err:
   str_fin(&stEnt->filename);
   str_fin(&stEnt->entryname);
   return false;
}




/* Directory. */
/* -------------------------------------------------------------------------- */


/*!:: Allocate memory for struct and its members, open directory stream.
 * ::Preconditions:
 * ::Errors:
 *    Nothin created.
 *    ENOMEM      -- no memory available
 *    ds_open()   -- this function errors [ds_basic.c]
*/
st_dir*                 //! directory structure
ds_dir_create (
   const char* dname    //! directory name
){
   ___Dassert(dname, "'dname' parameter invalid");

   st_dir* Dir_new = NULL;    // dir entry to create
   char* dirname = NULL;      // dir name
   DIR* ds = NULL;            // dirstream

   /*: Allocate struct: */
   Dir_new = mem_alloc(sizeof(*Dir_new));
      HandleErr(handle_err);

   dirname = cs_cre(dname);      //: assign dirname
      HandleErr(handle_err);

   /*: Open dirstream: */
   ds = ds_open(dname);
      HandleErr(handle_err);

   /*: Assign struct members: */
   Dir_new->this_entry = (st_ent){0};
   str_fill(&Dir_new->this_entry.filename, dirname);
   Dir_new->this_entry.ds = ds;
   Dir_new->Dentrs = NULL;
   Dir_new->nels = 0;

   /*: Finish: */
   return Dir_new;

  handle_err:
   cs_des(&dirname);
   mem_free(Dir_new);
   return NULL;
}

/*!:: Release resources inside struct and struct itself. Will destroy each
 *    entry inside, will close dirstream.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
*/
void
ds_dir_destroy (
   st_dir** pstEntrs    //! pointer to directory structure
){
   ___Dassert(pstEntrs, "'pstEntrs' parameter invalid");
   ___Dassert(*pstEntrs, "'*pstEntrs' parameter invalid");

   ds_close(&(*pstEntrs)->this_entry.ds);
      HandleErr(handle_err);
  handle_err:
   //try to release resources on error

   dir_finish_(*pstEntrs);
   mem_free(*pstEntrs);
   return;
}

/*!:: Entries are placed in array inside directory struct. Array is expanded
 *    as needed.
 * ::Preconditions:
 *    - `stDir` must be created and dirstream inside open
 * ::Errors:
 *    Not all entries read or none.
 *    ENOMEM         -- no memory available
 *    E_OVERFLOWINT  -- overal number of entries would exceed SIZE_MAX
 *    EBADF          -- invalid dirstream inside struct
 *    stat_read_attrs() -- this function errors [stat_info.c]
*/
void
ds_read_directory (
   st_dir* stDir,    //! pointer to directory structure
   bool read_stats   //! read stats for each entry?
){
   ___Dassert(stDir, "'stDir' parameter invalid");

   st_ent* stEnt = NULL;
   size_t nels_put = 0;    // entries already read into dir struct

   /*: Expand struct: */
   dir_expand_ (stDir, 128);
      HandleErr(handle_err);

   /*: For each entry read from dirstream: */
   while (true)
   {
      /*: Alloc new entry struct: */
      stEnt = ds_ent_alloc(stDir->this_entry.ds);
         HandleErr(handle_err);

      /*: Read entry, break loop if no more entries: */
      bool present = ds_read_entry(stEnt, stDir->this_entry.ds, stDir->this_entry.filename.buf);
         HandleErr(handle_err);
      if (!present) { ds_ent_free(&stEnt); break; }

      /*: Read stats if requested: */
      if (read_stats) {
         // in case of not critical errors do not throw error
         // set 'attrs' to 0 instead to indicate that attributes
         // couldn't be read
         // 0 can't be valid inode number, its reserved for "no inode"
         stat_read_attrs(stEnt);
         if (errno == ELOOP || errno == ENAMETOOLONG || errno == EOVERFLOW) {
            stEnt->attrs = (struct stat){0};
            errno = 0;
         }
            HandleErr(handle_err);
         //maybe: check inodes in dirent and stat?
      }

      /*: Expand array of entries inside directory struct if needed: */
      if (nels_put >= stDir->nels) {
         dir_expand_ (stDir, 128);
            HandleErr(handle_err);
      }

      /*: Place entry struct inside directory struct: */
      stDir->Dentrs[nels_put] = stEnt;
      ++nels_put;
   }

   stDir->nels = nels_put; //: assign num of entries

   /*: Finish: */
   return;

  handle_err:
   if (stEnt != NULL) { ds_ent_free(&stEnt); }
   stDir->nels = nels_put;
   dir_finish_(stDir);
   return;
}

/*!:: Entries are placed in array inside directory struct. Array is expanded
 *    as needed.
 * ::Preconditions:
 *    - `stDir` must be created and dirstream inside open
 * ::Errors:
 *    Not all entries read or none.
 *    ENOMEM         -- no memory available
 *    E_OVERFLOWINT  -- overal number of entries would exceed SIZE_MAX
 *    EBADF          -- invalid dirstream inside struct
*/
void
ds_read_directory_stats (
   st_dir* stDir     //! pointer to directory structure
){
   ___Dassert(stDir, "'stDir' parameter invalid");
   /*: Use other function to perform task. */
   ds_read_directory(stDir, true);
   return;  // error is silently propagated (it stays untouched)
}

/*!:: If no entries are found or error occured 0 is returned.
 * ::Preconditions:
 * ::Errors:
 *    Not all entries read counted none.
 *    EBADF          -- invalid dirstream inside struct
 *    ds_open()   -- this function errors [ds_basic.c]
 *    ds_closr()  -- this function errors [ds_basic.c]
*/
size_t                     //! number of entries
ds_count_dir_entries (
   const char* dname       //! directory name
){
   ___Dassert(dname, "'dname' parameter invalid");

   size_t nentrs = 0;

   /*: Open directory: */
   DIR* ds = ds_open(dname);
      HandleErr(handle_err);

   /*: Count entries - read each entry in directory. */
   struct dirent* result = NULL;
   struct dirent dent = {0};
   while (true)
   {
      int err = readdir_r(ds, &dent, &result);
         ErrorTrigIf(err != 0, handle_err);
      if (result != NULL) {
         ++nentrs;
      } else {
         break;
      }
   }

   /*: Finish: */
   ds_close(&ds);
   return nentrs;

  handle_err:
   if (ds != NULL) { ds_close(&ds); }
   return 0;
}




/* Miscelaneous. */
/* -------------------------------------------------------------------------- */


/*!:: not finished not tested */
void
ds_wordexpand_name (
   strarr_t* sarr_mat,    //matches
   const char* name
){
   ___Dassert(sarr_mat, "'sarr_mat' parameter invalid");
   ___Dassert(name, "'name' parameter invalid");

   wordexp_t stWordexp = {0};

   int flags = GLOB_MARK;  //mark each dir with trailing '/'
#  if defined (_GNU_SOURCE)
   flags |= GLOB_BRACE;  //use {a,b} like expansion
#  endif
      errno = 0;
   int ret = wordexp(name, &stWordexp, flags);
      ErrorTrigIf(errno != 0 && errno != EACCES, handle_err);
      ErrorSetIf(ret == GLOB_NOSPACE, ENOMEM, handle_err);
      ErrorTrigIf(ret == GLOB_ABORTED, handle_err);
      if (ret == GLOB_NOMATCH) { goto no_match; }

   size_t i = 0;
   char* cs_match = NULL;
   while ((cs_match = stWordexp.we_wordv[i]) != NULL) {
      sarr_app(sarr_mat, stWordexp.we_wordv[i]);
         HandleErr(handle_err);
      ++i;
   }

   wordfree(&stWordexp);

   goto handle_succ;

  handle_succ:
   return;

  no_match:
   return;

  handle_err:
   wordfree(&stWordexp);
   return;
}

/*!:: Pointer to globbing errors.
 * :: Points to array of strings describing errors detected during globbing in
 *    `ds_glob_name` function.
 *    Instance of array is provided by caller, thus implementation is
 *    considered thread-safe.
 */
__tls strarr_t* glob_name_errors = NULL;

/*!:: Append error description to array of strings pointed to by thread-local
 *    variable `glob_name_errors`. Pointer is set in each call to
 *    `ds_glob_name()`.
 * .. Used exclusively by `ds_glob_name()` function
 * ::Preconditions:
 * ::Errors:
 *    Error description not appended to array.
 *    ENOMEM         -- no memory available
*/
static inline int glob_errfun (const char* epath, int errnum) {
   sarr_appf(glob_name_errors, "%s: %s", epath, strerror(errnum));
      HandleErr(handle_err);
   return 0;
  handle_err:
   return 1;
}

/*!:: Find file names (fullpaths) matching given pattern and place them in
 *    array of strings. If any errors during globbing occured place them in
 *    another array of strings.
 * ::Preconditions:
 * ::Errors:
 *    Error description not appended to array.
 *    ENOMEM         -- no memory available
 *    opendir()      -- this function errors, used internally by `glob()`
*/
void
ds_glob_name (
   strarr_t* sarr_mat,     //! matching fullpaths
   strarr_t* sarr_errs,    //! scanning errors - array of strings describing errors encountered but not caused to abort scanning
   const str_t* pattern    //! pattern
){
   ___Dassert(sarr_mat, "'sarr_mat' parameter invalid");
   ___Dassert(sarr_errs, "'sarr_errs' parameter invalid");
   ___Dassert(pattern, "'pattern' parameter invalid");

   /*: Global pointer to errors array: */
   glob_name_errors = sarr_errs;

   /*: Prepare: */
   glob_t stGlob = {0};

   //stGLob->gl_offs = 0;    // cstrs to reserve for own use in stGlob.gl_pathv
   //    requires GLOB_DOOFFS
   //int flags = GLOB_MARK;  //mark each match dir with trailing '/'
   int flags = 0;    // globbing flags
   /*
#  if defined (_GNU_SOURCE)
   flags |= GLOB_BRACE;  //use {a,b} like expansion
   flags |= GLOB_NOMAGIC;//no metachars in pattern: return pattern even if no file of that pattern
#  endif
*/

   /*: Perform globbing: */
      errno = 0;
   int ret = glob(pattern->buf, flags, glob_errfun, &stGlob);
      ErrorTrigIf(errno != 0 && errno != EACCES && errno != ENOENT, handle_err); // other error
      ErrorSetIf(ret == GLOB_NOSPACE, ENOMEM, handle_err);  // no memory
      ErrorTrigIf(ret == GLOB_ABORTED, handle_err);         // read error
      if (ret == GLOB_NOMATCH) { ClrErrCode(); goto no_match; }  //sets ENOENT too
      //ENOENT - dangling link
      //Probably need cover cases when acces is denied - add faults to array

   /*: For each matched file name: */
   size_t i = 0;
   char* cs_match = NULL;
   while ((cs_match = stGlob.gl_pathv[i]) != NULL) {
      //: add to array of matches
      sarr_app(sarr_mat, cs_match);
         HandleErr(handle_err);
      ++i;
   }

   /*: Finish: */
   globfree(&stGlob);
   return;

  no_match:
   globfree(&stGlob);
   sarr_app(sarr_mat, "");
      HandleErr(handle_err);
   return;

  handle_err:
   globfree(&stGlob);
   return;
}
