/*!F::Operations on directories - main implementation file.
 *
 *  ::Purpose:
 *    Operations changing the state of directories.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - making entries, links, removing entries, copying, moving
 *
 *  ::Limitations:
 *    - more operations may be implemented
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-02
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* symlink():  _BSD_SOURCE || _XOPEN_SOURCE >= 500 || _XOPEN_SOURCE
 *             && _XOPEN_SOURCE_EXTENDED || _POSIX_C_SOURCE >= 200112L */
//#define _POSIX_C_SOURCE 200112L

/* lstat():    _BSD_SOURCE || _XOPEN_SOURCE >= 500 || _XOPEN_SOURCE &&
 *             _XOPEN_SOURCE_EXTENDED || _POSIX_C_SOURCE >= 200112L */

/* dirfd():    _BSD_SOURCE || _SVID_SOURCE
 *             || /Since glibc 2.10:/ (_POSIX_C_SOURCE >= 200809L || _XOPEN_SOURCE >= 700)*/

/* fchmod():   _BSD_SOURCE || _XOPEN_SOURCE >= 500 || _XOPEN_SOURCE && _XOPEN_SOURCE_EXTENDED
 *             || Since glibc 2.12: _POSIX_C_SOURCE >= 200809L */
#define _POSIX_C_SOURCE 200809L




/* current project */
#include "common_mhlfs.h"
#include "fs_storage.h"
//#include "ds_basic.h"
//#include "ds_info.h"
#include "stat_info.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"
#include "cstrings.h"

/* libc */
#include <sys/stat.h>   // mkdir(), open(), fchmod()
#include <sys/types.h>  // mkdir(), open()
#include <fcntl.h>      // open()
#include <unistd.h>     // close(), symlink()
#include <libgen.h>     // basename() - POSIX version (there is GNU ver)
#include <stdlib.h>     // mkdtemp()
#include <string.h>     // strcmp()

/* current module */
#include "dstream_operate.h"
#include "dstream_operate_internal.h.inc"
#include "dstream_operate_aux.h.inc"




/* Make. */
/* -------------------------------------------------------------------------- */


/* :: Create file with given mode.
 * ::Preconditions:
 * ::Errors:
 *    Nothing created.
 *    open()      -- this libc function errros
*/
FD
ds_mkfile (
   const char* cs_fname,
   mode_t mode
){
   ___Dassert(cs_fname, "'cs_fname' parameter invalid");

   int fd = open(cs_fname, O_WRONLY | O_CREAT | O_EXCL, mode);
      ErrorTrigIf(fd == -1, handle_err);

   /*: Finish: */
   return fd;

  handle_err:
   return -1;
}

/* :: Same as 'ds_mkfile()' but close fd.
*/
void
ds_mkfile2 (
   const char* cs_fname,
   mode_t mode
){
   ___Dassert(cs_fname, "'cs_fname' parameter invalid");

   FD fd = ds_mkfile (cs_fname, mode);
      HandleErr(handle_err);
   int ret = close(fd);
      ErrorTrigIf(ret == -1, handle_err);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* :: Create directory with given mode.
 * ::Preconditions:
 * ::Errors:
 *    Nothing created.
 *    mkdir()     -- this libc function errros
*/
void
ds_mkdir (
   const char* cs_dname,
   mode_t mode
){
   ___Dassert(cs_dname, "'cs_dname' parameter invalid");

   int ret = mkdir(cs_dname, mode);
      ErrorTrigIf(ret == -1, handle_err);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* :: Create hard link to given entry.
 * ::Preconditions:
 * ::Errors:
 *    Nothing created.
 *    link()      -- this libc function errros, among others:
 *    . EEXIST    -- `cs_lname` already exist
 *    . EPERM     -- attempt ot hl directory by non-root user OR fs does not allow hlinks
 *    . EXDEV     -- different mountpoints or filesystems
*/
void
ds_mkhlink (
   const char* cs_hname,
   const char* cs_tar
){
   ___Dassert(cs_hname, "'cs_hname' parameter invalid");
   ___Dassert(cs_tar, "'cs_tar' parameter invalid");

   int ret = link(cs_tar, cs_hname);
      ErrorTrigIf(ret == -1, handle_err);

   /*: Finish */
   return;

  handle_err:
   return;
}

/* :: Create symbolic link to given entry.
 * ::Preconditions:
 * ::Errors:
 *    Nothing created.
 *    symlink()   -- this libc function errros, among others:
 *    . EEXIST    -- `cs_lname` already exist
 *    . EPERM     -- fs not support symlinks
*/
void
ds_mkslink (
   const char* cs_lname,
   const char* cs_tar
){
   ___Dassert(cs_lname, "'cs_lname' parameter invalid");
   ___Dassert(cs_tar, "'cs_tar' parameter invalid");

   int ret = symlink(cs_tar, cs_lname);
      ErrorTrigIf(ret == -1, handle_err);

   /*: Finish */
   return;

  handle_err:
   return;
}




/* Remove. */
/* -------------------------------------------------------------------------- */


/* :: Remove directory and all entries within.
 * ::Preconditions:
 * ::Errors:
 *    Nothing removed by some temporary directories may be left.
 *    many errors
*/
void
ds_rmdir (
   const char* dname                //! directory name to remove
){
   ___Dassert(dname, "'dname' parameter invalid");
   ds_rmdir2(dname, NULL);
}

/* :: Same as 'ds_rmdir()' except:
 *    caller may supply attributes of directory - it will be used internally
 *    instead of retrieving them.
*/
void
ds_rmdir2 (
   const char* dname,               //! directory name to remove
   const struct stat* attrs_orig    //! dir attributes for internal checks taht directory is same as original, may be NULL
){
   ___Dassert(dname, "'dname' parameter invalid");

   str_t tempdir_name = {0};
   str_t tentry_name = {0};
   DIR* ds = NULL;

   /*: Create temporary directory and move tobe-deleted dir into there */
   create_temp_dir_for_file(&tempdir_name, dname);                //mkdtemp
      HandleErr(handle_err);
   move_entry_to_temp_dir(&tentry_name, dname, tempdir_name.buf); //rename
      HandleErr(handle_err);

   /*: Check if directory is the same as original */
   if (attrs_orig != NULL) {
      struct stat attrs = {0};
      int res2 = lstat(tentry_name.buf, &attrs);
         ErrorTrigIfm(res2 == -1, handle_err, "failed read file attributes of '%s'", tentry_name.buf);
      if (!files_same(attrs_orig, &attrs, 0)) {
         ErrorSetm(E_STATE_CHANGE, handle_err, "directory '%s' replaced during operation", dname);
      }
   }

   /*: Remove each entry in directory */
   ds = opendir(tentry_name.buf);
      ErrorTrigIf(ds == NULL, handle_err);
   while (true)
   {
      //: Allocate entry struct
      st_ent* stEnt = ds_ent_alloc(ds);
         HandleErr(handle_err_while);

      //: Read next entry
      bool present = ds_read_entry(stEnt, ds, tentry_name.buf);
         HandleErr(handle_err_while);
      if (!present) { ds_ent_free(&stEnt); break; }
      if (strcmp(stEnt->dent->d_name, "..") == 0
      || strcmp(stEnt->dent->d_name, ".") == 0) {
         /*cleanup*/ ds_ent_free(&stEnt);
         continue;
      }

      //: Remove entry from file system
      ds_rmentry(stEnt->filename.buf);
         HandleErr(handle_err_while);

      /*cleanup*/ ds_ent_free(&stEnt);
      continue;

      /*: Loop error: */
     handle_err_while:
      if (stEnt != NULL) { ds_ent_free(&stEnt); }
      goto handle_err;
   }
   int res = closedir(ds);
      ErrorTrigIf(res == -1, handle_err);

   /* Remove directory itself and cleanup temp directory */
   res = rmdir(tentry_name.buf);
      ErrorTrigIfm(res == -1, handle_err, "failed remove '%s' directory or '%s'", tentry_name.buf, dname);
   remove_temp_dir_for_file(&tempdir_name);                    //rmdir
      HandleErr(handle_err);

   /*cleanup*/ str_fin(&tentry_name);
   /*cleanup*/ str_fin(&tempdir_name);

   /*: Finish: */
   return;

  handle_err:
   if (ds != NULL) { closedir(ds); }
   if (tentry_name.nc > 0) {
      int res2 = rename(tentry_name.buf, dname);
      if (res2 == -1) {
         err_append_msg(", unable to move back temporary directory '%s'", tentry_name.buf);
      }
      str_fin(&tentry_name);
   }
   if (tempdir_name.nc > 0) {
      int res2 = rmdir(tempdir_name.buf);
      if (res2 == -1) {
         err_append_msg(", unable to remove temporary directory '%s'", tempdir_name.buf);
      }
      str_fin(&tempdir_name);
   }
   return;
}

/* :: Remove file from file system.
 * ::Preconditions:
 * ::Errors:
 *    Nothing removed by some temporary files may be left.
*/
void
ds_rmfile (
   const char* fname                //! file name to remove
){
   ___Dassert(fname, "'fname' parameter invalid");
   ds_rmfile2(fname, NULL);
}

/* :: Same as 'ds_rmfile()' except:
 *    caller may supply attributes of file - it will be used internally
 *    instead of retrieving them.
*/
void
ds_rmfile2 (
   const char* fname,               //! file name to remove
   const struct stat* attrs_orig    //! dir attributes for internal checks that file is same as original, may be NULL
){
   ___Dassert(fname, "'fname' parameter invalid");

   str_t tempdir_name = {0};
   str_t tentry_name = {0};

   /* Create temporary directory and move tobe-deleted file into there */
   create_temp_dir_for_file(&tempdir_name, fname);                //mkdtemp
      HandleErr(handle_err);
   move_entry_to_temp_dir(&tentry_name, fname, tempdir_name.buf); //rename
      HandleErr(handle_err);

   /* Check if file is the same as original */
   if (attrs_orig != NULL) {
      struct stat attrs = {0};
      int res = lstat(tentry_name.buf, &attrs);
         ErrorTrigIfm(res == -1, handle_err, "failed read file attributes of '%s'", tentry_name.buf);
      if (!files_same(attrs_orig, &attrs, 0)) {
         ErrorSetm(E_STATE_CHANGE, handle_err, "directory '%s' replaced during operation", fname);
      }
   }

   /* unlink file and cleanup */
   int res = unlink(tentry_name.buf);
      ErrorTrigIfm(res == -1, handle_err, "failed unlink entry '%s'", tentry_name.buf);
   remove_temp_dir_for_file(&tempdir_name);                    //rmdir
      HandleErr(handle_err);

   /*cleanup*/ str_fin(&tentry_name);
   /*cleanup*/ str_fin(&tempdir_name);

   /*: Finish: */
   return;

  handle_err:
   if (tentry_name.nc > 0) {
      int res2 = rename(tentry_name.buf, fname);
      if (res2 == -1) {
         err_append_msg(", unable to move back temporary directory '%s'", tentry_name.buf);
      }
      str_fin(&tentry_name);
   }
   if (tempdir_name.nc > 0) {
      int res2 = rmdir(tempdir_name.buf);
      if (res2 == -1) {
         err_append_msg(", unable to remove temporary directory '%s'", tempdir_name.buf);
      }
      str_fin(&tempdir_name);
   }
   return;
}

/* :: Same as 'ds_rmfile()' except:
 *    caller may supply attributes of file - it will be used internally
 *    instead of retrieving them.
*/
void
ds_rmentry (
   const char* ename                //! entry name to remove
){
   ___Dassert(ename, "'ename' parameter invalid");
   ds_rmentry2(ename, NULL);
}

/* :: Use 'ds_rmfile' for files or 'ds_rmdir' for directories.
 * ::Preconditions:
 * ::Errors:
 *    Nothing removed by some temporary files may be left.
*/
void
ds_rmentry2 (
   const char* ename,               //! entry name to remove
   const struct stat* attrs_orig    //! dir attributes for internal checks that file is same as original, may be NULL
){
   ___Dassert(ename, "'ename' parameter invalid");

   /*: Get attrs of tobe-removed entry */
   struct stat attrs = {0};
   int res = lstat(ename, &attrs);
      ErrorTrigIfm(res == -1, handle_err, "can not read attributes of '%s'", ename);

   /* Check if file is the same as original */
   if (attrs_orig != NULL) {
      if (!files_same(attrs_orig, &attrs, 0)) {
         ErrorSetm(E_STATE_CHANGE, handle_err, "'%s' replaced during operation", ename);
      }
   }

   /*: Remove entry with adequate function: */
   if (S_ISDIR(attrs.st_mode)) {
      ds_rmdir2(ename, &attrs);
         HandleErr(handle_err);
   } else {
      ds_rmfile2(ename, &attrs);
         HandleErr(handle_err);
   }

   /*: Finish: */
   return;

  handle_err:
   return;
}




/* Copy/Remove. */
/* -------------------------------------------------------------------------- */


/* :: Both files can be of any type. Content and permissions are copied.
 *    Default behavior if destination file exist is to fail with EEXIST.
 *    Default behavior if destination is directory is to fail with EEXIST.
 * ::Flags:
 *    FSF_OVERW      -- overwrite target if exist (also directory)
 *    FSF_INTO_DIR   -- copy into directory if destination is directory, has precedence over FSF_OVERW
 * ::Note:
 *    If destination is directory and FSF_OVERW flag is specified directory
 *    will be removed. But if FSF_INTO_DIR flag is specified, file will be
 *    copied into directory regardless FSF_OVERW is set or not.
 * ::Preconditions:
 * ::Errors:
 *    Nothing copied or directories may be partialy copied and some temporary
 *    files may be left unlinked.
 *    many errors
*/
void
ds_copy (
   const char* fname_src,     //! source file name
   const char* fname_dest,    //! destination file name
   int flags                  //! file system operations flags
){
   ___Dassert(fname_src, "'fname_src' parameter invalid");
   ___Dassert(fname_dest, "'fname_dest' parameter invalid");
   ds_copy2(fname_src, fname_dest, NULL, NULL, flags);
}


/* :: Same as 'ds_copy()' except:
 *    caller may supply attributes of one or both of files - they will be used
 *    internally instead of retrieving them.
*/
void
ds_copy2 (
   const char* fname_src,     //! source file name
   const char* fname_dest,    //! destination file name
   struct stat* at_s,         //! source entry attributes for internal checks that file is same as original, may be NULL
   const struct stat* at_d,   //! destination entry attributes for internal checks that file is same as original, may be NULL
   int flags                  //! file system operations flags
){
   ___Dassert(fname_src, "'fname_src' parameter invalid");
   ___Dassert(fname_dest, "'fname_dest' parameter invalid");

   char typ_s = '\0';
   char typ_d = '\0';
   bool can_tar = false;   // do not canonicalize target

   /*: Initial attributes retrieval for involved files */
   struct stat at_s2 = {0};
   struct stat at_d2 = {0};

   if (at_s == NULL) {
      int res = lstat (fname_src, &at_s2);
         ErrorTrigIf(res == -1, handle_err);        //src nonexist - error
      at_s = &at_s2;
   }

   if (at_d == NULL) {
      int res = lstat (fname_dest, &at_d2);
         if (res == -1 && errno == ENOENT) {
            errno = 0; typ_d = 'N';
         } else {
            ErrorTrigIf(res == -1, handle_err);
            at_d = &at_d2;
         }
   }
   if (at_d != NULL && at_s->st_ino == at_d->st_ino) { at_s->st_ino = 0; return; }
      //set ino to 0 to avoide removing file in ds_move2()

   stat_extract_ftype(&typ_s, at_s->st_mode);
   if (typ_d == '\0') { stat_extract_ftype(&typ_d, at_d->st_mode); }

   //printf("st=%c dt=%c\n", typ_s, typ_d);

   /*: Take action according to operands type */
   switch (typ_s)
   {
      /*: source file is symlink */
      case 'l':
         if (flags & FSF_CANON_TARGET) { can_tar = true; }
         switch (typ_d) {
            case 'N':
               cp_l_N(fname_src, fname_dest, at_s, can_tar);
                  HandleErr(handle_err);
               break;
            case 'l':   /*- \/ THROUGH -*/
            case '-':
            case 's':
            case 'f':
            case 'b':
            case 'c':
               if (flags & FSF_OVERW) {
                  ds_rmfile2(fname_dest, at_d);
                     HandleErr(handle_err);
                  cp_l_N(fname_src, fname_dest, at_s, can_tar);
                     HandleErr(handle_err);
               } else {
                  ErrorSet(EEXIST, handle_err);
               }
               break;
            case 'd':
               if (flags & FSF_INTO_DIR) {
                  char* ftmp = cs_cre(fname_src);
                  char* fdest = cs_cref("%s/%s", fname_dest, basename(ftmp));

                  ds_copy2(fname_src, fdest, at_s, NULL, flags & ~FSF_INTO_DIR);
                     HandleErr(handle_err);
/*
                  if (flags & FSF_OVERW) {
                     ds_copy2(fname_src, fdest, at_s, NULL, FSF_OVERW);
                        HandleErr(handle_err);
                  } else {
                     cp_l_N(fname_src, fdest, at_s, can_tar);
                        HandleErr(handle_err);
                  }*/
                  /*cleanup*/ cs_des(&fdest);
                  /*cleanup*/ cs_des(&ftmp);
               }
               else if (flags & FSF_OVERW) {
                  char* fdest = cs_cref("%s%s", fname_dest, ".tmp");
                     HandleErr(handle_err);
                  while (stat_file_exists (fdest))
                  {
                     cs_des(&fdest);
                     fdest = cs_cref("%s%s", fdest, ".tmp");
                        HandleErr(handle_err);
                  }
                  cp_l_N(fname_src, fdest, at_s, can_tar);
                     HandleErr(handle_err);

                  struct stat at_src = {0};
                  int res = lstat (fdest, &at_src);
                     ErrorTrigIf(res == -1, handle_err);

                  ds_rmdir2(fname_dest, at_d);
                     HandleErr(handle_err);
                  cp_l_N(fdest, fname_dest, &at_src, can_tar);
                     HandleErr(handle_err);
                  ds_rmfile2(fdest, &at_src);
                     HandleErr(handle_err);
                  cs_des(&fdest);
               } else {
                  ErrorSet(EEXIST, handle_err);
               }
               break;
            default: ___Sentinel("Wrong file type"); abort();
         }
         break;
      /*: source file is directory */
      case 'd':
         switch (typ_d) {
            case 'N':
               cp_d_N(fname_src, fname_dest, at_s, flags&FSF_ERR_CONT);
                  HandleErr(handle_err);
               break;
            case 'l':   /*- \/ THROUGH -*/
            case '-':
            case 's':
            case 'f':
            case 'b':
            case 'c':
               if (flags & FSF_OVERW) {
                  ds_rmfile2(fname_dest, at_d);
                     HandleErr(handle_err);
                  cp_d_N(fname_src, fname_dest, at_s, flags&FSF_ERR_CONT);
                     HandleErr(handle_err);
               } else {
                  ErrorSet(EEXIST, handle_err);
               }
               break;
            case 'd':
               if (flags & FSF_INTO_DIR) {
                  char* ftmp = cs_cre(fname_src);
                  char* fdest = cs_cref("%s/%s", fname_dest, basename(ftmp));
                  ds_copy2(fname_src, fdest, at_s, NULL, flags & ~FSF_INTO_DIR);
                     HandleErr(handle_err);
                     /*
                  if (flags & FSF_OVERW) {
                     ds_copy2(fname_src, fdest, at_s, NULL, FSF_OVERW);
                        HandleErr(handle_err);
                  } else {
                     cp_d_N(fname_src, fdest, at_s, flags&FSF_ERR_CONT);
                        HandleErr(handle_err);
                  }
                  */
                  /*cleanup*/ cs_des(&fdest);
                  /*cleanup*/ cs_des(&ftmp);
               }
               else if (flags & FSF_OVERW) {
                  char* fdest = cs_cref("%s%s", fname_dest, ".tmp");
                     HandleErr(handle_err);
                  while (stat_file_exists (fdest))
                  {
                     cs_des(&fdest);
                     fdest = cs_cref("%s%s", fdest, ".tmp");
                        HandleErr(handle_err);
                  }
                  cp_d_N(fname_src, fdest, at_s, flags & FSF_ERR_CONT);
                     HandleErr(handle_err);

                  struct stat at_src = {0};
                  int res = lstat (fdest, &at_src);
                     ErrorTrigIf(res == -1, handle_err);

                  ds_rmdir2(fname_dest, at_d);
                     HandleErr(handle_err);
                  cp_d_N(fdest, fname_dest, &at_src, flags & FSF_ERR_CONT);
                     HandleErr(handle_err);
                  ds_rmdir2(fdest, &at_src);
                     HandleErr(handle_err);
                  cs_des(&fdest);
               } else {
                  ErrorSet(EEXIST, handle_err);
               }
               break;
            default: ___Sentinel("Wrong file type"); abort();
         }
         break;
      /*: source file is regular file */
      case '-':
         switch (typ_d) {
            case 'N':
               cp_f_N(fname_src, fname_dest, at_s);
                  HandleErr(handle_err);
               break;
            case 'l':   /*- \/ THROUGH -*/
            case '-':
            case 's':
            case 'f':
            case 'b':
            case 'c':
               if (flags & FSF_OVERW) {
                  ds_rmfile2(fname_dest, at_d);
                     HandleErr(handle_err);
                  cp_f_N(fname_src, fname_dest, at_s);
                     HandleErr(handle_err);
               } else {
                  ErrorSet(EEXIST, handle_err);
               }
               break;
            case 'd':
               if (flags & FSF_INTO_DIR) {
                  char* ftmp = cs_cre(fname_src);
                  char* fdest = cs_cref("%s/%s", fname_dest, basename(ftmp));
                  ds_copy2(fname_src, fdest, at_s, NULL, flags & ~FSF_INTO_DIR);
                     HandleErr(handle_err);
                     /*
                  if (flags & FSF_OVERW) {
                     ds_copy2(fname_src, fdest, at_s, NULL, FSF_OVERW);
                        HandleErr(handle_err);
                  } else {
                     cp_f_N(fname_src, fdest, at_s);
                        HandleErr(handle_err);
                  }
                  */
                  /*cleanup*/ cs_des(&fdest);
                  /*cleanup*/ cs_des(&ftmp);
               } else if (flags & FSF_OVERW) {
                  char* fdest = cs_cref("%s%s", fname_dest, ".tmp");
                     HandleErr(handle_err);
                  printf("dest=%s\n", fdest);
                  while (stat_file_exists (fdest))
                  {
                     cs_des(&fdest);
                     fdest = cs_cref("%s%s", fdest, ".tmp");
                        HandleErr(handle_err);
                  }
                  printf("dest=%s\n", fdest);
                  cp_f_N(fname_src, fdest, at_s);
                     HandleErr(handle_err);

                  struct stat at_src = {0};
                  int res = lstat (fdest, &at_src);
                     ErrorTrigIf(res == -1, handle_err);

                  ds_rmdir2(fname_dest, at_d);
                     HandleErr(handle_err);
                  cp_f_N(fdest, fname_dest, &at_src);
                     HandleErr(handle_err);
                  ds_rmfile2(fdest, &at_src);
                     HandleErr(handle_err);
                  cs_des(&fdest);
               } else {
                  ErrorSet(EEXIST, handle_err);
               }
               break;
            default: ___Sentinel("Wrong file type"); abort();
         }
         break;
      /* copying of this file types is not implemented yet */
      case 's':   /*- \/ THROUGH -*/
      case 'f':
      case 'b':
      case 'c':
         ErrorSet(E_FUNC_NOTIMPL, handle_err);
         break;
      default: ___Sentinel("Wrong file type"); abort();
   }

   if (!stat_file_exists(fname_src)) {
      at_s->st_ino = 0;    //for ds_move() - to avoid removing non existant source
   }

   return;

  handle_err:
   return;
}

/* :: Both files can be of any type.
 *    Default behavior if destination file exist is to fail with EEXIST.
 *    Default behavior if destination is directory is to fail with EEXIST.
 * ::Flags:
 *    FSF_OVERW      -- overwrite target if exist (also directory)
 *    FSF_INTO_DIR   -- move into directory if destination is directory, has precedence over FSF_OVERW
 * ::Note:
 *    If destination is directory and FSF_OVERW flag is specified directory
 *    will be removed. But if FSF_INTO_DIR flag is specified, file will be
 *    moved into directory regardless FSF_OVERW is set or not.
 * ::Preconditions:
 * ::Errors:
 *    Nothing moved or directories may be partialy moved and some temporary
 *    files may be left unlinked.
 *    many errors
*/
void
ds_move (
   const char* old_file,      //! old file name
   const char* new_file,      //! new file name
   int flags                  //! file system operations flags
){
   ___Dassert(old_file, "'old_file' parameter invalid");
   ___Dassert(new_file, "'new_file' parameter invalid");
   ds_move2(old_file, new_file, NULL, NULL, flags);
}

/* :: Same as 'ds_move()' except:
 *    caller may supply attributes of one or both of files - they will be used
 *    internally instead of retrieving them.
*/
void
ds_move2 (
   const char* old_file,      //! old file name
   const char* new_file,      //! new file name
   struct stat* at_o,   //! old entry attributes for internal checks that file is same as original, may be NULL
   const struct stat* at_n,   //! new entry attributes for internal checks that file is same as original, may be NULL
   int flags                  //! file system operations flags
){
   ___Dassert(old_file, "'old_file' parameter invalid");
   ___Dassert(new_file, "'new_file' parameter invalid");

   struct stat at_o2 = {0};
   struct stat at_n2 = {0};

   /*: Retrieve involved files attributes if not provided with fun call */
   if (at_o == NULL) {
      int res = lstat (old_file, &at_o2);
         ErrorTrigIf(res == -1, handle_err);        //src nonexist - error
      at_o = &at_o2;
   }

   if (at_n == NULL) {
      int res = lstat (new_file, &at_n2);
         if (res == -1 && errno == ENOENT) {
            errno = 0;
         } else {
            ErrorTrigIf(res == -1, handle_err);
            at_n = &at_n2;
         }
   }

   /*: Perform move: copy first, then remove entry */
   //printf("%s->%s\n", old_file, new_file);
   //getchar();
   ds_copy2(old_file, new_file, at_o, at_n, flags);
      HandleErr(handle_err);
   //printf("%s->%s\n", old_file, new_file);
   //getchar();
   if (at_o->st_ino != 0) {   //ds_copy2() sets this to 0 if file does not exist anymore
      ds_rmentry2(old_file, at_o);     //due to itself overwriten its parent
         HandleErr(handle_err);
   }

   /* Finish: */
   return;

  handle_err:
   return;
}
