/* This file is part of program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MHCL_FS_SRCDIRS_DS_OPERATE_AUX_H_INC
#define MHCL_FS_SRCDIRS_DS_OPERATE_AUX_H_INC


/*!F::Operations on directories - auxillary file.
 *
 *  ::Provides helper functions for main implementation file.
 */




/* employs includes from ds_operate.c */




/* :: Compute size of libc entry structure 'struct dirent' for given directory
 *    stream. Size depends on file system entry name maximum length.
 * ::Preconditions:
 * ::Errors:
 *    Nothing moved.
 *    ENOMEM      -- no memory available
 *    rename()    -- this libc function errors
 *
*/
static inline
void
move_entry_to_temp_dir (
   str_t* tentry_name,        //! [out] temporary entry name
   const char* ename,         //! entry to move
   const char* tdir_name      //! temporary directory name
){
   char* tename = cs_cref ("%s/e.toremove", tdir_name);  //: create temporary entry name
      HandleErr(handle_err);
   int res = rename(ename, tename);                      //: move
      ErrorTrigIfm(res == -1, handle_err, "unable move '%s' to temporary directory", ename);
   str_fill(tentry_name, tename);      //: inform caller about new entry name

   /*: Finish: */
   return;

  handle_err:
   cs_des(&tename);
   return;
}

/* :: Compare inode numbers to check if attributes belong to same file. Compare
 *    file types if required too.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
*/
static inline
bool                       //! 'true' if files are same
files_same (
   const struct stat* fA,  //! attrs of 1st file
   const struct stat* fB,  //! attrs of 2nd file
   mode_t f_type           //! if not 0: check if both files are of specified type
){
   ___Dassert(fA, "'fA' parameter invalid");
   ___Dassert(fB, "'fB' parameter invalid");

   /*: Check types if required: */
   if (f_type != 0) {
      if ((fA->st_mode & S_IFMT) != f_type) { return false; }
      if ((fB->st_mode & S_IFMT) != f_type) { return false; }
   }
   /*: Check inodes: */
   return (fA->st_dev == fB->st_dev && fA->st_ino == fB->st_ino);
}

/* :: Open file, get attributes, prepare entry struct.
 * ::Preconditions:
 * ::Errors:
 *    Nothing opened and initialized.
 *    ENOMEM      -- no memory available
 *    open()      -- this libc function errors
 *    fstat()     -- this libc function errors
*/
static inline
void
initiate_source_file (
   st_ent* stFile,         //! [out] place for file info
   const char* fname       //! file name
){
   ___Dassert(stFile, "'stFile' parameter invalid");
   ___Dassert(fname, "'fname' parameter invalid");

   /*: Open file: */
   int fd = open(fname, O_RDONLY | O_NOFOLLOW); // open source file
      ErrorTrigIfm(fd == -1, handle_err, "unable open source file '%s'", fname);

   /*: Get attributes: */
   struct stat as = {0};
   int res = fstat(fd, &as);
      ErrorTrigIfm(res == -1, handle_err, "unable read attributes of source file '%s'", fname);

   /*: Set entry struct members: */
   ent_setall_(stFile, NULL, &as, NULL, fname, NULL, fd, NULL);
      HandleErr(handle_err);

   /*: Finish: */
   return;

  handle_err:
   if (fd > -1) { close(fd); }
   return;
}

/* :: Close file descriptor, finalize entry struct.
 * ::Preconditions:
 * ::Errors:
 *    close()        -- this libc function errors
*/
static inline
void
finalize_file (
   st_ent* stFile       //! file (entry struct)
){
   ___Dassert(stFile, "'stFile' parameter invalid");

   int ret = close(stFile->fd);
      ErrorTrigIf(ret == -1, handle_err);

   ent_finish_(stFile);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* :: Create temporary directory for file in file's parent directory. Temporary
 *    directory is in format 'parent_dir/XXXXXX'.
 * ::Preconditions:
 * ::Errors:
 *    No directory created.
 *    ENOMEM      -- no memory available
 *    dirname()   -- this libc function errors
 *    mkdtemp()   -- this libc function errors
*/
static inline
void
create_temp_dir_for_file (
   str_t* tdir,               //! [out] created temporary dir name
   const char* fname_dest     //! destination file name
){
   ___Dassert(tdir, "'tdir' parameter invalid");
   ___Dassert(fname_dest, "'fname_dest' parameter invalid");

   char* cs_dir = NULL;
   char* res = NULL;

   /*: Figure out parent directory of destination file. */
   char* cs = cs_cre(fname_dest);// temporary cstring with dest file name (dirname may modify path)
      HandleErr(handle_err);
   cs_dir = dirname(cs);

   /*: Create temporary directory: */
   cs_dir = cs_cref ("%s/XXXXXX", cs_dir);   //: create name in proper format
      HandleErr(handle_err);
   /*cleanup*/ cs_des(&cs);
   res = mkdtemp(cs_dir);
      ErrorTrigIfm(res == NULL, handle_err, "failed create temporary directory");

   /*: Inform caller about dir name: */
   str_fill(tdir, cs_dir);

   /*: Finish: */
   return;

  handle_err:
   if (cs != NULL)  { cs_des(&cs); }
   if (cs_dir != NULL)  { cs_des(&cs_dir); }
   // not needed: if (res != NULL) { rmdir(cs_dir); }
}

/* :: Create 'f.toremove' file in specified directory.
 * ::Preconditions:
 *    - directory where file have to be created must exist
 * ::Errors:
 *    No file created.
 *    ENOMEM      -- no memory available
 *    open()      -- this libc function errors
*/
static inline
void
create_temp_dest_file (
   st_ent* stFile,         //! [out] file's entry struct
   const char* tdir_name   //! directory name where to create file
){
   ___Dassert(stFile, "'stFile' parameter invalid");
   ___Dassert(tdir_name, "'tdir_name' parameter invalid");

   /*: Compose temporary file name. */
   int fd = -1;
   char* tfname = cs_cref ("%s/f.toremove", tdir_name);
      HandleErr(handle_err);

   /*: Open temp file and set entry struct: */
   fd = open(tfname, O_WRONLY | O_CREAT | O_EXCL, S_IWUSR);  // temp dest with minimal perms
      ErrorTrigIfm(fd == -1, handle_err, "unable create temporary file '%s'", tfname);

   ent_setall_(stFile, NULL, NULL, NULL, tfname, NULL, fd, NULL);
      HandleErr(handle_err);

   /*cleanup*/ cs_des(&tfname);

   /*: Finish: */
   return;

  handle_err:
   if (fd > -1) {
      close(fd);
      int res = unlink(tfname);
      if (res == -1) {
         err_append_msg(", unable to unlink temporary file '%s'", tfname);
      }
   }
   if (tfname != NULL) { cs_des(&tfname); }
   return;
}

/* :: Remove specified directory.
 * ::Preconditions:
 *    - directory where file have to be created must exist
 * ::Errors:
 *    No directory removed.
 *    rmdir()        -- this libc function errors
*/
static inline
void
remove_temp_dir_for_file (
   str_t* tempdir_name
){
   ___Dassert(tempdir_name, "'tempdir_name' parameter invalid");

   int res = rmdir(tempdir_name->buf);
      ErrorTrigIfm(res == -1, handle_err, "failed remove temporary directory '%s'", tempdir_name->buf);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* :: Copy content and permissions of source file to destination file.
 * ::Preconditions:
 * ::Errors:
 *    Part or nothing copied.
 *    read()      -- this libc function errors
 *    write()     -- this libc function errors
 *    fchmod()    -- this libc function errors
*/
static inline
uint                          //! 0-all ok, 1-error copying content 2-error copying permissions
copy_file_content_and_perms (
   st_ent* stFileDest,        //! destination file struct
   st_ent* stFileSrc          //! source file struct
){
   ___Dassert(stFileDest, "'stFileDest' parameter invalid");
   ___Dassert(stFileSrc, "'stFileSrc' parameter invalid");

   char buf[4096];      //: buffer for content
   /*: Copy content: */
   while (true)
   {
      /*: Read source file content to buffer: */
      ssize_t nbr = read(stFileSrc->fd, buf, sizeof buf);
         ErrorTrigIfm(nbr == -1 && errno != EINTR, handle_err_cont, "unable to complete file content copying");
      if (nbr == -1) { continue; }  //EINTR - we retry
      if (nbr == 0) { break; }

      /*: Write buffer to destination file: */
      char* pbuf = buf;
      while (true)
      {
         ssize_t nbw = write (stFileDest->fd, pbuf, nbr);
            ErrorTrigIfm(nbw == -1 && errno != EINTR, handle_err_cont, "unable to complete file content copying");
         if (nbr == -1) { continue; }  //EINTR - we retry
         nbr -= nbw;
         if (nbr <= 0) { break; }
         pbuf += nbw;
      }
   }

   /*: Copy permissions: */
   int res = fchmod(stFileDest->fd, stFileSrc->attrs.st_mode);
      ErrorTrigIfm(res == -1, handle_err_perms, "cannot set permissions on destination file");

   /*: Finish: */
   return 0;

  handle_err_cont:
   return 1;
  handle_err_perms:
   return 2;
}

/* :: Move temporary file entry to proper place.
 * ::Preconditions:
 * ::Errors:
 *    Nothing moved or old file not unlinked.
 *    link()      -- this libc function errors
 *    unlink()    -- this libc function errors
*/
static inline
void
move_temp_file_to_dest (
   const char* fname_dest, //! destination file name
   st_ent* stFileTemp      //! source (temp) file
){
   ___Dassert(fname_dest, "'fname_dest' parameter invalid");
   ___Dassert(stFileTemp, "'stFileTemp' parameter invalid");

   /*: Create new link and remove old link: */
   char* cs = stFileTemp->filename.buf;
   int res = link(cs, fname_dest);
      ErrorTrigIfm(res == -1, handle_err, "failed move temporary file '%s'", cs);

   res = unlink(cs);
      ErrorTrigIfm(res == -1, handle_err, "failed unlink temporary file '%s'", cs);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* :: Copy regular file to non-existent entry.
 * ::Preconditions:
 *    - source file must exist and be regular file
 *    - destination file must not exist
 * ::Errors:
 *    Nothing copied or copied but temporary dir and/or file not unlinked.
 *    Errors are written to message in '__err'.
 *    E_STATE_CHANGE       -- source file replaced during operation
 *    many other errors
*/
static inline
void
cp_f_N (
   const char* fname_src,  //! source file name
   const char* ename_dest, //! destination name
   const struct stat* at_s //! source file attributes - internal checks that file was not replaced
){
   ___Dassert(fname_src, "'fname_src' parameter invalid");
   ___Dassert(ename_dest, "'ename_dest' parameter invalid");
   ___Dassert(at_s, "'at_s' parameter invalid");

   st_ent stFileSrc = {0};
   str_t tempdir_name = {0};
   st_ent stFileTemp = {0};

   /*: Prepare source file: */
   initiate_source_file(&stFileSrc, fname_src);          //open+fstat
      HandleErr(handle_err);
   //: make sure source file was not replaced
   if (!files_same(at_s, &stFileSrc.attrs, S_IFREG)) {
      ErrorSetm(E_STATE_CHANGE, handle_err, "source file '%s' replaced during operation", fname_src);
   }

   /*: Create tmp destination and copy source file there: */
   create_temp_dir_for_file(&tempdir_name, ename_dest);  //mkdtemp
      HandleErr(handle_err);
   create_temp_dest_file(&stFileTemp, tempdir_name.buf); //open
      HandleErr(handle_err);
   copy_file_content_and_perms(&stFileTemp, &stFileSrc); //read+write+fchmod
      HandleErr(handle_err);
   /*cleanup*/ finalize_file(&stFileSrc);
                  //no error check - fds will be auto closed at process termination

   /*: Move to destination and cleanup: */
   move_temp_file_to_dest(ename_dest, &stFileTemp);      //link+unlink
      HandleErr(handle_err);
   /*cleanup*/ finalize_file(&stFileTemp);
   remove_temp_dir_for_file(&tempdir_name);              //rmdir
      HandleErr(handle_err);
   /*cleanup*/ str_fin(&tempdir_name);

   /*: Finish: */
   return;

  handle_err:
   if (stFileSrc.init == true) { ent_finish_(&stFileSrc); }
   if (stFileTemp.init == true) {
      int res = unlink(stFileTemp.filename.buf);
      if (res == -1) {
         err_append_msg(", unable to unlink temporary file '%s'", stFileTemp.filename.buf);
      }
      ent_finish_(&stFileTemp);
   }
   if (tempdir_name.nc > 0) {
      int res = rmdir(tempdir_name.buf);
      if (res == -1) {
         err_append_msg(", unable to remove temporary directory '%s'", tempdir_name.buf);
      }
   }
   str_fin(&tempdir_name);
   return;
}

/* :: Create temporary link for file.
 * ::Preconditions:
 * ::Errors:
 *    Link may be created but entry struct may not be. Errors are written to
 *    message in '__err'.
 *    ENOMEM      -- no memory available
 *    link()      -- this libc function errors
 *    unlink()    -- this libc function errors
 *    lstat()     -- this libc function errors
*/
static inline
void
prepare_temp_link (
   st_ent* stLinkTemp,     //! [out] entry struct for temporary link
   const char* fname_src,  //! file to link to
   const char* tdir_name,  //! temporary directory for link
   bool canon_tar          //! canonicalize target before symlinking? - makes target absolute path
){
   ___Dassert(fname_src, "'fname_src' parameter invalid");
   ___Dassert(tdir_name, "'tdir_name' parameter invalid");

   int lcreated = -1;
   char* tar = NULL;

   /*: Create link: */
   char* tlname = cs_cref ("%s/hl.toremove", tdir_name);  //temp link name
      HandleErr(handle_err);

   lcreated = link(fname_src, tlname);
      ErrorTrigIfm(lcreated == -1, handle_err, "failed create temporary link '%s'", tlname);

   struct stat attrs = {0};
   int res = lstat (tlname, &attrs);
      ErrorTrigIfm(res == -1, handle_err, "failed read '%s' attributes", tlname);

   /*: Canonicalize link's name if required: */
   if (canon_tar) {
      stat_absolutepath (&tar, fname_src);
         HandleErrm(handle_err, "failed read '%s' attributes", tlname);
   } else {
      stat_readlink(&tar, attrs.st_blksize, tlname);
         HandleErrm(handle_err, "failed read '%s' attributes", tlname);
   }

   /*: Set entry struct: */
   ent_setall_(stLinkTemp, NULL, &attrs, NULL, tlname, tar, -1, NULL);
      HandleErr(handle_err);

   /*cleanup*/ cs_des(&tlname);
   /*cleanup*/ cs_des(&tar);

   /*: Finish: */
   return;

  handle_err:
   if (tar != NULL) { cs_des(&tar); }
   if (lcreated > -1) {
      int res2 = unlink(tlname);
         ErrorTrigIfm(res2 == -1, handle_err, "unable unlink temporary file '%s'", tlname);
   }
   if (tlname != NULL) { cs_des(&tlname); }
   return;
}

/* :: Copy symbolic link to non-existent entry.
 * ::Preconditions:
 *    - source file must exist and be symlink
 *    - destination file must not exist
 * ::Errors:
 *    Nothing copied or copied but temporary dir and/or file not unlinked.
 *    Errors are written to message in '__err'.
 *    E_STATE_CHANGE       -- source file replaced during operation
 *    many other errors
*/
static inline
void
cp_l_N (
   const char* lname_src,  //! source symlink name
   const char* ename_dest, //! destination name
   const struct stat* at_s,//! source file attributes - internal checks that file was not replaced
   bool canon_tar          //! canonicalize target before symlinking? - makes target absolute path
){
   ___Dassert(lname_src, "'lname_src' parameter invalid");
   ___Dassert(ename_dest, "'ename_dest' parameter invalid");
   ___Dassert(at_s, "'at_s' parameter invalid");

   str_t tempdir_name = {0};
   st_ent stTmpLink = {0};

   /*: Create tmp directory and link source */
   create_temp_dir_for_file(&tempdir_name, lname_src);         //mkdtemp
      HandleErr(handle_err);
   prepare_temp_link(&stTmpLink, lname_src, tempdir_name.buf, canon_tar); //link+lstat+readlink
      HandleErr(handle_err);
   //: make sure source file was not replaced
   if (!files_same(at_s, &stTmpLink.attrs, S_IFLNK)) {
      ErrorSetm(E_STATE_CHANGE, handle_err, "source file '%s' replaced during operation", lname_src);
   }

   /*: Create symlink */
   int res = symlink(stTmpLink.target.buf, ename_dest);        //symlink
      ErrorTrigIfm(res == -1, handle_err, "failed create symlink '%s'", ename_dest);

   /*: Cleanup temporary dir */
   res = unlink(stTmpLink.filename.buf);                       //unlink
      ErrorTrigIfm(res == -1, handle_err, "unable unlink '%s'", stTmpLink.filename.buf);
   /*cleanup*/ ent_finish_(&stTmpLink);
                  //no error check - fds will be auto closed at process termination
   remove_temp_dir_for_file(&tempdir_name);                    //rmdir
      HandleErr(handle_err);
   /*cleanup*/ str_fin(&tempdir_name);

   /*: Finish: */
   return;

  handle_err:
   if (stTmpLink.init == true) {
      int res2 = unlink(stTmpLink.filename.buf);
      if (res2 == -1) {
         err_append_msg(", unable to unlink temporary file '%s'", stTmpLink.filename.buf);
      }
      ent_finish_(&stTmpLink);
   }
   if (tempdir_name.nc > 0) {
      int res2 = rmdir(tempdir_name.buf);
      if (res2 == -1) {
         err_append_msg(", unable to remove temporary directory '%s'", tempdir_name.buf);
      }
   }
   str_fin(&tempdir_name);
   return;
}

/* :: Initiate entry struct with specified directory. Open stream and get
 *    attributes.
 * ::Preconditions:
 * ::Errors:
 *    Directory not opened, struct not initiated.
 *    ENOMEM      -- not enough memory available
 *    opendir()      -- this libc function errors
 *    fstat()        -- this libc function errors
 *    stat()         -- this libc function errors
*/
static inline
void
initiate_source_dir (
   st_ent* stDir,       //! [out] entry struct - place for directory info
   const char* dname    //! file name
){
   ___Dassert(stDir, "'stDir' parameter invalid");
   ___Dassert(dname, "'dname' parameter invalid");

   /*: Open directory: */
   DIR* ds = opendir(dname);     //: open stream
      ErrorTrigIfm(ds == NULL, handle_err, "failed open source directory '%s'", dname);
   int dd = dirfd(ds);           //: get descriptor

   /*: Get directory attributes: */
   struct stat as = {0};
   if (dd != -1) {
      int res = fstat(dd, &as);
         ErrorTrigIfm(res == -1, handle_err, "unable read attributes of source directory '%s'", dname);
   } else {
      int res = stat(dname, &as);
         ErrorTrigIfm(res == -1, handle_err, "unable read attributes of source directory '%s'", dname);
   }

   /*: Set entry struct: */
   ent_setall_(stDir, NULL, &as, NULL, dname, NULL, dd, ds);
      HandleErr(handle_err);

   /*: Finish: */
   return;

  handle_err:
   if (ds != NULL) { closedir(ds); }
   return;
}

/* :: Close dirstream and finish entry struct.
 * ::Preconditions:
 * ::Errors:
 *    Dirstream may be not closed bu it is ignored.
 *    closedir()  -- this libc function errors
*/
static inline
void
finalize_dir (
   st_ent* stDir       //! directory - entry struct
){
   ___Dassert(stDir, "'stDir' parameter invalid");

   int ret = closedir(stDir->ds);
      ErrorTrigIf(ret == -1, handle_err);

   ent_finish_(stDir);

   return;

  handle_err:
   return;
}

/* :: Create 'd.toremove' directory in specified directory.
 * ::Preconditions:
 *    - directory where temprary destination directory have to be created must
 *      exist
 * ::Errors:
 *    No directory created.
 *    ENOMEM      -- no memory available
 *    mkdir()     -- this libc function errors
*/
static inline
void
create_temp_dest_dir (
   st_ent* stDir,          //! [out] temporary directory (entry struct)
   const char* tdir_name   //! directory that will contain temporary destination dir
){
   ___Dassert(stDir, "'stDir' parameter invalid");
   ___Dassert(tdir_name, "'tdir_name' parameter invalid");

   /*: Create directory: */
   int res = -1;
   char* tdname = cs_cref ("%s/d.toremove", tdir_name);     //temp dir name
      HandleErr(handle_err);
   res = mkdir (tdname, S_IRWXU);
      ErrorTrigIfm(res == -1, handle_err, "failed make temporary destination directory '%s'", tdname);

   /*: Set struct members: */
   ent_setall_(stDir, NULL, NULL, NULL, tdname, NULL, -1, NULL);
      HandleErr(handle_err);

   /*cleanup*/ cs_des(&tdname);

   /*: Finish: */
   return;

  handle_err:
   if (res != -1) {
      int res2 = rmdir(tdname);
      if (res2 == -1) {
         err_append_msg(", unable to remove temporary directory '%s'", tdname);
      }
   }
   if (tdname != NULL) { cs_des(&tdname); }
   return;
}

/* :: Copy directory link to non-existent entry.
 * ::Preconditions:
 *    - source dir must exist and be directory
 *    - destination entry must not exist
 *    - destination entry can't be subdirectory of source dir
 * ::Errors:
 *    Nothing copied or partialy copied or copied but temporary dir and/or file
 *    not unlinked
 *    Errors are written to message in '__err'.
 *    E_STATE_CHANGE       -- source file replaced during operation
 *    many other errors
*/
static inline
void
cp_d_N (
   const char* dname_src,  //! source directory name
   const char* ename_dest, //! destination name
   const struct stat* at_s,//! source file attributes - internal checks that file was not replaced
   bool err_cont
){
   ___Dassert(dname_src, "'dname_src' parameter invalid");
   ___Dassert(ename_dest, "'ename_dest' parameter invalid");
   ___Dassert(at_s, "'at_s' parameter invalid");

   st_ent stDirSrc = {0};
   str_t tempdir_name = {0};
   st_ent stTempDir = {0};

   /*: Prepare source directory */
   initiate_source_dir(&stDirSrc, dname_src);            //opendir+fstat/stat
      HandleErr(handle_err);
   //: make sure source file was not replaced
   if (!files_same(at_s, &stDirSrc.attrs, S_IFDIR)) {
      ErrorSetm(E_STATE_CHANGE, handle_err, "source directory '%s' replaced during operation", dname_src);
   }

   /*: Make tmp directory and create temp destination dir */
   create_temp_dir_for_file(&tempdir_name, ename_dest);  //mkdtemp
      HandleErr(handle_err);
   create_temp_dest_dir(&stTempDir, tempdir_name.buf);   //mkdir
      HandleErr(handle_err);

   /*: Copy dir entires: */
   DIR* ds = stDirSrc.ds;
   while (true)
   {
      //: Allocate entry struct
      st_ent* stEnt = ds_ent_alloc(ds);
         HandleErr(handle_err_while);

      //: Read next entry from source dir
      bool present = ds_read_entry(stEnt, ds, stDirSrc.filename.buf);
         HandleErr(handle_err_while);
      if (!present) { ds_ent_free(&stEnt); break; }   //: no more entries
      if (strcmp(stEnt->dent->d_name, "..") == 0      //: ignore . and ..
      || strcmp(stEnt->dent->d_name, ".") == 0) {
         ds_ent_free(&stEnt);
         continue;
      }

      //: Copy entry to destination dir
      char* entr_dest = cs_cref("%s/%s", stTempDir.filename.buf, stEnt->dent->d_name);
      ds_copy2(stEnt->filename.buf, entr_dest, NULL, NULL, 0);
         //: Return on error ONLY if continuation not required by caller
         if (!err_cont && errno != E_FUNC_NOTIMPL) {
            HandleErr(handle_err_while);
         }

      /*cleanup*/ cs_des(&entr_dest);
      /*cleanup*/ ds_ent_free(&stEnt);
      continue;

      /*: Loop error: */
     handle_err_while:
      if (stEnt != NULL) { ds_ent_free(&stEnt); }
      if (entr_dest != NULL) { cs_des(&entr_dest); }
      goto handle_err;
   }

   /*: Copy permissions: */
   int res = chmod(stTempDir.filename.buf, stDirSrc.attrs.st_mode);
      ErrorTrigIfm(res == -1, handle_err, "cannot set permissions on destination directory");
   /*cleanup*/ finalize_dir(&stDirSrc);
                  //no error check - fds will be auto closed at process termination

   /*: Move temp destination dir to proper destination */
   res = mkdir (ename_dest, 0);
      ErrorTrigIfm(res == -1, handle_err, "failed make dummy destination directory '%s'", ename_dest);
   res = rename(stTempDir.filename.buf, ename_dest);
      ErrorTrigIfm(res == -1, handle_err, "failed move temporary directory to proper destination '%s'", stTempDir.filename.buf);

   /*: Cleanup temporary dir */
   /*cleanup*/ ent_finish_(&stTempDir);
                  //no error check - fds will be auto closed at process termination
   remove_temp_dir_for_file(&tempdir_name);                    //rmdir
      HandleErr(handle_err);
   /*cleanup*/ str_fin(&tempdir_name);

   /*: Finish: */
   return;

  handle_err:
   if (stTempDir.init == true) {
      int res2 = rmdir(stTempDir.filename.buf);
      if (res2 == -1) {
         err_append_msg(", unable to remove temporary destination directory '%s'", stTempDir.filename.buf);
      }
      ent_finish_(&stTempDir);
   }
   if (tempdir_name.nc > 0) {
      int res2 = rmdir(tempdir_name.buf);
      if (res2 == -1) {
         err_append_msg(", unable to remove temporary directory '%s'", tempdir_name.buf);
      }
      str_fin(&tempdir_name);
   }
   if (stDirSrc.init == true) {
      ent_finish_(&stDirSrc);
   }
   return;
}




#endif /*- MHCL_FS_SRCDIRS_DS_OPERATE_AUX_H_INC -*/
