/*!F::File stream read functions - main implementation file.
 *
 *  ::Purpose:
 *    Various functions for reading data from file streams.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-02
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* getline(), getdelim():
 *       _POSIX_C_SOURCE >= 200809L || _XOPEN_SOURCE >= 700
 */
#define _POSIX_C_SOURCE 200809L




/* current project */
#include "common_mhlfs.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"

/* libc */
#include <errno.h>      // errno
#include <stdio.h>      // getline(), printf(), putc()

/* current module */
#include "fstream_read.h"
//#include "fstream_read_internal.h.inc"
//#include "fstream_read_aux.h.inc"




/*!:: Return subsequent lines read from file stream, place number of bytes read
 *    in `nb_r`.
 *    Line is read upto '\n' or EOF, including '\n'.
 *    Line is terminated with '\0'.
 *    If EOF is encountered NULL is returned and 0 is placed in `nb_r`.
 * ::Preconditions:
 * ::Errors:
 *    Return NULL, place 0 in `nb_r`, set error in '__err'.
 *    ENOMEM      -- no memory available
 *    chown()     -- this libc function errors
*/
char*             //! line read
fs_rline (
   FILE* fs,    //! file stream
   size_t* nb_r   //! [out] number of bytes read
){
   ___Dassert(fs != NULL, "'fs parameter invalid");

   char* line = NULL;
   size_t nb = 0;
   errno = 0;
   //line=NULL and size=0 => allocate buf inside getline
   ssize_t res = getline(&line, &nb, fs);
      ErrorTrigIf(errno != 0, handle_err);
      if (res == -1) { mem_free(line); res = 0; goto handle_succ; } // EOF

   /*: Finish: */
  handle_succ:
   if (nb_r != NULL) { *nb_r = res; } //: inform caller about number of Bytes read
   return line;

  handle_err:
   mem_free(line);
   if (nb_r != NULL) { *nb_r = 0; } //: inform caller about number of Bytes read
   return NULL;
}
