#ifndef MHCL_FS_SRCMAIN_FS_STORAGE_H
#define MHCL_FS_SRCMAIN_FS_STORAGE_H


/*!F::Common storage for file system routines - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* current project */
#include "common_mhlfs.h"
#include "dstream_basic.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"
#include "strings.h"

/* libc */
#include <string.h>     // memcpy()
#include <dirent.h>     // DIR, struct dirent
#include <sys/stat.h>   // struct stat




/*!:: File system operations flags.
 */
enum fs_flags_en {
   FSF_NONE = 0,           //! no flags
   FSF_RSL_SRC = 1,        //! resolve symlinks for source path - NOT IMPL
   FSF_RSL_TAR = 2,        //! resolve symlinks for destination path - NOT IMPL
   FSF_OVERW = 4,          //! if target exist - overwrite instead error
   FSF_INTO_DIR = 8,       //! if target is DIR append source filename to it (move/copy into dir)
   FSF_CANON_TARGET = 16,  //! canonicalize (make abs path) target
   FSF_ERR_CONT = 32,      //! try continuing operation on errors
   fs_flags_count };




typedef struct st_ent st_ent;
/*!:: Entry structure.
 *    File system entry in-program representation.
 */
struct st_ent
{
   bool init;
   struct dirent* dent;    //! libc entry struct
   struct stat attrs;      //! libc file statistics struct
   str_t entryname;        //! entry name
   str_t filename;         //! full file/dirname (relative or absolute)
   str_t target;           //! target if entry is symlink
   int fd;                 //! file descriptor
   DIR* ds;                //! directory stream (for file it may be parent dir)
};

typedef struct st_dir st_dir;
/*!:: Directory structure.
 *    Directory in-program representation.
 */
struct st_dir
{
   st_ent this_entry;//! this directory entry
   st_ent** Dentrs;  //! array of pointers to entry structures
   size_t nels;      //! number of elements
};




static inline void ent_zero_ (st_ent* stEnt)
                     __attr_nonnull__(1)
                     ;

static inline void ent_init_ (st_ent* stEnt, size_t dirent_size)
                     __attr_nonnull__(1)
                     ;

static inline void ent_finish_ (st_ent* stEnt)
                     __attr_nonnull__(1)
                     ;

static inline void ent_setall_ (st_ent* stEnt,
               struct dirent* dent, const struct stat* attrs,
               const char* entryname, const char* filename, const char* target,
               int fd, DIR* ds)
                     __attr_nonnull__(1)
                     ;




static inline void dir_zero_ (st_dir* stDirEntrs)
                     __attr_nonnull__(1)
                     ;

static inline void dir_finish_ (st_dir* stDirectory)
                     __attr_nonnull__(1)
                     ;

static inline void dir_expand_ (st_dir* stDir, size_t nels_add)
                     __attr_nonnull__(1)
                     ;




/*!1:: Initialize entry struct.
 * :: Set members of struct to empty values.
 */
static inline
void
ent_zero_ (
   st_ent* stEnt      //! entry structure
){
   ___Dassert(stEnt, "'pstEnt' parameter invalid");

   /*: Set members. */
   stEnt->init = false;
   stEnt->dent = NULL;
   stEnt->attrs = (struct stat){0};
   stEnt->entryname = (str_t){0};
   stEnt->filename = (str_t){0};
   stEnt->target = (str_t){0};
   stEnt->fd = -1;
   stEnt->ds = NULL;

   /*: Finish: */
   return;
}

/*!:: Initialize entry structure, allocate memory for its members, particulary
 *    libc entry struct ('struct dirent').
 * ::Errors:
 *    Nothing created.
 *    ENOMEM      -- no memory available
*/
void
ent_init_ (
   st_ent* stEnt,       //! [out] entry structure
   size_t dirent_size   //! size of 'dirent struct'
){
   ___Dassert(stEnt, "'stEnt' parameter invalid");

   /*: Allocate libc entry structure. */
   struct dirent* dent = mem_alloc(dirent_size);
      HandleErr(handle_err);
   *dent = (struct dirent){0};

   stEnt->dent = dent;

   /*: Finish: */
   return;

  handle_err:
   mem_free(dent);
   return;
}

/*!1:: Finalize entry struct.
 * :: Release resources inside struct.
 */
static inline
void
ent_finish_ (
   st_ent* stEnt      //! entry structure
){
   ___Dassert(stEnt, "'pstEnt' parameter invalid");

   /*: Release resources inside struct. */
   stEnt->init = false;
   mem_free(stEnt->dent);
   stEnt->attrs = (struct stat){0};
   str_fin(&stEnt->entryname);
   str_fin(&stEnt->filename);
   str_fin(&stEnt->target);
   stEnt->fd = -1;
   stEnt->ds = NULL;

   /*: Finish: */
   return;
}

/*!1:: Set entry struct.
 *  .. Set all members.
 * ::Errors:
 *    ENOMEM      -- not enough memory available
 */
static inline
void
ent_setall_ (
   st_ent* stEnt,     //! entry structure
   struct dirent* dent,
   const struct stat* attrs,
   const char* entryname,
   const char* filename,
   const char* target,
   int fd,
   DIR* ds
){
   ___Dassert(stEnt, "'stEnt' parameter invalid");

   /*: Set members. */
   stEnt->init = true;
   stEnt->dent = dent;
   if (attrs != NULL) { stEnt->attrs = *attrs; }
   if (entryname != NULL) { str_ini(&stEnt->entryname, entryname); }
   if (filename != NULL) { str_ini(&stEnt->filename, filename); }
   if (target != NULL) { str_ini(&stEnt->target, target); }
      HandleErr(handle_err);
   stEnt->fd = fd;
   stEnt->ds = ds;

   /*: Finish: */
   return;

  handle_err:
   return;
}




/*!1:: Zero directory struct.
 * :: Set all struct members to empty values.
 */
static inline
void
dir_zero_ (
   st_dir* stDir     //! directory struct
){
   ___Dassert(stDir, "'stDir' parameter invalid");
   stDir->this_entry = (st_ent){0};
   stDir->Dentrs = NULL;
   stDir->nels = 0;
   return;
}

/*!1:: Finalize directory struct.
 * :: Destroy every entry struct inside, close dir stream, release other
 *    resources and zero all data.
 */
static inline
void
dir_finish_ (
   st_dir* stDir     //! directory struct
){
   ___Dassert(stDir, "'stDir' parameter invalid");
   for (size_t i = 0; i < stDir->nels; ++i) {
      ent_finish_(stDir->Dentrs[i]);
      mem_free(stDir->Dentrs[i]);
   }
   mem_free(stDir->this_entry.dent);
   ent_finish_(&stDir->this_entry);
   if (stDir->this_entry.ds != NULL) {
      ds_close(&stDir->this_entry.ds);
   }
   mem_free(stDir->Dentrs);
   dir_zero_(stDir);
   return;
}

/*!1:: Expand array of entries inside directory struct.
 *  .. Previous content is untouched.
 * ::Errors:
 *    Nothing created.
 *    ENOMEM         -- no memory available
 *    E_OVERFLOWINT  -- integer overflow - too many elements to add
 */
static inline
void
dir_expand_ (
   st_dir* stDir,    //! directory struct
   size_t nels_add   //! number of elements to add
){
   ___Dassert(stDir, "'stDir parameter invalid");
   ___Dassert(nels_add > 0, "'nels_add' parameter invalid");

   st_ent** Dentrs_new = NULL;
   size_t nels_new = stDir->nels + nels_add;

   size_t nb_pst = sizeof(*Dentrs_new);
      // overflow protection
   OverflowIntIf(nb_pst != 0 && nels_new > SIZE_MAX/nb_pst, handle_err);
   size_t nb = nels_new * nb_pst;

   Dentrs_new = mem_alloc(nb);
      HandleErr(handle_err);

   nb = stDir->nels * nb_pst;
   memcpy(Dentrs_new, stDir->Dentrs, nb);

   mem_free(stDir->Dentrs);
   stDir->Dentrs = Dentrs_new;
   stDir->nels = nels_new;

   goto handle_succ;

  handle_succ:
   return;

  handle_err:
   mem_free(Dentrs_new);
   return;
}




#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool fsstor_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- MHCL_FS_SRCMAIN_FS_STORAGE_H -*/
