/*!F::File statistics change - main implementation file.
 *
 *  ::Purpose:
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-02
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */


//for S_ISVTX: which is avail only on BSD and derived systems
//#define _BSD_SOURCE


#if defined(BSD)
#  define _BSD_SOURCE
#else
#  define _POSIX_C_SOURCE 200112L
#endif




/* current project */
#include "common_mhlfs.h"
#include "stat_info.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"

/* libc */
#include <sys/stat.h>   // chmod()
#include <unistd.h>     // chown()

/* current module */
#include "stat_change.h"
#include "stat_change_aux.h.inc"




/*!:: Use 'mode_t' type or c-string in format "rwxrwxrwx".
 * ::Preconditions:
 * ::Errors:
 *    EINVAL      -- `cs_mode` invalid (not in format "rwxrwxrwx")
 *    chmod()     -- this libc function errors
*/
void
stat_chmod (
   const char* file,    //! file name
   mode_t fmode,        //! mode as 'mode_t' type
   const char* cs_mode  //! mode in "rwxrwxrwx" format (precedence over `fmode`)
){
   ___Dassert(file, "'file' parameter invalid");

   /*: Compute mode: */
   mode_t m = fmode;
   if (cs_mode != NULL) {
      m = compute_mode_(cs_mode);
   }

   /*: Change mode: */
   int ret = chmod(file, m);
      ErrorTrigIf(ret == -1, handle_err);

   /* Finish: */
   return;

  handle_err:
   return;
}

/*!:: Use user id as 'uid_t' type or c-string with user name.
 * ::Preconditions:
 * ::Errors:
 *    ENOMEM      -- no memory available
 *    chown()     -- this libc function errors
*/
void
stat_chusr (
   const char* file,    //! file name
   uid_t uid,           //! user id
   const char* cs_user  //! user name (precedence over `uid`)
){
   ___Dassert(file, "'file' parameter invalid");

   /* determine uid */
   uid_t u = uid;                   // use given uid
   if (cs_user != NULL) {
      u = stat_look_uid(cs_user);   // or look it up if user name given
         HandleErr(handle_err);
   }

   /* perform changing user */
   int ret = chown(file, u, -1);
      ErrorTrigIf(ret == -1, handle_err);

   /* Finish: */
   return;

  handle_err:
   return;
}

/*!:: Use group id as 'gid_t' type or c-string with group name.
 * ::Preconditions:
 * ::Errors:
 *    ENOMEM      -- no memory available
 *    chown()     -- this libc function errors
*/
void
stat_chgrp (
   const char* file,    //! file name
   gid_t gid,           //! group id
   const char* cs_group //! group name (precedence over `gid`)
){
   ___Dassert(file, "'file' parameter invalid");

   /* Determine gid */
   gid_t g = gid;                   // use given gid
   if (cs_group != NULL) {
      g = stat_look_gid(cs_group);  // or look it up if user name given
         HandleErr(handle_err);
   }

   /* Perform changing user */
   int ret = chown(file, -1, g);
      ErrorTrigIf(ret == -1, handle_err);

   /* Finish: */
   return;

  handle_err:
   return;
}

