/*!F::File statistics retrieve - main implementation file.
 *
 *  ::Purpose:
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *
 *  ::Limitations:
 *    - to implement: access and maybe other
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-02
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */

/* getpwuid_r(), getpwnam_r():
 *          _POSIX_C_SOURCE >= 1 || _XOPEN_SOURCE || _BSD_SOURCE
 *          || _SVID_SOURCE || _POSIX_SOURCE
 */

/* realpath():
 *    _BSD_SOURCE || _XOPEN_SOURCE >= 500 ||
 *    _XOPEN_SOURCE && _XOPEN_SOURCE_EXTENDED
 */

/* S_IF* constants:
 * S_ISLNK, S_ISSOCK macros:
 * lstat(), readlink():
 *          _BSD_SOURCE || _XOPEN_SOURCE >= 500
 *          || _XOPEN_SOURCE && _XOPEN_SOURCE_EXTENDED
 *          || / Since glibc 2.10: / _POSIX_C_SOURCE >= 200112L
 */
//#define _POSIX_C_SOURCE 200112L

//for S_ISVTX: which is avail only on BSD and derived systems
//#define _BSD_SOURCE

#if defined(BSD)
#  define _BSD_SOURCE
#else
#  define _POSIX_C_SOURCE 200112L
#endif

//for realpath - need to replace realpath
#define _XOPEN_SOURCE 500



/* current project */
#include "common_mhlfs.h"
#include "dstream_info.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"
#include "cstrings.h"

/* libc */
#include <errno.h>   // ERANGE
#include <unistd.h>  // sysconf()
#include <pwd.h>     // getpwuid()
#include <grp.h>     // getgrgid()
#include <sys/stat.h>   // stat(), struct stat
#include <stdlib.h>  // realpath()
#include <libgen.h>  // dirname()

/* current module */
#include "stat_info.h"




/* Test file. */
/* -------------------------------------------------------------------------- */


/* :: Return true if entry exist in file system.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
*/
bool                    //! true if exist
stat_file_exists (
   const char* fname    //! entry name
){
   ___Dassert(fname, "'fname parameter invalid");

   struct stat buffer;

   lstat(fname, &buffer);
   if (errno == ENOENT || errno == ENOTDIR) {
      errno = 0;
      return false;
   }
      ErrorTrigIf(errno != 0, handle_err);

   return true;

  handle_err:
   return false;
}

/* :: Return true if entry exist in file system. Entry name is subject to
 *    symlinks resolving.
 * ::Preconditions:
 * ::Errors:
 *    No errors.
*/
bool                    //! true if exist
stat_file_exists2 (
   const char* fname    //! entry name
){
   ___Dassert(fname, "'fname parameter invalid");

   struct stat buffer;

   stat(fname, &buffer);
   if (errno == ENOENT || errno == ENOTDIR) {
      errno = 0;
      return false;
   }
      ErrorTrigIf(errno != 0, handle_err);

   return true;

  handle_err:
   return false;
}

/* :: Return true if entry exist in file system and is directory.
 * ::Preconditions:
 * ::Errors:
 *    False returned.
 *    opendir()   -- this libc function errors
*/
bool
stat_file_is_dir (
   const char* fname
){
   ___Dassert(fname, "'fname' parameter invalid");
   DIR* ds = opendir(fname);//open on fifo | tape device causes DoS, see open() O_DIRECTORY for solution
   if (errno == ENOTDIR) { //not dir
      errno = 0;
      return false;
   }
   ErrorTrigIf(ds == NULL, handle_err);

   closedir(ds);

   /*: Finish: */
   return true;

  handle_err:
   return false;
}

bool
stat_fs_is_dir (
   FILE* fs
){
   ___Dassert(fs, "'fs' parameter invalid");
   struct stat st_stat;

   int ret = -1;
   ret = fstat(fileno(fs), &st_stat);
      ErrorTrigIf(ret != 0, HANDLE_ERR);

   if ((st_stat.st_mode & S_IFMT) == S_IFDIR) {
      return true;
   } else {
      return false;
   }

  HANDLE_ERR:
   return false;
}

/* :: Return true if both paths refers to same file.
 * ::Preconditions:
 * ::Errors:
 *    False returned.
 *    opendir()   -- this libc function errors
*/
// NOT FINISHED
bool
stat_files_are_same (
   const char* file1,
   const char* file2
){
   ___Dassert(file1, "'file1 parameter invalid");
   ___Dassert(file2, "'file2 parameter invalid");

   char* abs_f1 = NULL;
   char* abs_f2 = NULL;

   char* f1_cpy = cs_cre(file1);
   char* f2_cpy = cs_cre(file2);
      HandleErr(HANDLE_ERR);

   stat_absolutepath (&abs_f1, dirname(f1_cpy));
      HandleErr(HANDLE_ERR);
   stat_absolutepath (&abs_f2, dirname(f2_cpy));
      HandleErr(HANDLE_ERR);

   cs_des(&f1_cpy);
   cs_des(&f2_cpy);

   printf("f1: %s -> %s\n", file1, abs_f1);
   printf("f2: %s -> %s\n", file2, abs_f2);


   return true;

  HANDLE_ERR:
   if (f1_cpy != NULL) { cs_des(&f1_cpy); }
   if (f2_cpy != NULL) { cs_des(&f2_cpy); }
   return false;
}




/* File name modification. */
/* -------------------------------------------------------------------------- */


/* :: Removes all '.', '..' and redundant '/' from path name. Path do not need
 *    to exist.
 *    If path is malformed (too many ..), NULL is set.
 * ::Preconditions:
 * ::Errors:
 *    ENOMEM         -- no memory available
*/
void
stat_replacedots (
   char** new_path,
   const char* path
){
   ___DParam(new_path, != NULL);
   ___DParam(path, != NULL);

   char* new_path_new = NULL;
   bool ending_slash = false;

   strarr_t comps = STRARR_T_Z;  //: path components
   sarr_split_line(&comps, path, "/");
      HandleErr(HANDLE_ERR);

   if (sarr_el_cmp(&comps, comps.ns-1, "") == 0) {
      //: last element is empty => last char was '/'
      ending_slash = true;
   }

   if (path[0] == '\0') {
      //: path is empty
      new_path_new = cs_cre("");
      goto HANDLE_SUCC;
   }

   size_t i = 0;
   if (sarr_el_cmp(&comps, 0, ".") == 0
    || sarr_el_cmp(&comps, 0, "..") == 0
    || sarr_el_cmp(&comps, 0, "") == 0
   ){
      sarr_el_rets(&comps, 0)->arb = 1;
      ++i;
   }

   while (i < comps.ns)
   {
      //: get previous, not removed, component
      size_t j = i;
      if (j > 0) {
         --j;
         while (sarr_el_rets(&comps, j)->arb == 0) {
            --j;
         }
      }

      if (sarr_el_cmp(&comps, i, ".") == 0
       || sarr_el_cmp(&comps, i, "") == 0)
      {
         sarr_el_rets(&comps, i)->arb = 0;
      }
      else if (sarr_el_cmp(&comps, i, "..") == 0)
      {
         sarr_el_rets(&comps, i)->arb = 1;
         if (sarr_el_cmp(&comps, j, "") == 0) {
            goto HANDLE_SUCC;
         } else if (sarr_el_cmp(&comps, j, ".") == 0) {
            sarr_el_rets(&comps, j)->arb = 0;
         } else if (sarr_el_cmp(&comps, j, "..") == 0) {
         } else {
            sarr_el_rets(&comps, i)->arb = 0;
            sarr_el_rets(&comps, j)->arb = 0;
         }
      }
      else
      {
         sarr_el_rets(&comps, i)->arb = 1;
      }
      ++i;
   }

   str_t* el = NULL;
   uint ii = 0;
   do
   {
      el = sarr_el_rets(&comps, ii);
      if (el->arb == 1) {
         new_path_new = cs_cre2(el->buf, el->nc);
            HandleErr(HANDLE_ERR);
      }
      ++ii;
   }
   while (new_path_new == NULL);

   for (; (el = sarr_its(&comps, &ii)) != NULL;)
   {
      if (el->arb == 1) {
         cs_appf(&new_path_new, "/%s", el->buf);
            HandleErr(HANDLE_ERR);
      }
   }

   if (new_path_new[0] == '\0' || ending_slash == true) {
      cs_appf(&new_path_new, "/");
         HandleErr(HANDLE_ERR);
   }


#if 0
   printf("%s\n", path);
   sarr_printp(stdout, "\t|", &comps, "|\n");
   printf("\n");
#endif

  HANDLE_SUCC:
   *new_path = new_path_new;
   sarr_fin(&comps);
   return;

  HANDLE_ERR:
   if (new_path_new != NULL) { cs_des(&new_path_new); }
   if (comps.na > 0) { sarr_fin(&comps); }
   return;
}

/* :: String is made shorter (but not reallocated).
 * ::Preconditions:
 * ::Errors:
 *    ENOMEM         -- no memory available
*/
void
stat_remove_endslashess (
  str_t* name
){
  ___DParam(name, != NULL);

  while (name->buf[name->nc-1] == '/') {
    if (name->nc == 1) { break; }
    name->buf[name->nc-1] = '\0';
    --name->nc;
  }
}




/* Symlinks. */
/* -------------------------------------------------------------------------- */


/* :: Place target name in `ptar`.
 * ::Preconditions:
 * ::Errors:
 *    Target name not read.
 *    ENOMEM         -- no memory available
 *    readlink()     -- this libc function errors
*/
void
stat_readlink (
   char** ptar,         //! [out] name of symlink target
   uint tar_len,        //! presumable target name length, 0 if unknown
   const char* lname    //! symlink name
){
   ___Dassert(ptar, "'ptar' parameter invalid");
   ___Dassert(lname, "'lname' parameter invalid");

   char* buf = NULL;

   /*: Prepare initial target length */
   if (tar_len == 0) { tar_len = 4096; } else { tar_len += 1; }

   /*: Keep reading target name until length is big enough. */
   ssize_t nbw = 0;
   while (true)
   {
      buf = mem_alloc(tar_len);
      nbw =  readlink(lname, buf, tar_len);
         ErrorTrigIf(nbw == -1, handle_err);
      if (nbw < tar_len) { break; } //: done - target name fit into buffer
      mem_free(buf);
      tar_len += tar_len;
   }
   buf[nbw] = '\0';

   /*: Finish: */
   *ptar = buf;
   return;

  handle_err:
   mem_free(buf);
   return;
}

/* :: Resolve all symlinks, '.', '..' and redundant '/' from path name. Place
 *    new path in `abs_p`.
 * ::Preconditions:
 * ::Errors:
 *    Nothing resolved.
 *    realpath()     -- this libc function errors
*/
void
stat_absolutepath (
   char** abs_p,
   const char* fname
){
   ___Dassert(abs_p, "'abs_p parameter invalid");
   ___Dassert(fname, "'fname parameter invalid");

   char* abs_path = realpath(fname, NULL);
      ErrorTrigIf(abs_path == NULL, handle_err);
   *abs_p = abs_path;

   return;

  handle_err:
   return;
}




/* Attributes. */
/* -------------------------------------------------------------------------- */


/* :: Place attributes in struct within entry struct.
 * ::Preconditions:
 * ::Errors:
 *    Attributes not read.
 *    lstat()     -- this libc function errors
*/
void
stat_read_attrs (
   st_ent* stEnt
){
   ___Dassert(stEnt, "'stEnt parameter invalid");

   int err = lstat(stEnt->filename.buf, &stEnt->attrs);
      ErrorTrigIf(err == 0, handle_err);

   return;

  handle_err:
   return;
}

/* :: Return character denoting type.
 *    '-' regular file, 'd' directory, 'l' symlink,
 *    'b' block, 'c' character device
 *    'f' fifo, 's' socket
 *    'N' - non-existant
 * ::Preconditions:
 * ::Note:
 *    Caller mus have permissions to enter (x) all components of path.
 *    No permissions on file needed.
 * ::Errors:
 *    \0 returned.
 *    stat, lstat()  -- this libc function errors
*/
char                    //! file type
stat_file_type_is (
   const char* fname,   //! entry name
   bool deref_sl        //! dereference symlinks in name?
){
   ___Dassert(fname, "'fname parameter invalid");

   char type = '\0';
   struct stat st_stat;

   int ret = -1;
   if (deref_sl == true) {
      ret = stat(fname, &st_stat);
   } else {
      ret = lstat(fname, &st_stat);
   }
      if (ret == -1 && (errno == ENOENT || errno == ENOTDIR)) {
         errno = 0;
         type = 'N'; goto handle_succ;
      }
      ErrorTrigIf(ret != 0, handle_err);

   switch (st_stat.st_mode & S_IFMT) {
      case S_IFREG: type = '-'; break;    // regular file
      case S_IFDIR: type = 'd'; break;    // directory
      case S_IFLNK: type = 'l'; break;    // symlink
      case S_IFBLK: type = 'b'; break;    // block device
      case S_IFCHR: type = 'c'; break;    // character device
      case S_IFIFO: type = 'f'; break;    // fifo
      case S_IFSOCK: type = 's'; break;   // socket
      default: ___Sentinel("wrong file type\n"); abort();
   }

   /*: Finish: */
  handle_succ:
   return type;

  handle_err:
   return '\0';
}

/* :: Place type in `pch` and return it.
 *    '-' regular file, 'd' directory, 'l' symlink,
 *    'b' block, 'c' character device
 *    'f' fifo, 's' socket
 * ::Preconditions:
 * ::Errors:
 *    No errors.
*/
char                 //! return type
stat_extract_ftype (
   char* pch,        //! [out] character dentoing file type
   mode_t fmode      //! file mode
){
   ___Dassert(pch, "'pch' parameter invalid");
   if (S_ISREG(fmode)) { *pch='-'; return *pch; }
   if (S_ISDIR(fmode)) { *pch='d'; return *pch; }
   if (S_ISLNK(fmode)) { *pch='l'; return *pch; }
   if (S_ISBLK(fmode)) { *pch='b'; return *pch; }
   if (S_ISCHR(fmode)) { *pch='c'; return *pch; }
   if (S_ISFIFO(fmode)) { *pch='f'; return *pch; }
   if (S_ISSOCK(fmode)) { *pch='s'; return *pch; }
   *pch = '?'; return *pch;
}

/* :: Place permissions c-string in `perms` and return it.
 *    Format is of "rwxrwxrwx"
 * ::Preconditions:
 * ::Errors:
 *    No errors.
*/
char*                      //! return permissions
stat_extract_perms (
   char (*perms)[10],      //! [out] permissions
   mode_t fmode            //! file mode
){
   ___Dassert(perms, "'perms' parameter invalid");
   ___Dassert(sizeof(*perms) / sizeof((*perms)[0]) == 10, "'perms[10]' must have 10 elements");

   // USER r
   if (S_IRUSR & fmode) {
      (*perms)[0] = 'r';
   } else {
      (*perms)[0] = '-';
   }
   // USER w
   if (S_IWUSR & fmode) {
      (*perms)[1] = 'w';
   } else {
      (*perms)[1] = '-';
   }
   // USER x/s
   if (S_IXUSR & fmode) {
      if (S_ISUID & fmode) {
         (*perms)[2] = 's';
      } else {
         (*perms)[2] = 'x';
      }
   } else {
      if (S_ISUID & fmode) {
         (*perms)[2] = 'S';
      } else {
         (*perms)[2] = '-';
      }
   }
   // GROUP r
   if (S_IRGRP & fmode) {
      (*perms)[3] = 'r';
   } else {
      (*perms)[3] = '-';
   }
   // GROUP w
   if (S_IWGRP & fmode) {
      (*perms)[4] = 'w';
   } else {
      (*perms)[4] = '-';
   }
   // GROUP x/s
   if (S_IXGRP & fmode) {
      if (S_ISGID & fmode) {
         (*perms)[5] = 's';
      } else {
         (*perms)[5] = 'x';
      }
   } else {
      if (S_ISGID & fmode) {
         (*perms)[5] = 'S';
      } else {
         (*perms)[5] = '-';
      }
   }
   // OTHERS r
   if (S_IROTH & fmode) {
      (*perms)[6] = 'r';
   } else {
      (*perms)[6] = '-';
   }
   // OTHERS w
   if (S_IWOTH & fmode) {
      (*perms)[7] = 'w';
   } else {
      (*perms)[7] = '-';
   }
   // OTHERS x/t
#if defined(BSD)
   if (S_IXOTH & fmode) {
      if (S_ISVTX & fmode) {
         (*perms)[8] = 't';
      } else {
         (*perms)[8] = 'x';
      }
   } else {
      if (S_ISVTX & fmode) {
         (*perms)[8] = 'T';
      } else {
         (*perms)[8] = '-';
      }
   }
#else
   if (S_IXOTH & fmode) {
      (*perms)[8] = 'x';
   } else {
      (*perms)[8] = '-';
   }
#endif
   (*perms)[9] = '\0';
   return *perms;
}




/* File owners. */
/* -------------------------------------------------------------------------- */


/* :: Look for user name in users database, place result in `puser` and return
 *    it as c-string.
 * ::Preconditions:
 * ::Errors:
 *    ENOMEM         -- no memory available
 *    getpwuid()     -- this libc function errors
*/
char*                //! user name
stat_look_user (
   char** puser,     //! [out] user name
   uid_t uid         //! user id
){
   ___Dassert(puser, "'puser' parameter invalid");

   struct passwd* result = NULL;
   struct passwd pwd = {0};
   char* buf = NULL;    //buf used by getpwuid_r() to hold cstrings
   char* user = NULL;

   /* initial suggested size of `buf` */
   int nb_buf = sysconf(_SC_GETPW_R_SIZE_MAX);
   if (nb_buf == -1) { nb_buf = 1; }

   /* get info about user */
   while (true)
   {
      buf = mem_alloc(nb_buf);
         HandleErr(handle_err);
      int ret = getpwuid_r(uid, &pwd, buf, nb_buf, &result);
      if (ret == ERANGE) {    //: buffer needs to be bigger
         mem_free(buf);
         nb_buf += nb_buf;
         continue;
      }
      ErrorTrigIf(ret != 0, handle_err);
      break;
   }

   /* create cstring with user name */
   if (result == NULL) {
      user = cs_cre("NN");
         HandleErr(handle_err);
   } else {
      user = cs_cre(pwd.pw_name);
         HandleErr(handle_err);
   }

   /* Finish: */
   *puser = user;
   mem_free(buf);
   return *puser;

  handle_err:
   mem_free(buf);
   cs_des(&user);
   return NULL;
}

/* :: Look for group name in users database, place result in `pgroup` and return
 *    it as c-string.
 * ::Preconditions:
 * ::Errors:
 *    ENOMEM         -- no memory available
 *    getgrgid()     -- this libc function errors
*/
char*                //! group name
stat_look_group (
   char** pgroup,    //! [out] group name
   gid_t gid         //! group id
){
   ___Dassert(pgroup, "'pgroup' parameter invalid");

   struct group* result = NULL;
   struct group grp = {0};
   char* buf = NULL;    //buf used by getpwuid_r() to hold cstrings
   char* group = NULL;

   /* initial suggested size of `buf` */
   int nb_buf = sysconf(_SC_GETGR_R_SIZE_MAX);
   if (nb_buf == -1) { nb_buf = 1; }

   /* get info about user */
   while (true)
   {
      buf = mem_alloc(nb_buf);
         HandleErr(handle_err);
      int ret = getgrgid_r(gid, &grp, buf, nb_buf, &result);
      if (ret == ERANGE) {
         mem_free(buf);
         nb_buf += nb_buf;
         continue;
      }
      ErrorTrigIf(ret != 0, handle_err);
      break;
   }

   /* create cstring with user name */
   if (result == NULL) {
      group = cs_cre("NN");
         HandleErr(handle_err);
   } else {
      group = cs_cre(grp.gr_name);
         HandleErr(handle_err);
   }

   /* finish */
   *pgroup = group;
   mem_free(buf);
   return *pgroup;

  handle_err:
   mem_free(buf);
   cs_des(&group);
   return NULL;
}

/* :: Look for uid in user database and return it.
 * ::Preconditions:
 * ::Errors:
 *    ENOMEM         -- no memory available
 *    getpwnam()     -- this libc function errors
*/
uid_t                   //! user id
stat_look_uid (
   const char* cs_user  //! user name
){
   ___Dassert(cs_user, "'cs_user' parameter invalid");

   struct passwd* result = NULL;
   struct passwd pwd = {0};
   char* buf = NULL;    //buf used by getpwuid_r() to hold cstrings
   uid_t uid = 0;

   /* initial suggested size of `buf` */
   int nb_buf = sysconf(_SC_GETPW_R_SIZE_MAX);
   if (nb_buf == -1) { nb_buf = 1; }

   /* get info about cs_user */
   while (true)
   {
      buf = mem_alloc(nb_buf);
         HandleErr(handle_err);
      int ret = getpwnam_r(cs_user, &pwd, buf, nb_buf, &result);
      if (ret == ERANGE) { //buffer needs to be bigger
         mem_free(buf);
         nb_buf += nb_buf;
         continue;
      }
      ErrorTrigIf(ret != 0, handle_err);
      break;
   }

   uid = pwd.pw_uid;

   /* finish */
   mem_free(buf);
   return uid;

  handle_err:
   mem_free(buf);
   return 0;
}

/* :: Look for gid in user database and return it.
 * ::Preconditions:
 * ::Errors:
 *    ENOMEM         -- no memory available
 *    getgrnam()     -- this libc function errors
*/
gid_t                   //! group id
stat_look_gid (
   const char* cs_group //! group name
){
   ___Dassert(cs_group, "'cs_group' parameter invalid");

   struct group* result = NULL;
   struct group grp = {0};
   char* buf = NULL;    //buf used by getpwuid_r() to hold cstrings
   gid_t gid = 0;

   /* initial suggested size of `buf` */
   int nb_buf = sysconf(_SC_GETGR_R_SIZE_MAX);
   if (nb_buf == -1) { nb_buf = 1; }

   /* get info about cs_group */
   while (true)
   {
      buf = mem_alloc(nb_buf);
         HandleErr(handle_err);
      int ret = getgrnam_r(cs_group, &grp, buf, nb_buf, &result);
      if (ret == ERANGE) { //buffer needs to be bigger
         mem_free(buf);
         nb_buf += nb_buf;
         continue;
      }
      ErrorTrigIf(ret != 0, handle_err);
      break;
   }

   gid = grp.gr_gid;

   /* finish */
   mem_free(buf);
   return gid;

  handle_err:
   mem_free(buf);
   return 0;
}
