/*!F::Command-line arguments parser - main implementation file.
 *
 *  ::Purpose:
 *    Parse given array of c-strings as arguments specified on command-line.
 *  ..Parsing is done accordingly to cl-args specifications established by
 *    user. Values of options and operands are assigned to data inside 
 *    specification structures.
 *  ..Cl-args established in specifications can be: check for occurence in cl,
 *    get their indexes in occurence order, retrieve their values given on cl
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - parse command-line args, parse config file, parse config shell variable
 *    - get cl-arg values given on cl
 *
 *  ::Limitations:
 *    - surely more different ways of getting cl-args and their values can be
 *      implemented
 *
 *  ::Arguments format:
 *    operand                    -- argument not starting with hyphen is operand
 *    -o --opt                   -- short and long options starts with hyphens
 *    -o=val,val --opt=val,val   -- values separated with coma specified after '='
 *    ---ARG                     -- error
 *
 *  ::Faults:
 *    - invalid argument format (part. 3+ hyphens: '---ARG')
 *    - option not recognized
 *    - too many values for option or too many operands given
 *    - operand/option value not in valid range
 *    - mandatory operand/option not specified or not enough values given
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-01-12
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_prconf.h"
#include "prconf_spec.h"
#include "prconf_valid.h"

/* mh library */
#include "core.h"
#include "debug.h"
#include "errors.h"
#include "faults.h"
#include "cstrings.h"
#include "strings_arr.h"
#include "stat_info.h"
#include "fstream_basic.h"
#include "fstream_read.h"

/* libc */
#include <stddef.h>        // offsetof()
#include <stdio.h>         // printf()
#include <string.h>        // str*()

/* current module */
#include "prconf_parser.h"
#include "prconf_parser_internal.h.inc"
#include "prconf_parser_aux.h.inc"




void
cla_parse_clargs (
   st_specs_arr* stArgSpecs,  //! [out] arguments specification
   char* argv[],              //! cl arguments
   int argc,                  //! num of cl args
   st_faults_arr* Faults      //! [out] faults detected during parsing
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(argv, != NULL);
   ___DParam(argc, > 0);
   ___DParam(Faults, != NULL);

   /*: Parse each argument after program name: */
   for (int i = 1; i < argc; ++i)
   {
      const char* arg = argv[i];
      parse_argument(stArgSpecs, arg, Faults);
         HandleErr(HANDLE_ERR);
   }

   /*: Finish: */
   return;

  HANDLE_ERR:
   return;
}

/* Retrieve every line from file and parse it identical as one cl argument.
   '#' as first character in line denotes comment which is ignored.
 */
void
cla_parse_conffile (
   st_specs_arr* stArgSpecs,  //! [out] arguments specification
   const char* fname,         //! config file name
   st_faults_arr* Faults      //! [out] faults detected during parsing
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(fname, != NULL);
   ___DParam(Faults, != NULL);

   FILE* fs = NULL;
   char* arg = NULL;

   /*: Open config file: */
   fs_open(&fs, fname, "r");
   if (errno == EACCES || errno == ENOENT || errno == ENOTDIR || errno == EISDIR) {
      //: no access to file, file does not exist or is directory: ommit
      ClrErr();
      goto HANDLE_SUCC;
   }
      HandleErr(HANDLE_ERR); ___DInter(fs, != NULL);
   if (stat_fs_is_dir(fs)) { fclose(fs); goto HANDLE_SUCC; }

   /*: Each line read from file: */
   size_t nb_read = 0;
   while ((arg = fs_rline(fs, &nb_read)))
   {
      if (arg[nb_read - 1] == '\n') { arg[nb_read - 1] = '\0'; }  //: handle '\n'
      if (arg[0] == '#') { cs_des(&arg); continue; }              //: handle comments
      parse_argument(stArgSpecs, arg, Faults);                    //: Parse line as one argument
         HandleErr(HANDLE_ERR);

      cs_des(&arg);
   }
      HandleErr(HANDLE_ERR);  //handle errors of fs_rline()

   /*: Close config file: */
   fs_close(&fs);

   /*: Finish: */
  HANDLE_SUCC:
   return;

  HANDLE_ERR:
   if (arg != NULL) { cs_des(&arg); }
   if (fs != NULL) { fs_close(&fs); }
   return;
}

/* Read shell variable and parse it as cl arguments. Arguments are separated
   by ';', if arguments starts with '#' it is considered comment.
 */
void
cla_parse_shellvar (
   st_specs_arr* stArgSpecs,  //! [out] arguments specification
   const char* varname,       //! shell variable name
   st_faults_arr* Faults      //! [out] faults detected during parsing
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(varname, != NULL);
   ___DParam(Faults, != NULL);

   /*: Retrieve variable from shell environment: */
   char* shvar = getenv(varname);
   if (shvar == NULL) { goto HANDLE_SUCC; }  //: no variable found - ommit

   /*: Split value of variable to arguments: */
   strarr_t args = STRARR_T_Z;
   sarr_split_line2(&args, shvar, ";");
      HandleErr(HANDLE_ERR); ___DInter(args.buf, != NULL);

   /*: Parse every argument identicaly as cl argument: */
   const char* arg = NULL;
   uint ii = 0;
   while ((arg = sarr_it(&args, &ii)))
   {
      if (arg[0] == '#') { continue; }    //: commented-out - ommit
      parse_argument(stArgSpecs, arg, Faults);
         HandleErr(HANDLE_ERR);
   }

   sarr_fin(&args);

   /*: Finish: */
  HANDLE_SUCC:
   return;

  HANDLE_ERR:
   sarr_fin(&args);
   return;
}

void
cla_validate_arguments (
   st_specs_arr* stArgSpecs,  //! [out] arguments specification
   st_faults_arr* Faults      //! [out] faults detected during parsing
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(Faults, != NULL);

   /*: Check each option in arguments specification: */
   for (uint i = 0; i < stArgSpecs->opt_specs.ns; ++i) {
      st_opt_spec* ospec = arrptrs_el_ret(&stArgSpecs->opt_specs, i);
      if (ospec == NULL) { continue; }
      //: is mandatory and wasnt set:
      if (mandatory_and_not_set(ospec, NULL)) {
         faultsarr_add_fault(Faults, Faultcath_exit, "mandatory option '%s' not present", ospec->lname.buf);
            HandleErr(HANDLE_ERR);
      }
      //: num of values given:
      int res = num_of_vals_check(ospec, NULL);
      if (res == -1) {
         faultsarr_add_fault(Faults, Faultcath_exit, "not enough values for option '%s'", ospec->lname.buf);
            HandleErr(HANDLE_ERR);
      } else if (res == 1) {
         faultsarr_add_fault(Faults, Faultcath_exit, "too many values for option '%s'", ospec->lname.buf);
            HandleErr(HANDLE_ERR);
      }
   }
   /*: Check each operand in arguments specification: */
   for (uint i = 0; i < stArgSpecs->oper_specs.ns; ++i) {
      st_oper_spec* ospec = arrptrs_el_ret(&stArgSpecs->oper_specs, i);
      if (ospec == NULL) { continue; }
      //: is mandatory and wasnt set:
      if (mandatory_and_not_set(NULL, ospec)) {
         faultsarr_add_fault(Faults, Faultcath_exit, "mandatory operand number '%d' not present", ospec->id);
            HandleErr(HANDLE_ERR);
      }
      //: num of values given:
      int res = num_of_vals_check(NULL, ospec);
      if (res == -1) {
         faultsarr_add_fault(Faults, Faultcath_exit, "not enough values for operand number '%d'", ospec->id);
            HandleErr(HANDLE_ERR);
      } else if (res == 1) {
         faultsarr_add_fault(Faults, Faultcath_exit, "too many values for operand number '%d'", ospec->id);
            HandleErr(HANDLE_ERR);
      }
   }

   /*: Finish: */
   return;

  HANDLE_ERR:
   return;
}

void
cla_print_set_arguments (
   st_specs_arr* stArgSpecs   //! arguments specification
){
   ___DParam(stArgSpecs, != NULL);

   { /*--*/
   for (uint i = 0; i < stArgSpecs->opt_specs.ns; ++i)
   {
      st_opt_spec* el_opt = arrptrs_el_ret(&stArgSpecs->opt_specs, i);
      if (el_opt == NULL) { continue; }
      printf("id=%d:%s/%s: lvl=%d, vals=", el_opt->id, el_opt->sname.buf, el_opt->lname.buf,
            el_opt->lvl);
      sarr_printp (stdout, "", &el_opt->vals, ",");
      printf("\n");
   }
   } /*--*/

   { /*--*/
   for (uint i = 0; i < stArgSpecs->oper_specs.ns; ++i)
   {
      st_oper_spec* el_oper = arrptrs_el_ret(&stArgSpecs->oper_specs, i);
      if (el_oper == NULL) { continue; }
      printf("id=%d: vals=", el_oper->id);
      sarr_printp (stdout, "", &el_oper->vals, ",");
      printf("\n");
   }
   } /*--*/
}

bool
cla_opt_is_set (
   st_specs_arr* stArgSpecs,  //! arguments specification
   uint opt_id                //! option id
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(opt_id, < stArgSpecs->opt_specs.ns);
   st_opt_spec* opt = arrptrs_el_ret(&stArgSpecs->opt_specs, opt_id);
   ___DInter(opt, != NULL);
   if (opt->lvl > 0) { return true; }
   else { return false; }
}

bool
cla_oper_is_set (
   st_specs_arr* stArgSpecs,  //! arguments specification
   uint oper_id               //! operand id
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(oper_id, < stArgSpecs->oper_specs.ns);
   st_oper_spec* oper = arrptrs_el_ret(&stArgSpecs->oper_specs, oper_id);
   ___DInter(oper, != NULL);
   if (oper->vals.ns > 0) { return true; }
   else { return false; }
}

int
cla_ret_opt_lvl (
   st_specs_arr* stArgSpecs,  //! arguments specification
   uint opt_id                //! option id
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(opt_id, < stArgSpecs->opt_specs.ns);
   st_opt_spec* opt = arrptrs_el_ret(&stArgSpecs->opt_specs, opt_id);
   ___DInter(opt, != NULL);
   return opt->lvl;
}

char*
cla_ret_opt_value (
   st_specs_arr* stArgSpecs,  //! arguments specification
   uint opt_id,               //! option id
   uint vn                    //! value number
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(opt_id, < stArgSpecs->opt_specs.ns);
   st_opt_spec* opt = arrptrs_el_ret(&stArgSpecs->opt_specs, opt_id);
   ___DInter(opt, != NULL);
   if (vn >= opt->vals.ns) { return NULL; }
   return sarr_el_ret(&opt->vals, vn);
}

char*
cla_ret_oper_value (
   st_specs_arr* stArgSpecs,  //! arguments specification
   uint oper_id,              //! operand id
   uint vn                    //! value number
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(oper_id, < stArgSpecs->oper_specs.ns);
   st_oper_spec* oper = arrptrs_el_ret(&stArgSpecs->oper_specs, oper_id);
   ___DInter(oper, != NULL);
   if (vn >= oper->vals.ns) { return NULL; }
   return sarr_el_ret(&oper->vals, vn);
}

strarr_t*
cla_ret_opt_values_arr (
   st_specs_arr* stArgSpecs,  //! arguments specification
   uint opt_id                //! option id
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(opt_id, < stArgSpecs->opt_specs.ns);
   st_opt_spec* opt = arrptrs_el_ret(&stArgSpecs->opt_specs, opt_id);
   ___DInter(opt, != NULL);
   return &opt->vals;
}
