/*!F::Command-line arguments specificator - main implementation file.
 *
 *  ::Purpose:
 *    Establish options, operands and their parameters in cl-args specification.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - initialize and finish array of specifications
 *    - establish cl-arg specification
 *
 *  ::Limitations:
 *    - no internal destruction of valid values - it requires mechanism to
 *       handle multipe destructions because each valid values array can be
 *       assigned to multiple cl-arg specifications
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-01-12
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "memory.h"
#include "strings.h"
#include "strings_arr.h"

/* current project */
#include "common_prconf.h"
#include "clarg_storage.h.inc"     // structures to hold arguments data

/* libc */

/* current module */
#include "prconf_spec.h"
#include "prconf_spec_internal.h.inc"




void
cla_specs_initialize (
   st_specs_arr* stArgSpecs,
   uint nopts,
   uint nopers
){
   ___DParam(stArgSpecs, != NULL);

   arrptrs_ini2(&stArgSpecs->opt_specs, nopts, NULL);
   arrptrs_ini2(&stArgSpecs->oper_specs, nopers, NULL);
      HandleErr(HANDLE_ERR);

   return;

  HANDLE_ERR:
   return;
}

void
cla_opt_establish (
   st_specs_arr* stArgSpecs,
   uint id,
   behopt_en beh,
   const char* sname,
   const char* lname,
   sint vals_min,
   sint vals_max,
   strarr_t* p_vv,
   bool mand
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(sname, != NULL);
   ___DParam(lname, != NULL);

   st_opt_spec* ospec_new = optspec_constr_(id, beh, sname, lname, vals_min,
                                             vals_max, p_vv, mand);
      HandleErr(HANDLE_ERR); ___DInter(ospec_new, != NULL);

   ___DInter(id, < stArgSpecs->opt_specs.na);
   arrptrs_cpy(&stArgSpecs->opt_specs, id, ospec_new);

   return;

  HANDLE_ERR:
   return;
}

void
cla_oper_establish (
   st_specs_arr* stArgSpecs,
   uint id,
   sint vals_min,
   sint vals_max,
   strarr_t* p_vv,
   bool mand
){
   ___DParam(stArgSpecs, != NULL);

   st_oper_spec* ospec_new = operspec_constr_(id, vals_min, vals_max, p_vv, mand);
      HandleErr(HANDLE_ERR); ___DInter(ospec_new, != NULL);

   ___DInter(id, < stArgSpecs->oper_specs.na);
   arrptrs_cpy(&stArgSpecs->oper_specs, id, ospec_new);

   return;

  HANDLE_ERR:
   return;
}

void
cla_specs_cleanup (
   st_specs_arr* stArgSpecs
){
   ___DParam(stArgSpecs, != NULL);

   for (uint i = 0; i < stArgSpecs->opt_specs.ns; ++i) {
      st_opt_spec* ospec = arrptrs_el_ret(&stArgSpecs->opt_specs, i);
      if (ospec != NULL) { optspec_destr_(&ospec); }
   }

   for (uint i = 0; i < stArgSpecs->oper_specs.ns; ++i) {
      st_oper_spec* ospec = arrptrs_el_ret(&stArgSpecs->oper_specs, i);
      if (ospec != NULL) { operspec_destr_(&ospec); }
   }

   arrptrs_fin(&stArgSpecs->opt_specs, NULL);
   arrptrs_fin(&stArgSpecs->oper_specs, NULL);
}
