/*!F::Validate cl-arguments specification - main implementation file.
 *
 *  ::Purpose:
 *    Make sure established cl-arg specifications have valid format and
 *    parameters.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - validate variables existence and correctness, data format, conflicts
 *       between settings
 *
 *  ::Limitations:
 *    - for sure some more validations can be implemented
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-01-12
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_prconf.h"  // pr usng this lib must provide common.h
#include "clarg_storage.h.inc"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "strings_arr.h"

/* libc */

/* current module */
#include "prconf_valid.h"
#include "prconf_valid_aux.h.inc"




/*!:: Check validity of struct data including each cl-arg specification in array.
 * .. Each fault detected is added to `Faults` as separate string.
 *    If fault has critical nature (causing following checks impossible) function
 *    returns immediately, otherwise checks are continued.
 *
 * ::Preconditions:
 *  - `stClSpec` must be already created
 *  - `Faults` must be already created
 *
 * ::Errors:
 *    Fults are left in `Faults`.
 *    ENOMEM            -- no memory
 */
void
validate_clargs_specification (
   st_clarg_specs_arr* stClSpec,    //! array of cl-arg specifications
   strarr_t* Faults                 //! [out] faults detected
){
   ___Dassert(stClSpec, "'stClSpec' parameter invalid");
   ___Dassert(Faults, "'Faults' parameter invalid");

   sarr_app(Faults, "Faults detected:");
      HandleErr(handle_err);

   /*: Checks:
    *    . variable correctness (addresses etc.)
    *    . format validity (strings etc.)
    *    . logic (conflicts between settings)
    */
   check_variable(stClSpec, Faults);
      HandleErr(handle_err);
   check_format(stClSpec, Faults);
      HandleErr(handle_err);
   check_logic(stClSpec, Faults);
      HandleErr(handle_err);

   /*: Finish: */
   return;

  handle_err:
   return;
}
