/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MHCL_PRCONF_SRCSPEC_PRCONFVALID_AUX_H_INC
#define MHCL_PRCONF_SRCSPEC_PRCONFVALID_AUX_H_INC


/*!F::Validate cl-arguments specification - auxillary file.
 *
 *  ::Provides helper functions for main implementation file.
 */




/* employs includes from opts_valid.c */




/*!1:: Check cl-arg specifications - variables correctness */
static inline void check_variable (
      st_clarg_specs_arr* stClSpec, strarr_t* Faults)
                     __attr_nonnull__(1,2)
                     ;

/*!1:: Check cl-arg specifications - format correctness */
static inline void check_format (
      st_clarg_specs_arr* stClSpec, strarr_t* Faults)
                     __attr_nonnull__(1,2)
                     ;

/*!1:: Check cl-arg specifications - logic */
static inline void check_logic (
      st_clarg_specs_arr* stClSpec, strarr_t* Faults)
                     __attr_nonnull__(1,2)
                     ;




/*!:: Check members of cl-arguments specification struct: are required variables
 *    set, are sizes correct, is data inside members properly set etc.?
 * .. Each fault detected is added to `Faults` as separate string.
 *    If fault has critical nature (causing following checks impossible) function
 *    returns immediately, otherwise checks are continued.
 *    Errors: NOMEM
 */
static inline
void
check_variable (
   st_clarg_specs_arr* stClSpec,    //! array of cl-arg specifications
   strarr_t* Faults                 //! [out] faults detected
){
   ___Dassert(stClSpec, "'stClSpec' parameter invalid");
   ___Dassert(Faults, "'Faults' parameter invalid");

   /*: does array of option entry exist inside `stClSpec` ? */
   if (stClSpec->OptEntries == NULL) {
      sarr_app(Faults, "\tthere is no options specification detected");
         HandleErr(handle_err);
      goto critical_fault;
   }

   /*: For each cl-arg specification in array, check: */
   for (size_t i = 0; i < stClSpec->ns; ++i)
   {
      //: cstring inside short name not NULL
      if (str_is_empty(&stClSpec->OptEntries[i].sname)) {
         sarr_appf(Faults, "\tshort name of option %zu is corrupted (NULL)", i);
            HandleErr(handle_err);
         goto critical_fault;
      }
      //: cstring inside long name not NULL
      if (str_is_empty(&stClSpec->OptEntries[i].lname)) {
         sarr_appf(Faults, "\tlong name of option %zu is corrupted (NULL)", i);
            HandleErr(handle_err);
         goto critical_fault;
      }
      //: cstring inside each valid value not NULL
      strarr_t* p_valid_vals = stClSpec->OptEntries[i].p_valid_vals;
      if (p_valid_vals != NULL) {
          for (size_t j = 0; j < p_valid_vals->ns; ++j) {
            const char* cstr_tmp = sarr_el_ret(p_valid_vals, j);
            if (cstr_tmp == NULL) {
               sarr_appf(Faults, "\tvalue %zu of option %zu is corrupted, it is NULL", j, i);
                  HandleErr(handle_err);
               goto critical_fault;
            }
         }
      }
      // No checks for default values
   }

   /*: Finish: */
   return;

  handle_err:
   return;

  critical_fault:
   return;
}

/*!:: Check strings format and numbers range in cl-arguments specification
 *    struct.
 * .. Each fault detected is added to `Faults` as separate string.
 *    If fault has critical nature (causing following checks impossible) function
 *    returns immediately, otherwise checks are continued.
 *    Errors: NOMEM
 */
static inline
void
check_format (
   st_clarg_specs_arr* stClSpec,    //! array of cl-arg specifications
   strarr_t* Faults                 //! [out] faults detected
){
   ___Dassert(stClSpec, "'stClSpec' parameter invalid");
   ___Dassert(Faults, "'Faults' parameter invalid");

   /*: For each cl-arg specification in array, check: */
   for (size_t i = 0; i < stClSpec->ns; ++i)
   {
      const char* cstr_tmp = NULL;
      cstr_tmp = stClSpec->OptEntries[i].sname.buf;
      /*: short name format - doesn't start with hyphen? */
      if (cstr_tmp[0] == '-') {
         sarr_appf(Faults, "\tshort name %s of option %zu is invalid, starts with '-'", cstr_tmp, i);
            HandleErr(handle_err);
      }
      /*: short name format - one letter only? */
      if (cstr_tmp[0] != '\0' && cstr_tmp[1] != '\0') {
         sarr_appf(Faults, "\tshort name %s of option %zu is longer then one character", cstr_tmp, i);
            HandleErr(handle_err);
      }
      /*: short name format - doesn't contain '='? */
      if (cstr_tmp[0] == '=') {
         sarr_appf(Faults, "\tshort name %s of option %zu contains '=' character", cstr_tmp, i);
            HandleErr(handle_err);
      }

      cstr_tmp = stClSpec->OptEntries[i].lname.buf;
      /*: long name format - doesn't start with hyphen? */
      if (cstr_tmp[0] == '-') {
         sarr_appf(Faults, "\tlong name %s of option %zu is invalid, starts with '-'", cstr_tmp, i);
            HandleErr(handle_err);
      }
      /*: long name format - doesn't contain '='? */
      for (size_t j = 0; cstr_tmp[j] != '\0'; ++j)
      {
         if (cstr_tmp[j] == '=') {
            sarr_appf(Faults, "\tlong name %s of option %zu contains '=' character", cstr_tmp, i);
               HandleErr(handle_err);
         }
      }

      strarr_t* p_valid_vals = stClSpec->OptEntries[i].p_valid_vals;
      if (p_valid_vals != NULL) {
         for (size_t j = 0; j < p_valid_vals->ns; ++j) {
            cstr_tmp = sarr_el_ret(p_valid_vals, j);
            /*: each valid value format - doesn't start with hyphen? */
            if (cstr_tmp[0] == '-') {
               sarr_appf(Faults, "\tvalue %s of option %zu is invalid, starts with '-'", cstr_tmp, i);
                  HandleErr(handle_err);
            }
         }
      }
   }

   /*: Finish: */
   return;

  handle_err:
   return;
}

/*!:: Check relations and conflicts between values in cl-arguments specification
 *    struct.
 * .. Each fault detected is added to `Faults` as separate string.
 *    If fault has critical nature (causing following checks impossible) function
 *    returns immediately, otherwise checks are continued.
 *    Errors: NOMEM
 */
static inline
void
check_logic (
   st_clarg_specs_arr* stClSpec,    //! array of cl-arg specifications
   strarr_t* Faults                 //! [out] faults detected
){
   ___Dassert(stClSpec, "'stClSpec' parameter invalid");
   ___Dassert(Faults, "'Faults' parameter invalid");

   /*: For each cl-arg specification in array, check: */
   for (size_t i = 0; i < stClSpec->ns; ++i)
   {
      st_clarg_spec* stOpt = &stClSpec->OptEntries[i];
      /*: option min and max value - greater then -1? */
      if (stOpt->vn_min < -1 || stOpt->vn_max < -1) {
         sarr_appf(Faults, "\tmin and max value can't be less then -1 in option %zu", i);
            HandleErr(handle_err);
      }
      /*: option min and max value - min not greater then max? */
      if (stOpt->vn_max > -1 && stOpt->vn_min > stOpt->vn_max) {
         sarr_appf(Faults, "\tmin values number is greater then max in option %zu", i);
            HandleErr(handle_err);
      }
   }

   /*: Finish: */
   return;

  handle_err:
   return;
}




#endif /*- MHCL_PRCONF_SRCSPEC_PRCONFVALID_AUX_H_INC -*/
