/*!F::Main program file for this project.
 *
 *  ::Purpose:
 *    Contains function main() running project's proper program. Should be
 *    compiled and linked from Makefile by targets destined to build program.
 *  ..For tests there should be another source file with distinct function
 *    main() running module tests, that should be used by targets destined to
 *    build tests.
 *  ..For library projects there should be target in Makefile that creates
 *    library.
 *
 *  ::Note:
 *    This file should be excluded from linking for builds other then
 *    working program.
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of 'Tfman' program.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-24
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




#include "core.h"
#include "debug.h"
#include "prconf.h"
#include "fstream_basic.h"
#include "stat_info.h"

#include "common_tfman.h"
#include "opts.h"
#include "scan_img.h"
#include "scan_txt.h"
#include "parse_img.h"
#include "parse_acts.h"
#include "perform_actions.h"

#include <stdio.h>      // printf()
#include <stdlib.h>     // printf()
#include <string.h>     // strcmp()



#define TFMAN_OPEN_FILES_MAX 128




/*!:: Print usage and/or help. Depends on what was requested in options.
 */
static inline
bool
print_usage_help_if_required (
   st_tfman_opts* stTfmOptions,     //! 'tfman' options
   const char* cstr_usage,          //! usage c-string
   const char* cstr_help            //! help c-string
){
   bool ret = false;    //: true if anything was printed

   /*: Print: */
   if (stTfmOptions->usage && stTfmOptions->help)   //: both
   {
      int res = fprintf(stdout, cstr_usage);
      res = fprintf(stdout, "\n");
      res = fprintf(stdout, cstr_help);
         ErrorTrigIf(res < 0, handle_err);
      ret = true;
   }
   else if (stTfmOptions->usage)      //: usage
   {
      int res = fprintf(stdout, cstr_usage);
         ErrorTrigIf(res < 0, handle_err);
      ret = true;
   }
   else if (stTfmOptions->help)       //: help
   {
      int res = fprintf(stdout, cstr_help);
         ErrorTrigIf(res < 0, handle_err);
      ret = true;
   }

   /*: Finish: */
   return ret;

  handle_err:
   return false;
}


/*!:: Open file streams for file names given in array and as operands on
 *    command-line.
 */
static inline
void
open_input_files (
   FILE* Fss[],            //! [out] array of file streams opened
   uint* pn_files,         //! number of files opened
   strarr_t* opts_files,   //! [out] array of file names opened
   st_specs_arr* stArgSpecs,   //! specifications - needed to open operands given on command-line
   uint max_opened_files   //! stop opening when this number reached
){

   uint n_file = 0;  //num of files opened

   /*: Open each operand given on cl: */
   const char* clval = NULL;  //: cl-arg current value
   uint n_val = 0;
   while ((clval = cla_ret_oper_value(stArgSpecs, OPER_A, n_val++)) != NULL && n_file < max_opened_files) {
      fs_open (&Fss[n_file], clval, "r");
         HandleErrm(FUN_RETURN, "Can not open: %s", clval);
      sarr_app(opts_files, clval);
      ++n_file;
   }

   /*: Finish: */
  FUN_RETURN:
   if (n_file != 0) {
      *pn_files = n_file;
   }
   return;
}

/*!:: Close all given file streams. Ignore errors - continue closing.
 */
static inline
void
close_file_streams (
   FILE* Fss[]       //! array of file streams
){
   /*: Close all file streams in array. Ignore errors. */
   for (uint j = 0; Fss[j] != NULL; ++j) {
      /*int ret =*/ fclose(Fss[j]);
         //ErrorTrigIf(ret != 0, FUN_RETURN);
      Fss[j] = NULL;
   }
  //FUN_RETURN:
   return;
}




/*!:: Scan directory.
 */
void
scan_directory_tree (
   st_tfman_opts* stTfmOptions,  //! 'tfman' options structure
   st_fsimg* stFsImage,          //! [out] file system image
   st_fstxt* stFsText,           //! [out] file system text representation
   const char* dirname,          //! directory name to scan
   FILE* Fss[]                   //! file streams to write
){
   /*: Scan directory: */
   repimg_scan_tree(stFsImage, dirname, stTfmOptions); //: fs image
      HandleErr(FUN_RETURN);
   fstxt_image_to_text(stFsText, stFsImage);          //: text repr
      HandleErr(FUN_RETURN);

   /*: Write text representation to each given file stream: */
   uint i = 0;
   while (Fss[i] != NULL)
   {
      fstxt_print(Fss[i], stFsText, stTfmOptions->verb_lvl, stTfmOptions);   //: with given verbose level
         HandleErr(FUN_RETURN);
      int ret = fprintf(Fss[i], "\n");
         ErrorTrigIf(ret < 0, FUN_RETURN);
      ++i;
   }

   /*: Finish: */
  FUN_RETURN:
   repimg_cleanup_entries(stFsImage);  // make place for image of next fs tree
   fstxt_cleanup_text(stFsText);      // make place for next text representation
   return;
}

/*!:: Parse file cintainig text representation of file system.
 */
void
parse_fsrep_file (
   st_parser_main* stParserMain, //! main parser structure
   FILE* fs,                     //! file stream - file with text repr
   const char* filename,         //! file name
   bool show_only,               //! show action only, do not perform
   bool ask,                     //! ask before perform actions
   uint start_line,
   uint end_line,
   bool compute_names_only
){
   st_txt_img istTxtRep = ST_TXT_IMG_Z;
   st_actions_table istTable = ST_ACTIONS_TABLE_Z;
   strarr_t Names = STRARR_T_Z;

   /*: Input text ==> textual representation data: */
   pars_prepare_txtrep(&istTxtRep, fs, filename, stParserMain);
      HandleErr(FUN_RETURN);

   //: Print faults encountered during parsing
   if (stParserMain->ParsFaults.caths_num[Faultcath_continue] > 0) {
      fprintf(stderr, "\n::: Faults discovered:\n");
      faultsarr_print(Faultcath_continue, stderr, &stParserMain->ParsFaults, NULL, "", "\n", NULL);
      fprintf(stderr, "\n");
   }
   if (stParserMain->ParsFaults.caths_num[Faultcath_exit] > 0) {
      fprintf(stderr, "\n::: Critical faults discovered:\n");
      faultsarr_print(Faultcath_exit, stderr, &stParserMain->ParsFaults, NULL, "", "\n", NULL);
      fprintf(stderr, "::: Parsing terminated.\n\n");
      ErrorSetm(E_FORMAT_WRONG, FUN_RETURN,
               "wrong format of textual representation");
   }

   /* Text rep to names */
   if (compute_names_only == true) {
      pars_txtrepdata_to_names(&Names, &istTxtRep, stParserMain,
                                 start_line, end_line);
         HandleErr(FUN_RETURN);
      sarr_printn(stdout, &Names);
      goto FUN_RETURN;
   }


   /* Text rep to actions */
   pars_txtrepdata_to_actions(&istTable, &istTxtRep, stParserMain,
                              start_line, end_line);
      HandleErr(FUN_RETURN);

   size_t acts_num = acts_sift_and_present_to_user(&istTable, stParserMain);
      HandleErr(FUN_RETURN);

   if (show_only == true) { goto FUN_RETURN; }

   /*: Perform */
   if (acts_num > 0)
   {

      //: Redirect input to be able to read from stdin:
      if (fs == stdin)  //: we are reading input from stdin (redirection was made)
      {
         FILE* res = freopen("/dev/tty", "r", stdin); //: point stdin back to terminal
            ErrorTrigIf(res == NULL, FUN_RETURN);
      }
      //: Ask
      if (ask == true)
      {
         if (ask_yn("yY", "Continue y/n?")) {
            acts_perform(&istTable);
               HandleErr(FUN_RETURN);
         }
      }
      else
      {
         acts_perform(&istTable);
            HandleErr(FUN_RETURN);
      }

   }
   else
   {
      printf("No actions\n");
   }


   /*: Finish: */
  FUN_RETURN:
   pars_cleanup_actstable(&istTable);
   pars_cleanup_txtrep(&istTxtRep);
   sarr_fin(&Names);
   return;
}




int
main (int argc, char* argv[])
{
  st_faults_arr Faults = ST_FAULTS_ARR_Z;

  st_specs_arr stArgSpecs = ST_SPECS_ARR_Z;
  st_tfman_opts istTfmOptions = ST_TFMAN_OPTS_Z;
  strarr_t* opts_files = NULL;


  /*
  *    === CL ARGUMENTS ===
  */

  /* --- Cl-args specification --- */

  cla_specs_initialize(&stArgSpecs, 128, 4);
    HandleErr(handle_err);

  strarr_t* M_vals = sarr_crev("scan", "parse", NULL);
  strarr_t* L_vals = sarr_crev("block", "tree", NULL);
  strarr_t* S_vals = sarr_crev( "name", "nameb", "nameid", "nameidb",
        "type", "typeb", "dirfst", "dirfstb", "size", "sizeb",
        "atime", "atimeb", "mtime", "mtimeb", "ctime", "ctimeb",
        "user", "userb", "group", "groupb", NULL);
  strarr_t* F_vals = sarr_crev("name", "type", "perms", "user", "group",
      "proper", "lntar", "size", "size_sparse", "atime", "mtime", "ctime",
      "entrs", "separator", NULL);
     HandleErr(handle_err);

  //general
  cla_opt_establish(&stArgSpecs, O_HELP, Behopt_on, "h", "help", 0, 0, NULL, false);
  cla_opt_establish(&stArgSpecs, O_VERBOSE, Behopt_on, "v", "verbose", 0, 0, NULL, false);
  cla_opt_establish(&stArgSpecs, O_QUIET, Behopt_on, "q", "quiet", 0, 0, NULL, false);
  cla_opt_establish(&stArgSpecs, O_MODE, Behopt_on, "M", "mode", 1, 1, M_vals, false);
  /*not used*/cla_opt_establish(&stArgSpecs, O_INDENT, Behopt_on, "I", "indent", 1, 1, NULL, false);
  cla_oper_establish(&stArgSpecs, OPER_A, 1, -1, NULL, false);

  //scan
  cla_opt_establish(&stArgSpecs, O_RECURS, Behopt_on, "R", "recursive", 0, 1, NULL, false);
  cla_opt_establish(&stArgSpecs, O_LAYOUT, Behopt_on, "L", "layout", 1, 1, L_vals, false);
  cla_opt_establish(&stArgSpecs, O_SORT, Behopt_on, "S", "sort", 1, 4, S_vals, false);
  cla_opt_establish(&stArgSpecs, O_FORMAT, Behopt_on, "F", "format", 1, 4, F_vals, false);
  cla_opt_establish(&stArgSpecs, O_CLONE_NAMES, Behopt_on, "C", "clone-names", 0, 0, NULL, false);
  //parse
  cla_opt_establish(&stArgSpecs, O_NAMES_ONLY, Behopt_on, "N", "names-only", 0, 0, NULL, false);
  cla_opt_establish(&stArgSpecs, O_SHOW_ONLY, Behopt_on, "s", "show-only", 0, 0, NULL, false);
  cla_opt_establish(&stArgSpecs, O_NO_ASK, Behopt_on, "n", "no-ask", 0, 0, NULL, false);
  cla_opt_establish(&stArgSpecs, O_LINES, Behopt_on, "l", "lines", 1, 2, NULL, false);
    HandleErr(handle_err);

  static const char* cstr_usage =
      "Scan:      tfman [-M scan] [DIR...] [OPTION...]\n"
      "Parse:     tfman [-M parse] [FILE...] [OPTION...]\n"
      "\n"
      "  scan ex.\n"
      "     'tfman /home/mydir /root'             - print text representation of both directories to standard output\n"
      "     'tfman /root > root.tfman'            - write text representation of '/root' to 'root.tfman' file, existent file will be overwriten\n"
      "     'tfman -M scan /root > root.tfman'    - same as above example\n"
      "  parse ex.\n"
      "     'tfman tr1.tfman tr2.tfman'           - perform operations read from 'tr1.tfman' and 'tr2.tfman' files\n"
      "     'tfman -M parse tr1.tfman tr2.tfman'  - same as above example\n"
      ;

   static const char* cstr_help =
      "General options:\n"
      " -v, --verbose                 Verbose execution.\n"
      " -q, --quiet                   Quiet execution, overrides '-v' option.\n"
      " -h, --help                    Display help and exit.\n"
      " -M, --mode=MODE               Mode of program operation: scan or parse. It is auto detected anyway.\n"
      "\nScan options:\n"
      " -R, --recursive[=LVL]         Scan directories recursively, stop at LVL level.\n"
      "                               For 0 (def.) don't stop.\n"
      " -S, --sort=FIELD[,FIELD]...   Sort entries with fields in given order.\n"
      "                               def.: type,name,size.\n"
      " -F, --format=FMT[,FMT]...     Compose entries from fields in given order.\n"
      "                               def.: name,type,perms,entrs,user,group,lntar,separator,\n"
      " -C, --clone-names             Duplicate name for each entry.\n"
      "\nParse options:\n"
      " -N, --names-only              Print name for each entry, do not perform any operation.\n"
      " -s, --show-only               Show operations, do not perform them.\n"
      " -n, --no-ask                  Do not ask for confirmation before performing operations.\n"
      " -l, --lines=N[,M]             Perform operations only for specified lines. M defaults to last line.\n"
      ;

   /* --- Parse cl arguments --- */
   cla_parse_clargs(&stArgSpecs, argv, argc, &Faults);
      HandleErr(handle_err);
   cla_validate_arguments(&stArgSpecs, &Faults);
      HandleErr(handle_err);
   if (Faults.ns > 0)  //: print faults if any
   {
      faultsarr_print(0, stderr, &Faults, NULL, "", "\n", NULL);
      goto handle_err;
   }

   //cla_print_set_arguments(&stArgSpecs);

   /*
    *    === OPTIONS ===
    */

   /* --- Default option values --- */

   //: order of sorting entries in text representation
   sort_en Sort_order[sort_count] = TFM_SORT_ORDER;
      size_t Sort_order_els = 0; while (Sort_order[Sort_order_els] != Sort_none) { ++Sort_order_els; }
   //: order of fields composing entry in text representation
   format_en Format[format_count] = TFM_ENTRY_FIELDS;
      size_t Format_els = 0; while (Format[Format_els] != Format_none) { ++Format_els; }

   tfmopts_initialize (
      &istTfmOptions,
   /* general */
      Mode_none,
      TFM_VERBOSITY_LVL,
      false,               //usage
      false,               //help
      TFM_DEF_INDENT,
   /* scan */
      false,               //append to file in scan mode
      TFM_RECURSION_LVL,
      Sort_order, Sort_order_els,
      Format, Format_els,
      TFM_DEF_LAYOUT,
      NULL,                //output files str array
      false,               //clone entry names
   /* parse */
      false,               //print entry names only and exit
      false,               //show actions only and exit
      TFM_ASK_USER,        //ask user before perfroming actions
      0,                   //start retrieving actions at this line
      0,                   //end retrieving actions at this line
      (const char**)argv, argc, NULL, 0
   );

   /* --- Set options --- */
   if (cla_opt_is_set(&stArgSpecs, O_VERBOSE)) {
      istTfmOptions.verb_lvl = 2;
   }

   if (cla_opt_is_set(&stArgSpecs, O_QUIET)) {
      istTfmOptions.verb_lvl = 0;
   }

   if (cla_opt_is_set(&stArgSpecs, O_HELP)) {
      istTfmOptions.usage = true;
      istTfmOptions.help = true;
   }

   if (cla_opt_is_set(&stArgSpecs, O_RECURS)) {
      const char* val = cla_ret_opt_value(&stArgSpecs, O_RECURS, 0);
      if (val == NULL) {
         istTfmOptions.recurs_lvl = 0; //1-no recurs (only top), 2-recurs 1 subdir level, 3-lvl2, etc.
      } else {
         istTfmOptions.recurs_lvl = atoi(val);
      }
   }


   if (cla_opt_is_set(&stArgSpecs, O_SORT))
   {
      const char* val = NULL; size_t j = 0;
      while ((val = cla_ret_opt_value(&stArgSpecs, O_SORT, j)))
      {
         if (strcmp(val, "name") == 0) {
            Sort_order[j] = Sort_name;
         } else if (strcmp(val, "nameb") == 0) {
            Sort_order[j] = Sort_nameb;
         } else if (strcmp(val, "nameid") == 0) {
            Sort_order[j] = Sort_nameidot;
         } else if (strcmp(val, "nameidb") == 0) {
            Sort_order[j] = Sort_nameidotb;
         } else if (strcmp(val, "type") == 0) {
            Sort_order[j] = Sort_type;
         } else if (strcmp(val, "typeb") == 0) {
            Sort_order[j] = Sort_typeb;
         } else if (strcmp(val, "size") == 0) {
            Sort_order[j] = Sort_size;
         } else if (strcmp(val, "sizeb") == 0) {
            Sort_order[j] = Sort_sizeb;
         } else if (strcmp(val, "atime") == 0) {
            Sort_order[j] = Sort_atime;
         } else if (strcmp(val, "atimeb") == 0) {
            Sort_order[j] = Sort_atimeb;
         } else if (strcmp(val, "mtime") == 0) {
            Sort_order[j] = Sort_mtime;
         } else if (strcmp(val, "mtimeb") == 0) {
            Sort_order[j] = Sort_mtimeb;
         } else if (strcmp(val, "ctime") == 0) {
            Sort_order[j] = Sort_ctime;
         } else if (strcmp(val, "ctimeb") == 0) {
            Sort_order[j] = Sort_ctimeb;
         } else if (strcmp(val, "user") == 0) {
            Sort_order[j] = Sort_user;
         } else if (strcmp(val, "userb") == 0) {
            Sort_order[j] = Sort_userb;
         } else if (strcmp(val, "group") == 0) {
            Sort_order[j] = Sort_group;
         } else if (strcmp(val, "groupb") == 0) {
            Sort_order[j] = Sort_groupb;
         } else {
            continue;
         }
         ++j;
      }
      Sort_order[j] = Sort_none;
      /*: Set sorting order in 'tfman' options: */
      tfmopts_sort_set(&istTfmOptions, Sort_order, j+1);
   }

   if (cla_opt_is_set(&stArgSpecs, O_FORMAT))
   {
      const char* val = NULL; size_t j = 0;
      while ((val = cla_ret_opt_value(&stArgSpecs, O_FORMAT, j)))
      {
         if (strcmp(val, "name") == 0) {
            Format[j] = Format_name;
         } else if (strcmp(val, "type") == 0) {
            Format[j] = Format_type;
         } else if (strcmp(val, "perms") == 0) {
            Format[j] = Format_perms;
         } else if (strcmp(val, "user") == 0) {
            Format[j] = Format_user;
         } else if (strcmp(val, "group") == 0) {
            Format[j] = Format_group;
         } else if (strcmp(val, "proper") == 0) {
            Format[j] = Format_props;
         } else if (strcmp(val, "lntar") == 0) {
            Format[j] = Format_lntarget;
         } else if (strcmp(val, "size") == 0) {
            Format[j] = Format_size;
         } else if (strcmp(val, "size_sparse") == 0) {
            Format[j] = Format_size2;
         } else if (strcmp(val, "atime") == 0) {
            Format[j] = Format_atime;
         } else if (strcmp(val, "mtime") == 0) {
            Format[j] = Format_mtime;
         } else if (strcmp(val, "ctime") == 0) {
            Format[j] = Format_ctime;
         } else if (strcmp(val, "entrs") == 0) {
            Format[j] = Format_entrs;
         } else if (strcmp(val, "separator") == 0) {
            Format[j] = Format_separ;
         } else {
            continue;
         }
         ++j;
      }
      Format[j] = Format_none;
      /*: Set format fields in 'tfman' options: */
      tfmopts_format_set(&istTfmOptions, Format, j+1);
   }

   if (cla_opt_is_set(&stArgSpecs, O_LAYOUT))
   {
      const char* val = cla_ret_opt_value(&stArgSpecs, O_LAYOUT, 0);
      if (strcmp(val, "block") == 0) {
         istTfmOptions.layout = Lay_block;
      } else if (strcmp(val, "tree") == 0) {
         istTfmOptions.layout = Lay_tree;
      } else {
         ___Sentinel("wrong layout value passed through\n"); abort();
      }
   }

   if (cla_opt_is_set(&stArgSpecs, O_INDENT))
   {
      const char* val = cla_ret_opt_value(&stArgSpecs, O_INDENT, 0);
      istTfmOptions.indent = atoi(val);
   }

   if (cla_opt_is_set(&stArgSpecs, O_MODE))
   {
      const char* val = cla_ret_opt_value(&stArgSpecs, O_MODE, 0);
      if (strcmp(val, "scan") == 0) {
         istTfmOptions.mode = Mode_scan;
      } else if (strcmp(val, "parse") == 0) {
         istTfmOptions.mode = Mode_parse;
      } else {
         ___Sentinel("wrong mode value passed through\n"); abort();
      }
   }

   if (cla_opt_is_set(&stArgSpecs, O_CLONE_NAMES)) {
      istTfmOptions.clone_names = true;
   }

   if (cla_opt_is_set(&stArgSpecs, O_SHOW_ONLY)) {
      istTfmOptions.show_only = true;
   }

   if (cla_opt_is_set(&stArgSpecs, O_NO_ASK)) {
      istTfmOptions.ask = false;
   }

   if (cla_opt_is_set(&stArgSpecs, O_LINES))
   {
      const char* val1 = cla_ret_opt_value(&stArgSpecs, O_LINES, 0);
      const char* val2 = cla_ret_opt_value(&stArgSpecs, O_LINES, 1);

      istTfmOptions.start_line = atoi(val1);

      if (val2 != NULL) {
         istTfmOptions.end_line = atoi(val2);
      } else {
         istTfmOptions.end_line = atoi(val1);
      }
   }

   if (cla_opt_is_set(&stArgSpecs, O_NAMES_ONLY)) {
      istTfmOptions.names_only = true;
   }




   /*
    * === ACTIONS ===
    */

   /*: --- Help/usage --- */
   if (print_usage_help_if_required(&istTfmOptions, cstr_usage, cstr_help) == true) {
      goto handle_succ;
   }

   /*: --- Program settings --- */

   /*: --- Actions --- */
   /*: Deduce mode from operand, only if mode is not yet set (it can be set with -M option): */
   if (istTfmOptions.mode == Mode_none)
   {
      const char* fst_operand = NULL;
      fst_operand = cla_ret_oper_value(&stArgSpecs, OPER_A, 0);
      if (fst_operand != NULL && stat_file_is_dir(fst_operand)) {
         istTfmOptions.mode = Mode_scan;
      } else {
         istTfmOptions.mode = Mode_parse;
      }
         HandleErrm(handle_err, "Can not open: %s", fst_operand);
   }

   mode_en action = istTfmOptions.mode;         //: mode of operation

  opts_files = sarr_cre2(0);                    //: names of files opened for parsing
     HandleErr(handle_err);

   /*VERB*/    if (istTfmOptions.verb_lvl >= 2) {
                  int ret = fprintf(stdout, "starting '%s - %s'\n\n", TFMAN_PR_FULLNAME, TFMAN_PR_LONGNAME);
                     ErrorTrigIf(ret < 0, handle_err);
               }

   switch (action)
   {
      /*: --- Scan --- */
      case Mode_scan:{

         st_fsimg istFsImage = {0};
         st_fstxt istFsText = {0};
         FILE* Fss[TFMAN_OPEN_FILES_MAX+1] = {stdout, NULL};

         /*: Set strings (marks) used in text representation: */
         fstxt_set_pref_strings(&istFsText, TFM_SCAN_PREFIXES);
         fsrxt_set_field_strings(&istFsText, TFM_SCAN_MARKERS);
            HandleErr(scan_finish);

         /*: Scan each directory given as operand on cl: */
         size_t nval = 0;           //: cl-arg value number
         const char* clval = NULL;  //: cl-arg current value
         while ((clval = cla_ret_oper_value(&stArgSpecs, OPER_A, nval++)) != NULL)
         {
            scan_directory_tree(&istTfmOptions, &istFsImage, &istFsText, clval, Fss);
               HandleErr(scan_finish);
         }

         /*: Finish Scanning: */
         close_file_streams(Fss);
         repimg_cleanup(&istFsImage);    // cleanup all
         fstxt_cleanup(&istFsText);     // cleanup all
         goto handle_succ;

        scan_finish:
         close_file_streams(Fss);
         repimg_cleanup(&istFsImage);    // cleanup all
         fstxt_cleanup(&istFsText);     // cleanup all
         goto handle_err;

      }break;

      /*: --- Parse --- */
      case Mode_parse:{

         st_parser_main istParserMain = ST_PARSER_MAIN_Z;
         FILE* Fss[TFMAN_OPEN_FILES_MAX+1] = { stdin, NULL};

         istParserMain.ind_width = istTfmOptions.indent;

         /*: Open input files: */
         uint n_files = 1;    // num of opened files
         open_input_files(Fss, &n_files, opts_files, &stArgSpecs, TFMAN_OPEN_FILES_MAX);
            HandleErr(parse_finish);

         /*: Parse each file: */
         for (uint j = 0; j < n_files; ++j)
         {
            pars_cleanup_parser_main_next_run(&istParserMain);

            const char* filename = NULL;
            if (Fss[j] == stdin) { filename = "'stdin'"; }
            else { filename = sarr_el_ret(opts_files, j); }
            parse_fsrep_file(&istParserMain, Fss[j], filename,
                              istTfmOptions.show_only, istTfmOptions.ask,
                              istTfmOptions.start_line, istTfmOptions.end_line,
                              istTfmOptions.names_only);
               HandleErr(parse_finish);
         }

         /*: Finish Parsing: */
         if (Fss[0] != stdin) { close_file_streams(Fss); }
         pars_cleanup_parser_main(&istParserMain);
         goto handle_succ;

        parse_finish:
         if (Fss[0] != stdin) { close_file_streams(Fss); }
         pars_cleanup_parser_main(&istParserMain);
         goto handle_err;

      }break;

      case Mode_none:  /*- \/ through -*/
      case mode_count:
      default: ___Sentinel("wrong mode value detected\n"); abort();
   }




   /*: Finish: */
  handle_succ:
   faultsarr_finalize(&Faults);
   if (istTfmOptions.ofiles == NULL && opts_files != NULL) { sarr_des(&opts_files); }
   cla_specs_cleanup(&stArgSpecs);
   tfmopts_cleanup(&istTfmOptions);
   if (M_vals != NULL) { sarr_des(&M_vals); }
   if (L_vals != NULL) { sarr_des(&L_vals); }
   if (S_vals != NULL) { sarr_des(&S_vals); }
   if (F_vals != NULL) { sarr_des(&F_vals); }
   return EXIT_SUCCESS;

  handle_err:
   CatchError(E_ALL_TYPES, TFM_ERR_PRINT) {
      fprintf(stderr, "%s: %s, terminating\n", ErrDescr(), ErrMessage());
      goto handle_err_exit;
   }

  handle_err_exit:
   faultsarr_finalize(&Faults);
   if (stArgSpecs.opt_specs.ns > 0) { cla_specs_cleanup(&stArgSpecs); }
   if (istTfmOptions.ofiles == NULL && opts_files != NULL) { sarr_des(&opts_files); }
   tfmopts_cleanup(&istTfmOptions);
   if (M_vals != NULL) { sarr_des(&M_vals); }
   if (L_vals != NULL) { sarr_des(&L_vals); }
   if (S_vals != NULL) { sarr_des(&S_vals); }
   if (F_vals != NULL) { sarr_des(&F_vals); }
   return EXIT_FAILURE;
}
