/*!F::Options management for 'tfman' - main implementation file.
 *
 *  ::Purpose:
 *    Establish, parse, read nad set options used by 'tfman'.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - Initialize, set, get options
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of 'Tfman' program.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-12
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_tfman.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "cstrings.h"
#include "strings_arr.h"

/* libc */

/* current module */
#include "opts.h"
#include "opts_internal.h.inc"
//#include "opts_aux.h.inc"




/* Init options. */
/* -------------------------------------------------------------------------- */

void
tfmopts_initialize (
   st_tfman_opts* stOpts,
   mode_en mode,
   uint verb_lvl,
   bool usage,
   bool help,
   uint indent,
   bool append,
   uint recurs_lvl,
   sort_en* sort,
   uint els_s,
   format_en* format,
   uint els_f,
   uint layout,
   strarr_t* ofiles,
   bool clone_names,
   bool names_only,
   bool show_only,
   bool ask,
   uint start_line,
   uint end_line,
   const char** argv_cl,
   int argc_cl,
   const char** argv_cf,
   int argc_cf
){
   ___DParam(stOpts, != NULL);
   ___DParam(sort, != NULL);
   ___DParam(format, != NULL);
   stOpts->mode = mode;
   stOpts->verb_lvl = verb_lvl;
   stOpts->usage = usage;
   stOpts->help = help;
   stOpts->indent = indent;
   stOpts->append = append;
   stOpts->recurs_lvl = recurs_lvl;
   stOpts->stSort.ESort = sort; stOpts->stSort.els = els_s;
   stOpts->stFormat.EFormat = format; stOpts->stFormat.els = els_f;
   stOpts->layout = layout;
   stOpts->ofiles = ofiles;
   stOpts->clone_names = clone_names;
   stOpts->names_only = names_only;
   stOpts->show_only = show_only;
   stOpts->ask = ask;
   stOpts->start_line = start_line;
   stOpts->end_line = end_line;
   stOpts->argv[0] = argv_cl;
   stOpts->argv[1] = argv_cf;
   stOpts->argc[0] = argc_cl;
   stOpts->argc[1] = argc_cf;
   return;
}

/*!:: Release all resources and assign empty values to struct members.
 * ::Preconditions:
 *    - `stOpts` must be already created
 * ::Errors:
 *    No errors.
 */
void
tfmopts_cleanup (
   st_tfman_opts* stOpts      //! [out] 'tfman' options structure
){
   ___Dassert(stOpts, "'stOpts' parameter invalid");
   return;
}




/* Set / get options. */
/* -------------------------------------------------------------------------- */


void
tfmopts_sort_set (
   st_tfman_opts* stOpts,     //! [out] 'tfman' options structure
   sort_en* ESort,            //! array of fields used for entries sorting in scan mode
   size_t els_arr             //! number of elements in array
){
   ___Dassert(stOpts, "'stOpts' parameter invalid");
   stOpts->stSort.ESort = ESort;
   stOpts->stSort.els = els_arr;
   return;
}

void
tfmopts_format_set (
   st_tfman_opts* stOpts,     //! [out] 'tfman' options structure
   format_en* EFormat,        //! array of fields used for entries composing in scan mode
   size_t els_arr             //! number of elements in array
){
   ___Dassert(stOpts, "'stOpts' parameter invalid");
   stOpts->stFormat.EFormat = EFormat;
   stOpts->stFormat.els = els_arr;
   return;
}

void
tfmopts_file_set (
   st_tfman_opts* stOpts,     //! [out] 'tfman' options structure
   strarr_t* ofiles     //! output files
){
   ___Dassert(stOpts, "'stOpts' parameter invalid");
   stOpts->ofiles = ofiles;
   return;
}
