#ifndef TFMAN_SRCMAIN_OPTS_H
#define TFMAN_SRCMAIN_OPTS_H


/*!F::Options management for 'tfman' - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* current project */
#include "common_tfman.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "strings_arr.h"




/* Data. */
/* -------------------------------------------------------------------------- */


/*!:: Mode of program operation.
 *    Which mode program is invoced in.
 */
enum mode_en {
   Mode_none=0,   //! empty value
   Mode_parse,    //! parsing text representation file
   Mode_scan,     //! scannig file system
   mode_count };
typedef enum mode_en mode_en;

/*!:: How option was set.
 *    Source of particular option set.
 */
enum tfmopt_how_set_en {
   Tfmopt_notset,       //! yet not set
   Tfmopt_set_def,      //! default value
   Tfmopt_set_cl,       //! command line argument
   Tfmopt_set_cf,       //! config file
   Tfmopt_set_other,    //! other way
   tfmopt_how_set_count };
typedef enum tfmopt_how_set_en tfmopt_how_set_en;




/* ...Text.organization...................................................... */


/*!:: Layout of text representation.
 *    How textual representation created by filesystem scan should be visually
 *    organized.
 */
enum layout_en {
   Lay_none=0,    //! empty value
   Lay_block,     //! block - each directory is monolithic block, subdirs are separate blocks only indented
   Lay_tree,      //! tree - subdirs are organized as hierarchy tree
   layout_count };
typedef enum layout_en layout_en;


/*!:: Sort of entries.
 *    Field names used to sort entries in text representation created by
 *    filesystem scan.
 */
enum sort_en {
   Sort_none=0,   //! empty value
   Sort_name,     //! name
   Sort_nameb,    //! name backwards
   Sort_nameidot, //! name, ignore . at begining
   Sort_nameidotb,//! name backwards, ignore . at begining
   Sort_type,     //! entry type
   Sort_typeb,    //! as above but backwards
   Sort_dirfst,   //! dirs first, rest as is
   Sort_dirfstb,  //! as above but backwards
   Sort_size,     //! size
   Sort_sizeb,    //! as above but backwards
   Sort_atime,    //! access time
   Sort_atimeb,   //! as above but backwards
   Sort_mtime,    //! modify time
   Sort_mtimeb,   //! as above but backwards
   Sort_ctime,    //! change time
   Sort_ctimeb,   //! as above but backwards
   Sort_user,     //! user owner
   Sort_userb,    //! as above but backwards
   Sort_group,    //! group owner
   Sort_groupb,   //! as above but backwards
   sort_count };
typedef enum sort_en sort_en;

/*!:: Format of entries.
 *    Fields to compose entries in text representation created by filesystem
 *    scan.
 */
enum format_en {
   Format_none=0,       //! empty value
   Format_name,         //! name
   Format_type,         //! type
   Format_perms,        //! permissions
   Format_user,         //! user
   Format_group,        //! group
   Format_props,        //! properties for entry (bitrate etc.)
   Format_lntarget,     //! symlink target
   Format_size,         //! size
   Format_size2,        //! size with sparse areas and meta info
   Format_atime,        //! access time
   Format_mtime,        //! modify time
   Format_ctime,        //! change time
   Format_entrs,        //! num of entries in directory
   Format_separ,        //! separator
   format_count };
typedef enum format_en format_en;




typedef struct st_ent_sort st_ent_sort;
/*!:: Struct with order of entry sorting.
 *    Fields used to sort each entry in text representation. They are used in
 *    order of appearance in array.
 */
struct st_ent_sort
{
   sort_en* ESort;         //! array of sort fields
   uint els;             //! num of elements in array
};
#define ST_ENT_SORT_Z {NULL,0}


typedef struct st_ent_format st_ent_format;
/*!:: Struct with order of entry fields.
 *    Fields used to compose each entry in text representation. They are used in
 *    orfer of appearance in array.
 */
struct st_ent_format
{
   format_en* EFormat;     //! array of composing fields
   uint els;             //! num of elements in array
};
#define ST_ENT_FORMAT_Z {NULL,0}




/* ...Options................................................................ */


/*!:: How cloned name is composed.
 *    It may be clean or preceeded with specified action tag.
 */
enum clone_how_en {
  Clonehow_none=0,    //! no tag
  Clonehow_mv,        //! move
  Clonehow_mvo,       //! move overwrite
  Clonehow_cp,        //! copy
  Clonehow_cpo,       //! copy overwrite
  clonehow_count };
typedef enum clone_how_en clone_how_en;


typedef struct st_tfman_opts st_tfman_opts;
/*!:: 'Tfman' options.
 *    Set of all options used by 'tfman'.
 */
struct st_tfman_opts
{
   /* general */
   mode_en mode;          //! mode of program operation: 0-scan 1-parse
   uint verb_lvl;      //! program execution verbosity level
   bool usage;         //! print uage information
   bool help;          //! print help
   uint indent;        //! num of characters used in subdirs indentation
   /* scan */
   bool append;        // append to output file or overwrite
   uint recurs_lvl;    //! recursion level - how deep scan directory
                              //. 0-all the way, 1-only top dir, 2+-1 and more lvl deep
   st_ent_sort stSort;        //! entries sort order
   st_ent_format stFormat;    //! entries compose format
   uint layout;        //! text representation layout in scan mode: 0-block 1-tree
   strarr_t* ofiles;    //! files to output text representation to
   bool clone_names;    //! clone name in each line of text representation
   clone_how_en clone_how;  //! prepend action to cloned name: 0-no, 1-move, 2-move_overw, 3-copy, 4-copy_overw
   /* parse */
   bool names_only;     //! print entry names only and exit
   bool show_only;      //! show actions only, do not perform
   bool ask;            //! ask before perform actions
   uint start_line;     //! start retrieving actions at this line
   uint end_line;       //! end retrieving actions at this line
   /* admin */
   const char** argv[2];//! 0-cl options, 1-conf file options
   int argc[2];         //! as above
};

/*!:: Initializer for 'st_tfman_opts' struct. Example:
 *       'st_tfman_opts st = ST_TFMAN_OPTS_Z;'   */
#define ST_TFMAN_OPTS_Z {Mode_none,0,false,false,0,false,0, \
               ST_ENT_SORT_Z,ST_ENT_FORMAT_Z,0,NULL,false,Clonehow_none,false,false,false,0,0, \
               {NULL,NULL},{0,0}}




/* Init options. */
/* -------------------------------------------------------------------------- */


/*!1:: Initialize struct with 'tfman' options.  */
extern void tfmopts_initialize (st_tfman_opts* stOpts, mode_en mode, uint verb_lvl,
                  bool usage, bool help, uint indent, bool append, uint recurs_lvl,
                  sort_en* sort, uint els_s, format_en* format, uint els_f,
                  uint layout, strarr_t* ofiles, bool clone_names,
                  bool names_only, bool show_only, bool ask,
                  uint start_line, uint end_line,
                  const char** argv_cl, int argc_cl, const char** argv_cf, int argc_cf)
               __attr_nonnull__(1,9,11)

               ;

/*!1:: Cleanup struct with 'tfman' options.  */
extern void tfmopts_cleanup (st_tfman_opts* stOpts)
               __attr_nonnull__(1)
               ;




/* Set / get options. */
/* -------------------------------------------------------------------------- */


extern void tfmopts_sort_set (st_tfman_opts* stOpts, sort_en* sort_arr,
         size_t els_arr)
               __attr_nonnull__(1)
               ;

extern void tfmopts_format_set (st_tfman_opts* stOpts,
         format_en* format_arr, size_t els_arr)
               __attr_nonnull__(1)
               ;

extern void tfmopts_file_set (st_tfman_opts* stOpts, strarr_t* ofiles)
               __attr_nonnull__(1)
               ;




#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool opts_all_tests_run (int argc, char* argv[])
      __attr_wures__;
#endif




#endif /*- TFMAN_SRCMAIN_OPTS_H -*/
