/*!F::Transformator data-actions - main implementation file.
 *
 *  ::Purpose:
 *    Create file system actions from data structures representing part of file
 *    system described in given textual representation.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of 'Tfman' program.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-05-08
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_tfman.h"
#include "parse_img.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "cstrings.h"   //torem
#include "debug.h"
#include "faults.h"
#include "memory.h"
#include "strings_arr.h"
#include "string_operations.h"
#include "dstream_info.h"

/* libc */

/* current module */
#include "parse_acts.h"
#include "parse_acts_internal.h.inc"
#include "parse_acts_aux.h.inc"
#include "parse_img_internal.h.inc" //torem



/* Auxillary struct for pars_txtrepdata_to_actions(). */
typedef struct {
   arr_ptrs_t arrTags;     //: words with actions (st_word_ch)
   strarr_t entries_be;    //: names resulting from brace expansion
   strarr_t names_prep;    //: names resulting from prepending parent dirs
   strarr_t names_glob;    //: names resulting from globbing
} st_parser_line;
#define ST_PARSER_LINE_Z {ARR_PTRS_T_Z,STRARR_T_Z,STRARR_T_Z,STRARR_T_Z}


void
pars_txtrepdata_to_actions (
   st_actions_table* stTable,    //! [out] file system actions
   st_txt_img* stTxtRep,         //! textual representation
   st_parser_main* stParsMain,   //! main parser structure
   uint start_line,              //! retrieving actions start, 0-no limit
   uint end_line                 //! retrieving actions end, 0-no limit
){
   ___Dassert(stTable, "'stTable' parameter invalid");
   ___Dassert(stTxtRep, "'stTxtRep' parameter invalid");
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");


   st_parser_line PL = ST_PARSER_LINE_Z;


   /* ....................................................................... */
   /*: WORK-OUT EACH LINE CHARACTERISTICS: */

   uint i_lnch = 0;
   st_line_ch* el_stLnCh = NULL;
   while ((el_stLnCh = arrptrs_it(&stTxtRep->arrEntries, &i_lnch)) != NULL)
   {

      /*: Tags processing: */
      /* .................................................................... */

      /*: Entries prepare: */
      str_expand_braces(&PL.entries_be, &el_stLnCh->ename);
         HandleErr(handle_err);

      arrptrs_ini2(&PL.arrTags, acttype_count, NULL);
         HandleErr(handle_err);

      //: Retrieve actions only if line's group and parent dir is not disabled
      linemod_en grp_dis = el_stLnCh->p_grp->p_lmod_def->mod;
      linemod_en dir_dis = Linetype_none; if (el_stLnCh->p_dir != NULL) { dir_dis = el_stLnCh->p_dir->p_lmod_def->mod; }

      if (grp_dis != Linemod_disable && dir_dis != Linemod_disable
         && el_stLnCh->p_lmod_def->mod != Linemod_disable)
      {

      bool excl_group = false;    //: exclude entry from containig group?
      bool excl_dir = false;      //: exclude entry from containig dir?
      bool doubledot_dir = false; //: entry is '..'

      linemod_en lmod = el_stLnCh->p_lmod_def->mod;
      if (lmod == Linemod_excl_grp || lmod == Linemod_excl_both) { excl_group = true; }
      if (lmod == Linemod_excl_dir || lmod == Linemod_excl_both) { excl_dir = true; }

      if (strcmp(el_stLnCh->ename.buf, "..") == 0
       || strcmp(el_stLnCh->ename.buf, "../") == 0)
      {
        doubledot_dir = true;
      }

      /*: Line actions: */
      retrieve_tags_line(&PL.arrTags, el_stLnCh);

      /*: Group actions */
      if (doubledot_dir == false && excl_group == false)
      retrieve_tags_group(&PL.arrTags, el_stLnCh);

      /*: Directory actions */
      if (doubledot_dir == false && excl_dir == false)
      retrieve_tags_dir(&PL.arrTags, el_stLnCh);

      }


      /*: Names processing: */
      /* .................................................................... */

      st_line_ch* par_dir = el_stLnCh->p_dir;

      strarr_t* pdir_names = NULL;
      if (par_dir == NULL) {
         pdir_names = sarr_cre("");
            HandleErr(MAJOR_PROC_ERR);
      } else {
         pdir_names = &el_stLnCh->p_dir->fullnames;
      }

      /*: For each directory name: */
      for (size_t i = 0; i < pdir_names->ns; ++i)
      {
         str_t* el_dname = sarr_el_rets(pdir_names, i);

         //: prepend parent dirs if applicable:
         prepend_parent_dirs_to_names (&PL.names_prep, &PL.entries_be,
                                          el_stLnCh, el_dname);
            HandleErr(MAJOR_PROC_ERR);

         //: globbing (on all names resulted from prepending)
         glob_all_names(&PL.names_glob, &PL.names_prep);
            HandleErr(MAJOR_PROC_ERR);

         /*cu*/ sarr_fin(&PL.names_prep);

         //: create actions
         if ((start_line == 0 || (el_stLnCh->lnum_in_file >= start_line))
          && (end_line == 0 || (el_stLnCh->lnum_in_file <= end_line)) )
         {
            //: line is within given range:
            create_actions_for_all_words(stTable, el_stLnCh, &PL.names_glob, &PL.arrTags, el_dname);
               HandleErr(MAJOR_PROC_ERR);
         }

         sarr_appA(&el_stLnCh->fullnames, &PL.names_glob);

         /*cu*/ sarr_fin(&PL.names_glob);
      }
      /*cu*/ if (par_dir == NULL) { sarr_des(&pdir_names); }

      /*cu*/ sarr_fin(&PL.entries_be);
      /*cu*/ arrptrs_fin(&PL.arrTags, NULL);
      continue;

     MAJOR_PROC_ERR:
      if (par_dir == NULL) { sarr_des(&pdir_names); }
      goto handle_err;
   }

   return;

  handle_err:
   sarr_fin(&PL.names_prep);
   sarr_fin(&PL.entries_be);
   arrptrs_fin(&PL.arrTags, NULL);
   return;
}

void
pars_cleanup_actstable (
   st_actions_table* stTable
){
   ___Dassert(stTable, "'stTable' parameter invalid");

   { /*--*/
      uint ii = 0;
      st_action* el_act = NULL;
      while ((el_act = arrptrs_it(&stTable->arrActKindsPrior, &ii)) != NULL) {
         act_destr_(&el_act);
      }
      arrptrs_fin(&stTable->arrActKindsPrior, NULL);
   } /*--*/

   for (size_t i = 0; i < acttype_count; ++i)
   {
      arr_ptrs_t* parrActs = &stTable->arrActKinds[i];   //helper var

      uint ii = 0;
      st_action* el_act = NULL;
      while ((el_act = arrptrs_it(parrActs, &ii)) != NULL) {
         act_destr_(&el_act);
      }

      arrptrs_fin(parrActs, NULL);
   }
}

void
pars_txtrepdata_to_names (
   strarr_t* Names,              //! [out] entry names
   st_txt_img* stTxtRep,         //! textual representation
   st_parser_main* stParsMain,   //! main parser structure
   uint start_line,              //! retrieving actions start, 0-no limit
   uint end_line                 //! retrieving actions end, 0-no limit
){
   ___Dassert(Names, "'Names' parameter invalid");
   ___Dassert(stTxtRep, "'stTxtRep' parameter invalid");
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");


   st_parser_line PL = ST_PARSER_LINE_Z;


   /* ....................................................................... */
   /*: WORK-OUT EACH LINE CHARACTERISTICS: */

   uint i_lnch = 0;
   st_line_ch* el_stLnCh = NULL;
   while ((el_stLnCh = arrptrs_it(&stTxtRep->arrEntries, &i_lnch)) != NULL)
   {

      /*: Initial processing: */
      /* .................................................................... */

      /*: Entries prepare: */
      str_expand_braces(&PL.entries_be, &el_stLnCh->ename);
         HandleErr(handle_err);

      /*: Major processing: */
      /* .................................................................... */

      st_line_ch* par_dir = el_stLnCh->p_dir;

      strarr_t* pdir_names = NULL;
      if (par_dir == NULL) {
         pdir_names = sarr_cre("");
            HandleErr(MAJOR_PROC_ERR);
      } else {
         pdir_names = &el_stLnCh->p_dir->fullnames;
      }

      /*: For each directory name: */
      for (size_t i = 0; i < pdir_names->ns; ++i)
      {
         str_t* el_dname = sarr_el_rets(pdir_names, i);

         //: prepend parent dirs if applicable:
         prepend_parent_dirs_to_names (&PL.names_prep, &PL.entries_be,
                                          el_stLnCh, el_dname);
            HandleErr(MAJOR_PROC_ERR);

         //: globbing (on all names resulted from prepending)
         glob_all_names(&PL.names_glob, &PL.names_prep);
            HandleErr(MAJOR_PROC_ERR);

         /*cu*/ sarr_fin(&PL.names_prep);

         //: add names to result array
         if ((start_line == 0 || (el_stLnCh->lnum_in_file >= start_line))
          && (end_line == 0 || (el_stLnCh->lnum_in_file <= end_line))
         )  //: if line is within given range:
         {
            sarr_appA(Names, &PL.names_glob);
         }

         sarr_appA(&el_stLnCh->fullnames, &PL.names_glob);

         /*cu*/ sarr_fin(&PL.names_glob);
      }
      /*cu*/ if (par_dir == NULL) { sarr_des(&pdir_names); }

      /*cu*/ sarr_fin(&PL.entries_be);
      /*cu*/ arrptrs_fin(&PL.arrTags, NULL);
      continue;

     MAJOR_PROC_ERR:
      if (par_dir == NULL) { sarr_des(&pdir_names); }
      goto handle_err;
   }

   return;

  handle_err:
   sarr_fin(&PL.names_prep);
   sarr_fin(&PL.entries_be);
   arrptrs_fin(&PL.arrTags, NULL);
   return;
}
