/*!F::Parser of file system text representation - main implementation file.
 *
 *  ::Purpose:
 *    Parse text file containig text representation of directory or whole
 *    directories tree. Create data structures representing part of file
 *    system described in given textual representation.
 *
 *    Created data structures will serve to deduct file system actions to
 *    perform on entires.
 *
 *    Actions are marked in file content by special 'action strings'.
 *    Actions are deduced only for lines marked with 'effective strings'.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - read to-be parsed file to file text struct
 *    - parse file text struct to create data representing file system
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of 'Tfman' program.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-12
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* isblank():
 *       _XOPEN_SOURCE >= 600 || _ISOC99_SOURCE ||
 *       _POSIX_C_SOURCE >= 200112L;
 *       or cc -std=c99
 */




/* current project */
#include "common_tfman.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "faults.h"
#include "memory.h"
#include "string_operations.h"
#include "cstrings.h"
#include "strings.h"
#include "strings_arr.h"
#include "arrptrs.h"
#include "fstream_read.h"

/* libc */
#include <ctype.h>      // isblank()
#include <string.h>     // memset(), strcmp()
#include <stdio.h>      // printf(), putc()

/* current module */
#include "parse_img.h"
#include "parse_img_internal.h.inc"
#include "parse_img_aux.h.inc"




// name, line kind, line type, tag, tag length
const st_line_def StLineDefinitions[linetype_count] = {
  {"NONE          ", Linekind_none, Linetype_none,          "",   0},
  {"directory     ", Linekind_normal, Linetype_dir,         "#",  1},
  {"normal entry  ", Linekind_normal, Linetype_entry,       "|",  1},
  {"priority entry", Linekind_normal, Linetype_prior,       "$",  1},
  {"group beg     ", Linekind_nocontent, Linetype_grp_beg,  "{",  1},
  {"group end     ", Linekind_nomods, Linetype_grp_end,     "}",  1},
  {"group dir end ", Linekind_nocontent, Linetype_grpd_beg, "{{", 2},
  {"group dir end ", Linekind_nomods, Linetype_grpd_end,    "}}", 2},
  {"ignore errors ", Linekind_alone, Linetype_ign_errors,   "!!", 2},
  {"end parsing   ", Linekind_alone, Linetype_parsing_end,  "END",3}
};

// name, mod type, tag, tag length
const st_lmod_def StLinemodsDefinitions[linemod_count] = {
  {"NONE              ", Linemod_none,      "",   0},
  {"exclude from dir  ", Linemod_excl_dir,  "=",  1},
  {"exclude from group", Linemod_excl_grp,  "-",  1},
  {"exclude from both ", Linemod_excl_both, "E",  1},
  {"disable whole     ", Linemod_disable,   "x",  1}
};

// name, action kind, action type, tag, tag length
const st_act_def StActionDefinitions[acttype_count] = {
  {"no action        ", Actkind_none, Acttype_none,        "",    0},
  {"create           ", Actkind_nooper, Acttype_create,    "~",   1},
  {"slink            ", Actkind_oper, Acttype_slink,       "->",  2},
  {"slink overw      ", Actkind_oper, Acttype_slink_o,     "->*", 3},
  {"hlink            ", Actkind_oper, Acttype_hlink,       "=>",  2},
  {"hlink overw      ", Actkind_oper, Acttype_hlink_o,     "=>*", 3},
  {"change perms     ", Actkind_inactive, Acttype_chperms,     ":",    0},
  {"change user      ", Actkind_inactive, Acttype_chuser,      ":",    0},
  {"change group     ", Actkind_inactive, Acttype_chgroup,     ":",    0},
  {"change prop      ", Actkind_multioper, Acttype_prop,   ":",   1},
  {"copy             ", Actkind_oper, Acttype_copy,        ">",   1},
  {"copy overw       ", Actkind_oper, Acttype_copy_o,      ">*",  2},
  {"move             ", Actkind_oper, Acttype_move,        ">>",  2},
  {"move overw       ", Actkind_oper, Acttype_move_o,      ">>*", 3},
  {"swap names       ", Actkind_oper, Acttype_swnames,     "<>",  2},
  {"to trash         ", Actkind_nooper, Acttype_trash,     "+",   1},
  {"remove           ", Actkind_nooper, Acttype_remove,    "*",   1},
  {"COUNT            ", Actkind_none, opertype_count,       "",   0},
  {"name beg         ", Actkind_name, Actspec_name_beg,    "[",   1},
  {"name end         ", Actkind_name, Actspec_name_end,    "]",   1},
  {"dir content      ", Actkind_clear, Actspec_da_cont,    "=",   1},
  {"dir both         ", Actkind_clear, Actspec_da_both,    "==",  2},
  {"disable actions  ", Actkind_clear, Actspec_da_ignor,   "||",  2},
  {"ignore errors    ", Actkind_clear, Actspec_ign_errors, "!",   1},
  {"actions on target", Actkind_clear, Actspec_acts_target,"++",  2},
  {"DUMMY            ", Actkind_dummy, Actdummy_scan_slink,  "_>",   2},
  {"DUMMY            ", Actkind_dummy, Actdummy_scan_errlink,"!!>",  3},
  {"COUNT            ", Actkind_none, spectype_count,      "",    0}
};






/* Prepare routines. */
/* -------------------------------------------------------------------------- */



/*!:: Will clean all data inside struct..
 * ::Errors:
 *    No errors.
 */
void
pars_cleanup_parser_main (
   st_parser_main* stParsMain    //! main parser structure
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   parsmain_clean_(stParsMain);
   return;
}

/*!:: Will leave effective tags and action tags untouched. Rest is cleaned.
 * ::Errors:
 *    No errors.
 */
void
pars_cleanup_parser_main_next_run (
   st_parser_main* stParsMain    //! main parser structure
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   parsmain_clean_for_next_parse_(stParsMain);
   return;
}




/* Text rep data. */
/* -------------------------------------------------------------------------- */


/*!:: Textual representation is read from given file stream. It must conform to
 *    certain format described in documentation. It is transformed into data
 *    structures. Created data is suitable to deduct file system actions from it.
 * ::Preconditions:
 *    - `fs` must be opened for reading
 * ::Errors:
 *    ENOMEM               -- no memory available
 *    readin_parsed_file() -- this function errors
 *    assemble_groups_of_entries_to_txtrep() -- this function errors
 */
void
pars_prepare_txtrep (
   st_txt_img* stTxtRep,   //! [out] textual representation
   FILE* fs,               //! file stream
   const char* filename,   //! file name
   st_parser_main* stParsMain //! main parser structure
){
   ___Dassert(stTxtRep, "'stTxtRep' parameter invalid");
   ___Dassert(fs, "'fs' parameter invalid");
   ___Dassert(filename, "'filename' parameter invalid");
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");

   /*: Read file content: */
   strarr_t file_content = {0};  //lines read from file
   readin_parsed_file(&file_content, fs, filename);
      HandleErr(handle_err);

   /*: Create data representing file sys: */
   assemble_groups_of_entries_from_txtrep(stTxtRep, &file_content, stParsMain);
      HandleErr(handle_err);

   sarr_fin(&file_content);

   /*: Finish: */
   return;

  handle_err:
   txtrep_clean_(stTxtRep);
   sarr_fin(&file_content);
   return;
}

/*!:: Release memory.
 * ::Errors:
 *    No errors.
 */
void
pars_cleanup_txtrep (
   st_txt_img* stTxtRep    //! textual representation
){
   ___Dassert(stTxtRep, "'stTxtRep' parameter invalid");
   txtrep_clean_(stTxtRep);
   return;
}
