#ifndef TFMAN_SRCPARS_TXTREPTODATA_H
#define TFMAN_SRCPARS_TXTREPTODATA_H


/*!F::Parser of file system text representation - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* current project */
#include "common_tfman.h"

/* mh library */
#include "core.h"
#include "faults.h"
#include "strings.h"
#include "strings_arr.h"
#include "arrptrs.h"

/* libc */

/* current module */




/* Data. */
/* -------------------------------------------------------------------------- */


typedef struct st_group_ch st_group_ch;


/* ...Text.representation.....................................*/

/*!:: Kind of line.
 *    Does it take modification, does it have content (except tag)
 */
enum linekind_en {
  Linekind_none=0,
  Linekind_normal,    //! takes mod and has content, ex. #- line...
  Linekind_nomods,    //! no mods, has content, ex. } line
  Linekind_nocontent, //! takes mod, no content, ex. {x
  Linekind_alone,     //! no mods, no content, ex. END
  linekind_count};
typedef enum linekind_en linekind_en;

/*!:: Line tags
 *    Tags indicating type of line.
 * ::Usage in textual representation:
 *    String representing effective tag must be placed as first non-blank
 *    character in each line that should be parsed. If effective tag is not
 *    present line is ignored.
 * ::Tags:
 */
enum linetype_en {
   Linetype_none=0,       //! none
   Linetype_dir,       //! directory swap (block layout)
   Linetype_entry,        //! single entry
   Linetype_prior,  //! single priority entry
   Linetype_grp_beg,      //! group begin
   Linetype_grp_end,      //! group end
   Linetype_grpd_beg,     //! group begin - will include dirs
   Linetype_grpd_end,     //! group end - will include dirs
   Linetype_ign_errors,   //! do not terminate on errors - toggle
   Linetype_parsing_end,  //! end parsing now
   linetype_count };
typedef enum linetype_en linetype_en;


typedef struct st_line_def st_line_def;
/*!:: Line definition.
 *    Properties of single line (line template).
 */
struct st_line_def
{
  const char* const name; //! name
  linekind_en kind;       //! line kind
  linetype_en type;       //! line type
  const char* const tag;  //! tag string
  const size_t tag_len;   //! num of chars in tag
};
extern const st_line_def StLineDefinitions[linetype_count];


/*!:: Modifiers of line tags.
 *    Modify behavior of line tags.
 * ::Usage in textual representation:
 *    It must be placed immediately after effective tag.
 */
enum linemod_en {
   Linemod_none=0,       //! none
   Linemod_excl_dir,     //! exclude entry from parent dir
   Linemod_excl_grp,     //! exclude entry from containing group
   Linemod_excl_both,    //! exclude from both, dir and group
   Linemod_disable,      //! disable whole dir, group or single entry
   linemod_count };
typedef enum linemod_en linemod_en;

typedef struct st_lmod_def st_lmod_def;
/*!:: Line modification definition.
 *    Properties of modification (modification template).
 */
struct st_lmod_def
{
  const char* const name; //! name
  linemod_en mod;         //! line mod
  const char* const tag;  //! tag string
  const size_t tag_len;   //! num of chars in tag
};
extern const st_lmod_def StLinemodsDefinitions[linemod_count];







/*!:: Action kind.
 *    What operands it takes, how it can be written.
 */
enum actkind_en {
  Actkind_none=0,
  Actkind_dummy,      //! consume this actions when parsing but discard them
  Actkind_inactive,   //! ignore this actions when parsing
  Actkind_oper,       //! takes operand, may be preceeded by name, ex. name->oper
  Actkind_nooper,     //! no operand, may be preceeded/sucedeed by name, ex. * OR name*any
  Actkind_multioper,  //! may be mixed with many operands, ex. user:group:perms
  Actkind_clear,      //! always as separate word, ex. =
  Actkind_name,       //! tags sourrounding name
  actkind_count};
typedef enum actkind_en actkind_en;

/*!:: Action tags.
 *    Tags indicating action.
 * ::Usage:
 *    String representing action tag at begining of word in parsed line
 *    indicates action to perform.
 *    Word becomes action's operand. Action is performed on entry represented by
 *    this line.
 *    Action is performed only if this line is active (see effective tags).
 */
enum acttype_en {
   Acttype_none=0,    //! none
   //---- operations on filesystem ----
   Acttype_create,    //! create new entry
   Acttype_slink,     //! change symlink target / create sl
   Acttype_slink_o,   //! change symlink target / create sl, overwrite
   Acttype_hlink,     //! change hardlink target / create hl
   Acttype_hlink_o,   //! change hardlink target / create hl, overwrite

   Acttype_chperms,   //! change permissions
   Acttype_chuser,    //! change user
   Acttype_chgroup,   //! change group
   Acttype_prop,      //! 3 above combined into one tag

   Acttype_copy,     //! copy entry
   Acttype_copy_o,   //! copy entry
   Acttype_move,     //! move entry
   Acttype_move_o,   //! move entry
   Acttype_swnames,  //! swap two entries' names

   Acttype_trash,     //! move entry to wastebin
   Acttype_remove,    //! remove entry
   opertype_count,
   //---- special actions ----
   Actspec_name_beg,  //! group entries - begin
   Actspec_name_end,  //! group entries - end
   Actspec_da_cont,   //! directory actions relate to entries inside directory only
   Actspec_da_both,   //! directory actions relate to content and dir
   Actspec_da_ignor,  //! ignore following actions - toggle
   Actspec_ign_errors,  //! toggle stop on error
   Actspec_acts_target,//! toggle perform actions on last action's target (instead on entry)
   Actdummy_scan_slink,   //! to make slink not to be recognized as aother action with similar tag
   Actdummy_scan_errlink, //! to make slink not to be recognized as aother action with similar tag
   spectype_count,

   acttype_count};
typedef enum acttype_en acttype_en;




typedef struct st_act_def st_act_def;
/*!:: Action definition.
 *    Properties of single action (action type, action template, whatever).
 */
struct st_act_def
{
  const char* const name; //! name
  actkind_en kind;        //! action kind
  acttype_en type;        //! action type
  const char* const tag;  //! tag string
  const size_t tag_len;   //! num of chars in tag
};
extern const st_act_def StActionDefinitions[acttype_count];




#define STWORDCH_OPER_NUM 3
typedef struct st_word_ch st_word_ch;
/*!:: Word characteristics.
 *    Characteristics of one word in line.
 */
struct st_word_ch
{
   const st_act_def* p_act_def;         //! what action this word is
   str_t oper[STWORDCH_OPER_NUM]; //! operands values
   /* aux */
   st_word_ch* prev_file_word;   //! previous word that takes file as operand - needed for 'Actmod_tar_act'
   bool oper_prepared;           //! operand was prepared (dir prepended)
   bool ign_err;                  //! ignore errors in actions derived from this word
};
#define ST_WORD_CH_Z (st_word_ch){Acttype_none,Actspec_none,Actmod_none,STR_T_Z}


typedef struct st_line_ch st_line_ch;
/*!:: Entry line characteristics.
 *    Characteristics of one line in textual representation.
 */
struct st_line_ch
{
   size_t lnum_in_file;   //! line number in parsed file
   uint ind;              //! indentation of line
   const st_line_def* p_line_def;  //! what line this line is
   const st_lmod_def* p_lmod_def;  //! what mod this line has
   str_t ename;           //! entry name
   strarr_t fullnames;    //! names resulted from expanding braces, globbing and praparing entry name
   arr_ptrs_t arrWordsCh; //! line's words characteristics
   st_group_ch* p_grp;    //! pointer to containing group
   st_line_ch* p_dir;     //! pointer to parent directory (actions)
};
#define ST_LINE_CH_Z (st_line_ch){0,Linetype_none,Linemod_none,STR_T_Z,STR_T_Z,ARR_PTRS_T_Z,NULL,NULL}


//typedef struct st_group_ch st_group_ch; - declared earlier
/*!:: Group of entries characteristics.
 *    Characteristics of group of lines in textual representation.
 */
struct st_group_ch
{
   const st_line_def* p_line_def;  //! what line this group is
   const st_lmod_def* p_lmod_def;  //! what mod this group has
   arr_ptrs_t arrGroupsCh; //! nested groups
   arr_ptrs_t arrLinesCh;  //! lines in group
   arr_ptrs_t arrWordsCh;  //! words (actions) for whole group
   st_group_ch* p_grp_parent; //! pointer to parent group
};
#define ST_GROUP_CH_Z (st_group_ch){Effstr_none,ARR_PTRS_T_Z,ARR_PTRS_T_Z,ARR_PTRS_T_Z,NULL}


typedef struct st_txt_img st_txt_img;
/*!:: Textual representation.
 *    Textual representation of file system. Read from parsed file.
 */
struct st_txt_img
{
   arr_ptrs_t arrEntries;     //! table of lines in order encountered in parsed file - not storage, do not free pointed objects!
   st_group_ch* stGrpCh_Top;  //! top group in text rep
};
#define ST_TXT_IMG_Z (st_txt_img){ARR_PTRS_T_Z,NULL}




/* ...Main.parsing.struct.....................................*/


typedef struct st_parser_main st_parser_main;
/*!:: Main parser structure.
 *    Parsing settings and actual parsing state.
 */
struct st_parser_main
{
   /* parsing state */
   st_line_ch* p_cur_top_dir;     //! pointer to top directory (changed with '#')
   st_line_ch* p_cur_par_dir;    //! pointer to current parent directory
   st_group_ch* p_cur_group;  //! pointer to current group
   /* indent */
   uint ind_width;            //! width of indentation (spaces) used for subdirs
   /* record */
   strarr_t pars_record;      //! record of parsing
   /* faults */
   st_faults_arr ParsFaults;  //! array of pointers to parsing faults
};
#define ST_PARSER_MAIN_Z (st_parser_main){NULL,NULL,NULL,0,STRARR_T_Z,ST_FAULTS_ARR_Z}




/* Prepare routines. */
/* -------------------------------------------------------------------------- */



/*!1:: Cleanup struct internals. */
extern void pars_cleanup_parser_main (st_parser_main* stParsMain)
               __attr_nonnull__(1)
               ;

/*!1:: Cleanup some struct internals. */
extern void pars_cleanup_parser_main_next_run (st_parser_main* stParsMain)
               __attr_nonnull__(1)
               ;




/* Text rep data. */
/* -------------------------------------------------------------------------- */


/*!1:: Prepare data representing textual representation. */
extern void pars_prepare_txtrep (st_txt_img* stTxtRep, FILE* fs, const char* filename,
                  st_parser_main* stParsMain)
               __attr_nonnull__(1,2,3,4)
               ;

/*!1:: Cleanup struct internals. */
extern void pars_cleanup_txtrep (st_txt_img* stTxtRep)
               __attr_nonnull__(1)
               ;




#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool parsdata_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- TFMAN_SRCPARS_TXTREPTODATA_H -*/
