/*!F::File system data image - main implementation file.
 *
 *  ::Purpose:
 *    Create structure of data representing directory or whole tree of
 *    directories. For internal use by 'tfman' program.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of 'Tfman' program.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-12
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* readlink():  _BSD_SOURCE || _XOPEN_SOURCE >= 500
 *             || _XOPEN_SOURCE && _XOPEN_SOURCE_EXTENDED
 *             || / Since glibc 2.10: / _POSIX_C_SOURCE >= 200112L */
//#define _POSIX_C_SOURCE 200112L

/* strcasecmp():  _DEFAULT_SOURCE */
#define _DEFAULT_SOURCE




/* current project */
#include "common_tfman.h"
#include "opts.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "memory.h"
#include "cstrings.h"
#include "times.h"         // tm_stringize_time()
#include "dstream_info.h"       // ds_count_dir_entries()
#include "stat_info.h"     // stat_readlink()

/* libc */
#include <string.h>     // memset(), strcmp()
#include <stdio.h>      // printf(), putc()
#include <locale.h>     // setlocale()
#include <wchar.h>      // mbsrtowcs()

/* current module */
#include "scan_img.h"
#include "scan_img_internal.h.inc"
#include "scan_img_aux.h.inc"




/*!:: Start scanning in `dname` directory and go into subdirectories
 *    recursively as deep as specified in options. Compose and sort entries as
 *    specified in options. Place resulting file system image in `stFsImg`.
 * ::Errors:
 */
void
repimg_scan_tree (
   st_fsimg* stFsImg,      //! [out] file system image
   const char* dname,      //! name of top directory to scan
   st_tfman_opts* stOpts   //! scanning options
){
  ___Dassert(stFsImg, "'stFsImg parameter invalid");
  ___Dassert(dname, "'dname' parameter invalid");
  ___Dassert(stOpts, "'stOpts' parameter invalid");

  fsimg_zero_(stFsImg);
  stFsImg->stOpts = stOpts;
  str_ini(&stFsImg->filename, dname);
    HandleErr(handle_err);                    //NOMEM

  //: remove trailing '/' chars from dir name
  stat_remove_endslashess(&stFsImg->filename);


#if 0
  while (stFsImg->filename.buf[stFsImg->filename.nc] == '/') {
    stFsImg->filename.buf[stFsImg->filename.nc] = '\0';
    --stFsImg->filename.nc;
  }
#endif

  //: start scanning from top directory
  fsimg_scan_dir_(&stFsImg->n_entrs, &stFsImg->filename, &stFsImg->_subentry_first,
      stOpts->recurs_lvl, &stOpts->stSort, &stFsImg->_longest);
    HandleErr(handle_err);

  //no function resources to free
  return;

 handle_err:
  str_fin(&stFsImg->filename);
  return;
}


/*!:: Free some resources inside: entries tree and top directory name.
 * ::Preconditions:
 *  - `fs_image` must point to valid 'st_fsimg' struct, otherwise behavior is
 *    undefined
 * ::Errors:
 *    No errors.
 */
void
repimg_cleanup_entries (
   st_fsimg* stFsImg    //! file system image
){
   ___Dassert(stFsImg, "'stFsImg parameter invalid");
   if (stFsImg->_subentry_first != NULL) {
      imgent_destr_(&stFsImg->_subentry_first);
   }
   str_fin(&stFsImg->filename);
   return;
}

/*!:: Free all resources inside.
 * ::Preconditions:
 *  - `fs_image` must point to valid 'st_fsimg' struct, otherwise behavior is
 *    undefined
 * ::Errors:
 *    No errors.
 */
void
repimg_cleanup (
   st_fsimg* stFsImg    //! file system image
){
   ___Dassert(stFsImg, "'stFsImg parameter invalid");
   if (stFsImg->_subentry_first != NULL) { repimg_cleanup_entries(stFsImg); }
   return;
}
