/*!F::File system text representation - main implementation file.
 *
 *  ::Purpose:
 *    Create text representing directory or whole tree of directories, using
 *    file system data image. Each entry is represented by one line.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - create text representation of directory on the basis of image repr.
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of 'Tfman' program.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-02-12
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* mh library */
#include "core.h"
#include "errors.h"
#include "memory.h"
#include "cstrings.h"
#include "strings.h"
#include "strings_arr.h"
#include "stat_info.h"

/* current project */
#include "common_tfman.h"
#include "opts.h"
#include "scan_img.h"

/* libc */
#include <string.h>     // memset(), strcmp()
#include <stdio.h>      // printf(), putc()
#include <wchar.h>      // mbsrtowcs()

/* current module */
#include "scan_txt.h"
#include "scan_txt_internal.h.inc"
#include "scan_txt_aux.h.inc"




/*!:: Strings are used in text representation as tags to mark different directory
 *    layouts and different entry line states.
 *    Active entry - entry that will be scanned for action during parsing
 *    Inactive entry - will not be scanned even if contains actions
 * ::Preconditions:
 *    - 'st_fstxt' must point to valid struct
 * ::Errors:
 *    ENOMEM         -- no memory available
 */
void
fstxt_set_pref_strings (
   st_fstxt* stFsText,        //! [out] text representation struct
   const char* cs_dir_blk,    //! directory in block layout
   const char* cs_ent         //! inactive entry
){
   ___Dassert(stFsText, "'stFsText' parameter invalid");
   ___Dassert(cs_dir_blk, "'cs_dir_blk' parameter invalid");
   ___Dassert(cs_ent, "'cs_ent' parameter invalid");

   sarr_ini(&stFsText->prefstrs, cs_dir_blk);
   sarr_app(&stFsText->prefstrs, cs_ent);
      HandleErr(handle_err);

   return;

  handle_err:
   return;
}

/*!:: Strings are used in text representation as tags to mark different field
 *    types.
 * ::Preconditions:
 *    - 'st_fstxt' must point to valid struct
 * ::Errors:
 *    ENOMEM         -- no memory available
 */
void
fsrxt_set_field_strings (
   st_fstxt* stFsText,        //! [out] text representation struct
   const char* cs_name_beg,   //! begining of name
   const char* cs_name_end,   //! end of name
   const char* cs_tar,        //! target of symlink
   const char* cs_tar_nex,    //! target of symlink - non existant
   const char* cs_separ,      //! separator
   const char* cs_space,      //! space between fields
   const char* cs_atime,      //! access time
   const char* cs_mtime,      //! modificatin time
   const char* cs_ctime       //! change time
){
   ___Dassert(stFsText, "'stFsText' parameter invalid");
   ___Dassert(cs_name_beg, "'cs_name_beg' parameter invalid");
   ___Dassert(cs_name_end, "'cs_name_end' parameter invalid");
   ___Dassert(cs_tar, "'cs_tar' parameter invalid");
   ___Dassert(cs_tar_nex, "'cs_tar_nex' parameter invalid");
   ___Dassert(cs_separ, "'cs_separ' parameter invalid");
   ___Dassert(cs_space, "'cs_space' parameter invalid");
   ___Dassert(cs_atime, "'cs_atime' parameter invalid");
   ___Dassert(cs_mtime, "'cs_mtime' parameter invalid");
   ___Dassert(cs_ctime, "'cs_ctime' parameter invalid");

   sarr_ini(&stFsText->otherstrs, cs_name_beg);
   sarr_app(&stFsText->otherstrs, cs_name_end);
   sarr_app(&stFsText->otherstrs, cs_tar);
   sarr_app(&stFsText->otherstrs, cs_tar_nex);
   sarr_app(&stFsText->otherstrs, cs_separ);
   sarr_app(&stFsText->otherstrs, cs_space);
   sarr_app(&stFsText->otherstrs, cs_atime);
   sarr_app(&stFsText->otherstrs, cs_mtime);
   sarr_app(&stFsText->otherstrs, cs_ctime);
      HandleErr(handle_err);

   return;

  handle_err:
   return;
}

/*!:: Text is composed on the basis of file system image representation.
 * ::Preconditions:
 *    - 'st_fstxt' must point to valid struct
 *    - 'st_fsimg' must point to valid struct with created fs image
 * ::Errors:
 *    forge_dirname_line()    -- this function errors
 *    forge_other_lines()     -- this function errors
 */
void
fstxt_image_to_text (
   st_fstxt* stFsTxt,      //! [out] file system text representation
   st_fsimg* stFsImg       //! file system image
){
   ___Dassert(stFsTxt, "'stFsTxt parameter invalid");
   ___Dassert(stFsImg, "'stFsImg parameter invalid");

   /*: Init text representation */
   fstxt_init_(stFsTxt, stFsImg);

   /*: Forge text representation on the basis of fs image: */
   //forge_dirname_line(stFsTxt);
    //  HandleErr(handle_err);
   forge_other_lines(stFsTxt);
      HandleErr(handle_err);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/*!:: Release text lines resources, keep prefix and field strings.
 * ::Errors:
 *    No errors.
 */
void
fstxt_cleanup_text (
   st_fstxt* stFsTxt       //! file system text representation
){
   ___Dassert(stFsTxt, "'stFsTxt parameter invalid");
   fstxt_txt_finish_(stFsTxt);
   return;
}

/*!:: Release all resources and set members to empty values.
 * ::Errors:
 *    No errors.
 */
void
fstxt_cleanup (
   st_fstxt* stFsTxt       //! file system text representation
){
   ___Dassert(stFsTxt, "'stFsTxt parameter invalid");
   fstxt_finish_(stFsTxt);
   return;
}

/*!:: Prints dirname line and all entry lines. If verbose is at least 1 prints
 *    options used in text representation creation as first lines.
 * ::Errors:
 *    ENOMEM         -- no memory available
 */
void
fstxt_print (
   FILE* fs,               //! file stream
   st_fstxt* stFsTxt,      //! file system text representation
   uint verb,              //! how much info to print
   st_tfman_opts* stOpts   //! only cl options needed from this
){
   ___DParam(fs, != NULL);
   ___DParam(stFsTxt, != NULL);
   ___DParam(stOpts, != NULL);

   int ret = 0;

  if (verb >= 1) {
  /*: Print option lines: */
      const char** argv_cl = stOpts->argv[0];
      const int argc_cl = stOpts->argc[0];
      const char** argv_cf = stOpts->argv[1];
      const int argc_cf = stOpts->argc[1];

      ret = fprintf(fs, ":: cl options:");
      for (int i = 1; i < argc_cl; ++i) {
         ret = fprintf(fs, " %s", argv_cl[i]);
      }
      ret = fprintf(fs, "\n:: cf options:");
      for (int i = 1; i < argc_cf; ++i) {
        ret = fprintf(fs, " %s", argv_cf[i]);
      }
        ret = fprintf(fs, "\n");

  /*: Print actions remainder: */
    ret = fprintf(fs, "\n:: #Dir, |Entry, $Priority, {group} {{group dirs}}, !! no errors, END parsing\n");
    ret = fprintf(fs,   "::   = - E x         - line modifiers: excl. from dir/group/both, disable whole dir or group\n");
    ret = fprintf(fs,   ":: ~(create) slink-> hlink=> usr:grp:perms, copy> move>> *(remove) ...>* (overwrite)\n");
    ret = fprintf(fs,   "::   = == || ! ++    - special actions: content,entry and content,disable acts,no errors,on prev target\n");
  }
    ErrorTrigIf(ret < 0, handle_err);

  /*: Print dirname and all entry lines: */
  for (size_t i = 0; i < stFsTxt->nlns; ++i) {
    ret = fprintf(fs, "%s\n", stFsTxt->lines[i]);
  }
    ErrorTrigIf(ret < 0, handle_err);

   return;

  handle_err:
   return;
}
